@@ -1,16 +1,16 @@
-/***********************************************************************
+/******************************** -*- C -*- ****************************
  *
  *	Header for Smalltalk's xmalloc, xrealloc and xfree
  *
- *	$Revision: 1.6.2$
- *	$Date: 1999/08/31 11:23:18$
+ *	$Revision: 1.8.5$
+ *	$Date: 2000/12/27 10:45:49$
  *	$Author: pb$
  *
  ***********************************************************************/
 
 /***********************************************************************
  *
- * Copyright 1990, 91, 92, 94, 95, 99 Free Software Foundation, Inc.
+ * Copyright 1988-92, 1994-95, 1999, 2000 Free Software Foundation, Inc.
  * Written by Steve Byrne.
  *
  * This file is part of GNU Smalltalk.
@@ -31,8 +31,15 @@
  *
  ***********************************************************************/
 
+#ifndef __GSTALLOC__
+#define __GSTALLOC__
+
 char *		xmalloc();
 char *		xrealloc();
 void		xfree();
+void		nomemory();
 
-#define xcalloc(n,s)	xmalloc((n)*(s))
+#define obstack_chunk_alloc xmalloc
+#define obstack_chunk_free xfree
+
+#endif
\ No newline at end of file
