@@ -1,16 +1,18 @@
-/***********************************************************************
+/******************************** -*- C -*- ****************************
  *
- * Byte code array utility routines.
+ *	Byte code array utility routines.  This module also contains
+ *	optimization-related functions -- that is, everything where
+ *	knowing bytecode numbers is necessary.
  *
- * $Revision: 1.6.2$
- * $Date: 1999/08/31 11:23:18$
- * $Author: pb$
+ *	$Revision: 1.8.5$
+ *	$Date: 2000/12/27 10:45:49$
+ *	$Author: pb$
  *
  ***********************************************************************/
 
 /***********************************************************************
  *
- * Copyright 1990, 91, 92, 94, 95, 99 Free Software Foundation, Inc.
+ * Copyright 1988-92, 1994-95, 1999, 2000 Free Software Foundation, Inc.
  * Written by Steve Byrne.
  *
  * This file is part of GNU Smalltalk.
@@ -35,7 +37,9 @@
 #include "gst.h"
 #include "alloc.h"
 #include "byte.h"
+#include "comp.h"
 #include "dict.h"
+#include "memzero.h"
 #if STDC_HEADERS
 #include <string.h> /* for memcpy */
 #include <stdlib.h>
@@ -45,7 +49,7 @@
 /* Define this to disable the peephole bytecode optimizer.  It works
  * well and increases a bit performance, so there's no reason to do that
  * unless you're debugging the compiler. */
-#undef NO_OPTIMIZE
+/* #define NO_OPTIMIZE */
 
 
 #define BYTE_CODE_CHUNK_SIZE  64
@@ -68,7 +72,7 @@ int byteCodeSizeTable[256] = {
 1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1, /* 224 */
 1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1}; /* 240 */
 
-int isPush[256] = {
+static int isPush[256] = {
 1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1, /* 0 */
 1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1, /* 16 */
 1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1, /* 32 */
@@ -86,11 +90,20 @@ int isPush[256] = {
 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* 224 */
 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0}; /* 240 */
 
-ByteCodes  currentByteCodes = nil;
+static int  jumpOffsets[16] =  {
+  -1022, -766, -510, -254, 2, 258, 514, 770,
+      2,  258,  514,  770, 2, 258, 514, 770
+};
 
-static ByteCodes allocByteCodes();
-static mst_Boolean optimizeBasicBlock();
-static void  compileByteCodes(), reallocByteCodes();
+ByteCodes			currentByteCodes = nil;
+
+static ByteCodes		allocByteCodes();
+
+static mst_Boolean		optimizeBasicBlock();
+
+static void			compileByteCodes(), reallocByteCodes();
+
+static inline int		checkInnerBlock();
 
 
 /* These are used to decode special send byte codes. */
@@ -124,7 +137,7 @@ static char  *specialMessageName[] = {
   "atEnd",
   "==",
   "class",
-  "blockCopy:temporaries:",
+  "blockCopy:",
   "value",
   "value:",
   "do:",
@@ -134,6 +147,14 @@ static char  *specialMessageName[] = {
   "notNil"
 };
 
+struct ByteCodeArray {
+  Byte		*base;		/* base of the byte code array */
+  Byte		*ptr;		/* current byte+1 of byte code array */
+  int		maxLen;		/* max allocated len of byte code array */
+};				/* changed as byte code array grows, and */
+				/* used to tell when to grow the array. */
+
+
 /* This structure and the following one are used when optimizing the
  * bytecodes.
  *
@@ -156,9 +177,10 @@ typedef struct BlockBoundaryStruct {
  * the same way.
  * When single blocks are optimized, the sorted BlockBoundaries are
  * examined one at a time.  As we process blocks, we fill an array of
- * Jump structures with optimized bytecode offsets.  In the end, we'll
- * have paired BlockBoundaries whose id's absolute value is the same
- * (with different signs, of course).
+ * Jump structures with optimized bytecode offsets.  We fill a single
+ * `int' at a time -- the sign says which int is to be filled, the
+ * absolute value gives which Jump structure is to be filled. In the end,
+ * BlockBoundaries whose id's absolute value is the same are all paired.
  */
 typedef struct JumpStruct {
   int	from;  /* where the jump bytecode lies */
@@ -223,19 +245,19 @@ compileAndFreeByteCodes (byteCodes)
 }
 
 /*
- * ByteCodes getByteCodes()
+ *	ByteCodes getByteCodes()
  *
  * Description
  *
- * Called when byte code compilation is complete, this routine returns the
- * set of byte codes that were compiled.  Since compilation is complete,
- * this routine also resets the internal state of the byte code compiler
- * in preparation for next time.
+ *	Called when byte code compilation is complete, this routine returns the
+ *	set of byte codes that were compiled.  Since compilation is complete,
+ *	this routine also resets the internal state of the byte code compiler
+ *	in preparation for next time.
  *
  * Outputs
  *
- * A pointer to a byte code array that contains the bytes for the current
- * compilation.
+ *	A pointer to a byte code array that contains the bytes for the current
+ *	compilation.
  */
 ByteCodes
 getByteCodes ()
@@ -250,18 +272,18 @@ getByteCodes ()
 
 
 /*
- * ByteCodes saveByteCodeArray()
+ *	ByteCodes saveByteCodeArray()
  *
  * Description
  *
- * Called to save the set of byte codes currently being compiled and
- * prepare for a new compilation of byte codes.	 The current set of byte
- * codes being compiled is returned for the caller to keep and to later
- * use in a restoreByteCodeArray call.
+ *	Called to save the set of byte codes currently being compiled and
+ *	prepare for a new compilation of byte codes. The current set of byte
+ *	codes being compiled is returned for the caller to keep and to later
+ *	use in a restoreByteCodeArray call.
  *
  * Outputs
  *
- * A byte code array pointer to the current byte codes.
+ *	A byte code array pointer to the current byte codes.
  */
 ByteCodes
 saveByteCodeArray ()
@@ -276,20 +298,20 @@ saveByteCodeArray ()
 
 
 /*
- * void restoreByteCodeArray(byteCodes)
+ *	void restoreByteCodeArray(byteCodes)
  *
  * Description
  *
- * Restores the internal state of the byte code compiler so that it can
- * continue compiling byte codes into the byte code array "byteCodes",
- * which should have been returned at some previous point from
- * saveByteCodeArray().
+ *	Restores the internal state of the byte code compiler so that it can
+ *	continue compiling byte codes into the byte code array "byteCodes",
+ *	which should have been returned at some previous point from
+ *	saveByteCodeArray().
  *
  * Inputs
  *
- * byteCodes:
- *  The byte code array pointer to be restored.	 Should have come
- *  from a saveByteCodeArray call.
+ *	byteCodes:
+ *		The byte code array pointer to be restored. Should have come
+ *		from a saveByteCodeArray call.
  *
  */
 void
@@ -311,15 +333,15 @@ ByteCodes byteCodes;
 
 
 /*
- * int currentByteCodeLength()
+ *	int currentByteCodeLength()
  *
  * Description
  *
- * Return the current number of byte codes that have been compiled.
+ *	Return the current number of byte codes that have been compiled.
  *
  * Outputs
  *
- * Number of byte codes present in the current byte code array right now.
+ *	Number of byte codes present in the current byte code array right now.
  */
 int
 currentByteCodeLength ()
@@ -385,13 +407,20 @@ printIndexed (byte, literalVec, constOk)
     break;
   case 2:
     printf("Constant[%d]", byte & 63);
+    if (literalVec) {
+      printf(" = ");
+      printObject(literalVec[byte & 63]);
+    }
     if (!constOk) {
       printf(" -- INVALID!");
     }
     break;
   case 3:
-    printf("Global Variable[%d] = ", byte & 63);
-    printAssociationKey(literalVec[byte & 63]);
+    printf("Global Variable[%d]", byte & 63);
+    if (literalVec) {
+      printf(" = ");
+      printAssociationKey(literalVec[byte & 63]);
+    }
     break;
   }
 }
@@ -402,6 +431,8 @@ printByteCodeName (bp, ip, literalVec)
      int ip;
      OOP literalVec[];
 {
+  int lit;
+
   switch (bp[0]) {
   case	0: case	 1: case  2: case  3:
   case	4: case	 5: case  6: case  7:
@@ -426,6 +457,10 @@ printByteCodeName (bp, ip, literalVec)
   case 56: case 57: case 58: case 59:
   case 60: case 61: case 62: case 63:
     printf("push Literal[%d]", bp[0] & 31);
+    if (literalVec) {
+      printf(" = ");
+      printObject(literalVec[bp[0] & 31]);
+    }
     break;
 
   case 64: case 65: case 66: case 67:
@@ -436,8 +471,11 @@ printByteCodeName (bp, ip, literalVec)
   case 84: case 85: case 86: case 87:
   case 88: case 89: case 90: case 91:
   case 92: case 93: case 94: case 95:
-    printf("push Global Variable[%d] = ", bp[0] & 31);
-    printAssociationKey(literalVec[bp[0] & 31]);
+    printf("push Global Variable[%d]", bp[0] & 31);
+    if (literalVec) {
+      printf(" = ");
+      printAssociationKey(literalVec[bp[0] & 31]);
+    }
     break;
 
   case	96: case  97: case  98: case  99:
@@ -522,9 +560,13 @@ printByteCodeName (bp, ip, literalVec)
       break;
     }
     printf("[%d]", (bp[1] & 63) * 256 + bp[2]);
-    if (bp[1] >> 6) {
+    if (literalVec) {
       printf(" = ");
-      printAssociationKey(literalVec[(bp[1] & 63) * 256 + bp[2]]);
+      if (bp[1] >> 6) {
+	printAssociationKey(literalVec[(bp[1] & 63) * 256 + bp[2]]);
+      } else {
+	printObject(literalVec[(bp[1] & 63) * 256 + bp[2]]);
+      }
     }
     break;
 
@@ -548,23 +590,51 @@ printByteCodeName (bp, ip, literalVec)
     break;
 
   case 131:
-    printf("send selector %d, %d args = ", bp[1] & 31, bp[1] >> 5);
-    printSymbol(literalVec[bp[1] & 31]);
+    printf("send selector %d, %d args", bp[1] & 31, bp[1] >> 5);
+    if (literalVec) {
+      printf(" = ");
+      printSymbol(literalVec[bp[1] & 31]);
+    }
     break;
 
   case 132:
-    printf("send selector %d, %d args = ", bp[2], bp[1]);
-    printSymbol(literalVec[bp[2]]);
+    lit = bp[2] + ((bp[1] & 192) << 2);
+    printf("send %sselector %d, %d args",
+      bp[1] & extendedSendSuperFlag ? "to super " : "", lit, bp[1] & 31);
+    if (literalVec) {
+      printf(" = ");
+      printSymbol(literalVec[lit]);
+    }
     break;
 
   case 133:
-    printf("send to super selector %d, %d args = ", bp[1] & 31, bp[1] >> 5);
-    printSymbol(literalVec[bp[1] & 31]);
+    printf("send to super selector %d, %d args", bp[1] & 31, bp[1] >> 5);
+    if (literalVec) {
+      printf(" = ");
+      printSymbol(literalVec[bp[1] & 31]);
+    }
     break;
 
   case 134:
-    printf("send to super selector %d, %d args = ", bp[2], bp[1]);
-    printSymbol(literalVec[bp[2]]);
+    switch (bp[1] >> 6) {
+    case 0:
+      printf("pop and store");
+      break;
+    case 1:
+      printf("push");
+      break;
+    case 2:
+      printf("pop and store");
+      break;
+    case 3:
+      printf("store");
+      break;
+    }
+
+    printf(" Instance Variable[%d]", (bp[1] & 63) * 256 + bp[2]);
+    if ((bp[1] >> 6) == 0) {
+      printf(" of new stack top");
+    }
     break;
 
   case 135:
@@ -616,7 +686,7 @@ printByteCodeName (bp, ip, literalVec)
     break;
 
   case 143:
-    printf("Invalid bytecode %d!!!", bp[0]);
+    printf("terminate interpreter");
     break;
 
   case 144: case 145: case 146: case 147:
@@ -660,24 +730,37 @@ printByteCodeName (bp, ip, literalVec)
   case 212: case 213: case 214: case 215:
   case 216: case 217: case 218: case 219:
   case 220: case 221: case 222: case 223:
-    printf("send selector %d, 0 args = ", bp[0] & 15);
-    printSymbol(literalVec[bp[0] & 15]);
+    printf("send selector %d, 0 args", bp[0] & 15);
+    if (literalVec) {
+      printf(" = ");
+      printSymbol(literalVec[bp[0] & 15]);
+    }
     break;
 
   case 224: case 225: case 226: case 227:
   case 228: case 229: case 230: case 231:
   case 232: case 233: case 234: case 235:
   case 236: case 237: case 238: case 239:
-    printf("send selector %d, 1 arg = ", bp[0] & 15);
-    printSymbol(literalVec[bp[0] & 15]);
+    printf("send selector %d, 1 arg", bp[0] & 15);
+    if (literalVec) {
+      printf(" = ");
+      printSymbol(literalVec[bp[0] & 15]);
+    }
     break;
 
   case 240: case 241: case 242: case 243:
   case 244: case 245: case 246: case 247:
   case 248: case 249: case 250: case 251:
   case 252: case 253: case 254: case 255:
-    printf("send selector %d, 2 args = ", bp[0] & 15);
-    printSymbol(literalVec[bp[0] & 15]);
+    printf("send selector %d, 2 args", bp[0] & 15);
+    if (literalVec) {
+      printf(" = ");
+      printSymbol(literalVec[bp[0] & 15]);
+    }
+    break;
+
+  default:
+    printf("INVALID BYTECODE %d", bp[0]);
     break;
   }
 }
@@ -706,13 +789,13 @@ isSimpleReturn (byteCodes)
   /* check for ^instanceVariable */
   if (byteCodeLen == 2) {
     if ((bytes[0] & ~15) == pushReceiverVariable
-	&& bytes[1] == returnMethodStackTop) {
+	&& bytes[1] == returnContextStackTop) {
       return (((bytes[0] & 0x0F) << 8) | 2);
     }
   } else if (byteCodeLen == 3) {
     if (bytes[0] == pushIndexed
 	&& (bytes[1] & locationMask) == receiverLocation
-	&& bytes[2] == returnMethodStackTop) {
+	&& bytes[2] == returnContextStackTop) {
       return (((bytes[1] & ~locationMask) << 8) | 2);
     }
   }
@@ -734,7 +817,7 @@ isSimpleReturn (byteCodes)
   }
 
   if (byteCodeLen == 2) {
-    if (bytes[1] != returnMethodStackTop) {
+    if (bytes[1] != returnContextStackTop) {
       return (0);
     }
 
@@ -759,26 +842,224 @@ isSimpleReturn (byteCodes)
 }
 
 
-
 /*
- * static ByteCodes optimizeByteCodes(byteCodes)
+ *	void makeDestinationTable(byteCodes, size, destination)
  *
  * Description
  *
- * Divide the byte codes of a method in basic blocks, optimize each one,
- * reunite the optimized blocks and return a new vector of byte codes that
- * represents the optimized byte code stream.  Actual optimization of the
- * basic blocks is optimizeBasicBlock's task.
+ *	Fill a table that says to which bytecodes a jump lands.
  *
  * Inputs
  *
- * byteCodes:
- *  A vector of byte codes to be optimized.
+ *	bp:
+ *		A pointer to some byte codes for which the destination table
+ *		must be computed.
+ *	size:
+ *		The size of the bytecodes
+ *	dest:
+ *		A pointer to char; on output, dest[i] is true if and
+ *		only if bp[i] is the destination of a jump.
  *
  * Outputs
  *
- * *ANOTHER* vector of byte codes (even though it currently contains the
- * same bytecodes that were passed in).
+ *	dest is filled.
+ */
+
+void
+makeDestinationTable(bp, size, dest)
+     Byte	 *bp;
+     int	  size;
+     char	 *dest;
+{
+  Byte *end;
+  
+  memzero(dest, sizeof(char) * size);
+
+  for(end = bp + size; bp != end; 
+      dest += byteCodeSize(*bp), bp += byteCodeSize(*bp)) {
+
+    if (*bp >= 144) {
+      if (*bp < 160) {
+	/* short jumps */
+	dest[(*bp & 7) + 2] = true;
+      } else if (*bp < 176) {
+	/* long jumps */
+	dest[((signed int)bp[1]) + jumpOffsets[*bp & 15]] = true;
+      }
+    }
+  }
+}
+
+
+/*
+ *	int checkKindOfBlock(bc, literals)
+ *
+ * Description
+ *
+ *	Look whether the block compiled in the given bytecodes can be
+ *	optimized.  Answer a number that identifies the possibility to
+ *	optimize it.
+ *
+ * Inputs
+ *
+ *	bc:
+ *		The bytecodes to be examined.
+ *	literals:
+ *		The literals array for the block.
+ *
+ * Outputs
+ *
+ *	0:	The block is a clean block
+ *	1:	The block is self-contained but accesses the receiver
+ *	31:	The block contains a return from method or push thisContext
+ *	else	The result is the number of outer contexts which the receiver
+ *		needs access to, + 1.
+ */
+int
+checkKindOfBlock(bc, literals)
+     ByteCodes	bc;
+     OOP	*literals;
+{
+  int		status, newStatus;
+  Byte		*bp, *end;
+  OOP		blockClosureOOP;
+
+  status = 0;	/* clean block */
+  for (bp = bc->base, end = bc->ptr; bp != end; bp += byteCodeSize(*bp)) {
+    switch (*bp) {
+      case 0:   case 1:   case 2:   case 3:	/* push instance variable */
+      case 4:   case 5:   case 6:   case 7:
+      case 8:   case 9:   case 10:  case 11:
+      case 12:  case 13:  case 14:  case 15:
+      case 96:  case 97:  case 98:  case 99:	/* pop into instance var */
+      case 100: case 101: case 102: case 103:
+      case 112: case 120: case 140:		/* push self/return/set top */
+	if (status == 0) {
+	  status = 1;
+	}
+	break;
+
+      case 32: case 33: case 34: case 35:	/* push literal constant */
+      case 36: case 37: case 38: case 39:
+      case 40: case 41: case 42: case 43:
+      case 44: case 45: case 46: case 47:
+      case 48: case 49: case 50: case 51:
+      case 52: case 53: case 54: case 55:
+      case 56: case 57: case 58: case 59:
+      case 60: case 61: case 62: case 63:
+	newStatus = checkInnerBlock(literals[*bp & 31]);
+	if (newStatus > status) {
+	  if (newStatus == 31) {
+	    return (31);
+	  }
+	  status = newStatus;
+	}
+	break;
+
+      case 124:					/* return from method */
+        return (31);
+
+      case 126:					/* big literal operations */
+	if ((bp[1] & locationMask) != pushLiteral) {
+	  continue;
+	}
+	blockClosureOOP = literals[((bp[1] & ~locationMask) << 6) | bp[2]];
+	newStatus = checkInnerBlock(blockClosureOOP);
+	if (newStatus > status) {
+	  if (newStatus == 31) {
+	    return (31);
+	  }
+	  status = newStatus;
+	}
+	break;
+
+      case 137:					/* push this context */
+        if (bp[1] != blockCopyColonSpecial) {
+          return (31);
+        }
+        break;
+
+      case 128: case 129: case 130: case 142:	/* 2-byte stack ops */
+	if ((bp[1] & locationMask) == receiverLocation && status == 0) {
+	  status = 1;
+	} else if ((bp[1] & locationMask) == litConstLocation) {
+	  newStatus = checkInnerBlock(literals[bp[1] & ~locationMask]);
+	  if (newStatus > status) {
+	    if (newStatus == 31) {
+	      return (31);
+	    }
+	    status = newStatus;
+	  }
+	}
+	break;
+
+      case 138:					/* outer temp operation */
+	if (status < (1 + bp[2])) {
+	  status = 1 + bp[2];
+	  if (status > 31) {			/* ouch! how deep!! */
+	    return (31);
+	  }
+	}
+	break;
+    }
+  }
+  return (status);
+}
+
+int
+checkInnerBlock(blockClosureOOP)
+     OOP	blockClosureOOP;
+{
+  int		newStatus;
+  BlockClosure	blockClosure;
+  Block		block;
+
+  if (!isClass(blockClosureOOP, blockClosureClass)) {
+    return (0);
+  }
+
+  /* This case is the most complicated -- we must check the
+   * cleanness of the inner block and adequately change the status.
+   *
+   * full block:	no way dude -- exit immediately
+   * clean block:     	same for us
+   * receiver access:	same for us
+   * access to temps in the Xth context: from the perspective of the
+   * 		    block being checked here, it is like an access to
+   * 		    temps in the (X-1)th context
+   * access to this block's temps: our outerContext can be nil either,
+   *		    but for safety we won't be a clean block.
+   */
+  blockClosure = (BlockClosure) oopToObj(blockClosureOOP);
+  block = (Block) oopToObj(blockClosure->block);
+  newStatus = block->header.clean;
+  switch(newStatus) {
+    case 31: case 0: case 1:	return(newStatus);
+    default:			return(newStatus - 1);
+  }
+}
+
+
+/*
+ *	ByteCodes optimizeByteCodes(byteCodes)
+ *
+ * Description
+ *
+ *	Divide the byte codes of a method in basic blocks, optimize each one,
+ *	reunite the optimized blocks and return a new vector of byte codes
+ *	that contains the stream of optimized byte codes.  Actual optimization
+ *	of the basic blocks is optimizeBasicBlock's task; this function
+ *	takes care of handling basic blocks and optimizing jumps (which span
+ *	multiple basic blocks).
+ *
+ * Inputs
+ *
+ *	byteCodes:
+ *		A vector of byte codes to be optimized.
+ *
+ * Outputs
+ *
+ *	*ANOTHER* vector of byte codes.
  */
 
 static int compare_blocks();
@@ -794,15 +1075,11 @@ ByteCodes
 optimizeByteCodes (byteCodes)
      ByteCodes byteCodes;
 {
-  static int  jumpOffsets[] =  {
-    -1022, -766, -510, -254, 2, 258, 514, 770,
-	2,  258,  514,	770, 2, 258, 514, 770
-  };
   register BlockBoundary *blocks, *current;
-  register Jump	  *jumps;
-  register Byte	  *bp;
-  Byte	  *end, *first;
-  int	 num;
+  register Jump		 *jumps;
+  register Byte		 *bp;
+  Byte	  		 *end, *first;
+  int			 num;
 
 #ifdef NO_OPTIMIZE
   return (byteCodes);
@@ -845,7 +1122,7 @@ optimizeByteCodes (byteCodes)
 	  canOptimizeJump = (bp != dest);
 	  dest += *dest;
 	  dest -= 142;
-	  canOptimizeJump |= (*dest >= 120 || *dest <= 125);
+	  canOptimizeJump |= (*dest >= 120 && *dest <= 125);
 	  break;
 
 	/* pop and short jump if false */
@@ -876,15 +1153,13 @@ optimizeByteCodes (byteCodes)
 
 	case 120: case 121: case 122: case 123:
 	case 124: case 125:
-	  /* Return bytecode - patch the original jump to return directly */
-	  if (bp[-1] != blockCopyColonSpecial) {
-	    if (*bp >= jumpLong) {
-	      bp[1] = *dest;		/* fill both 2 bytes */
-	    }
-	    *bp = *dest;
-	    /* This in fact eliminated the jump, don't split in basic blocks */
-	    dest = bp;
+	  /* Return bytecodes - patch the original jump to return directly */
+	  if (*bp >= jumpLong) {
+	    bp[1] = *dest;		/* fill both 2 bytes */
 	  }
+	  *bp = *dest;
+	  /* This in fact eliminated the jump, don't split in basic blocks */
+	  dest = bp;
 	  break;
       }
     } while (canOptimizeJump);
@@ -968,23 +1243,25 @@ optimizeByteCodes (byteCodes)
  ***********************************************************************/
 
 /*
- * static ByteCodes optimizeBasicBlock(from, to)
+ *	static ByteCodes optimizeBasicBlock(from, to)
  *
  * Description
  *
- * Scan the bytecodes between the given addresses, and compile an
- * optimized version by calling compileByte().
+ *	Scan the bytecodes between the given addresses. As they are
+ *	scanned, they are overwritten with an optimized version; in
+ *	the end, compileByteCodes() is used to append them to the
+ *	stream of optimized bytecodes.
  *
  * Inputs
  *
- * from:
- *  Pointer to the first bytecode to be optimized.
- * to:
- *  Pointer past the last bytecode to be optimized.
+ *	from:
+ *		Pointer to the first bytecode to be optimized.
+ *	to:
+ *		Pointer past the last bytecode to be optimized.
  *
  * Outputs
  *
- * Return whether at least a bytecode was generated.
+ *	Return whether at least a bytecode was generated.
  */
 
 #define NEXT(size) {	\
@@ -1027,13 +1304,12 @@ optimizeBasicBlock (from, to)
   while (bp != to) {
     byte = *opt = *bp++;
     BEGIN
-      RANGE(120, 125)
-	/* this performs unreachable code optimization! */
+      RANGE(returnIndexed, returnContextStackTop)
 	opt++;
-	break;
+	break; /* this `break' performs unreachable code elimination! */
 
       BYTECODE(nopBytecode)
-	REPLACE(n);			/* used when compiling yourself */
+	REPLACE(n);
 
       BYTECODE(notEqualSpecial)
 	if (!n) NEXT(1);
