@@ -1,16 +1,16 @@
-/***********************************************************************
+/******************************** -*- C -*- ****************************
  *
  *	C - Smalltalk Interface module
  *
- *	$Revision: 1.6.2$
- *	$Date: 1999/08/31 11:23:18$
+ *	$Revision: 1.8.5$
+ *	$Date: 2000/12/27 10:45:49$
  *	$Author: pb$
  *
  ***********************************************************************/
 
 /***********************************************************************
  *
- * Copyright 1990, 91, 92, 94, 95, 99 Free Software Foundation, Inc.
+ * Copyright 1988-92, 1994-95, 1999, 2000 Free Software Foundation, Inc.
  * Written by Steve Byrne.
  *
  * This file is part of GNU Smalltalk.
@@ -35,6 +35,7 @@
 #include "gst.h"
 #include "alloc.h"
 #include "interp.h"
+#include "comp.h"
 #include "dict.h"
 #include "oop.h"
 #include "sym.h"
@@ -151,9 +152,11 @@ typedef union CParamUnionUnion {
 } CParamUnion;
 
 /* Element type for the name-to-C-function mapping table. */ 
+typedef void (*pVoidFunc)();
+
 typedef struct 	CFuncInfoStruct {
   char		*funcName;
-  void		(*funcAddr)();
+  pVoidFunc	funcAddr;
 } CFuncInfo;
 
 typedef struct GstStatStruct {
@@ -166,7 +169,6 @@ typedef struct GstStatStruct {
 
 
 
-extern mst_Boolean 		enableGC;
 extern void			initUserCFuncs();
 
 
@@ -205,8 +207,8 @@ static char			*cTypeName[] = {
   "...",			/* variadicType */
   "void *",			/* cObjectType -- this is misleading */
   "void *",			/* cObjectPtrType */
-  "void *",			/* smalltalkType */
-  "self",			/* selfType */
+  "OOP",			/* smalltalkType */
+  "OOP",			/* selfType */
 };
 
 static SymbolTypeMap	symbolTypeMap[] = {
@@ -252,6 +254,10 @@ static int		typeSizes[] = {
   sizeof(double)		/* doubleType */
 };
 
+/* brd Sun Jun 25 14:16:31 GMT-0800 1995 */
+/* When true, enable GC of Smalltalk objects referenced in callins/callouts. */
+mst_Boolean enableGC = false;
+
 
 
 /*
@@ -311,7 +317,8 @@ my_stat (name, out)
   static struct stat	statOut;
 
   result = stat(name, &statOut);
-  if (!errno) {
+  if (result == 0) {
+    errno = 0;
     out->st_mode  = statOut.st_mode;
     out->st_size  = statOut.st_size;
     out->st_aTime = adjustTimeZone(statOut.st_atime) - 86400*10959;
@@ -331,7 +338,8 @@ my_stat (name, out)
 /**/  static struct stat	statOut;
 /**/
 /**/  result = lstat(name, &statOut);
-/**/  if (!errno) {
+/**/  if (result != -1) {
+/**/    errno = 0;
 /**/    out->st_mode  = statOut.st_mode;
 /**/    out->st_size  = statOut.st_size;
 /**/    out->st_aTime = adjustTimeZone(statOut.st_atime) - 86400*10959;
@@ -368,15 +376,38 @@ my_putenv (str)
      char	*str;
 {
   char		*clone;
-  int		len;
 
-  len = strlen(str) + 1;	/* hold the null */
-  clone = (char *)xmalloc(len);
-  strcpy(clone, str);
+  clone = strdup(str);
   return (putenv(clone));
 }
 
 
+/*
+ *	static int my_chdir(str)
+ *
+ * Description
+ *
+ *	Does a chdir library call.  Exists because chdir under Cygwin can
+ *	set errno even if it succeeds.
+ *
+ * Inputs
+ *
+ *	str   : Path to new current directory.
+ *
+ * Outputs
+ *
+ *	Returned value from chdir() call.
+ */
+int my_chdir(char *dir)
+{
+  int status;
+
+  status = chdir(dir);
+ 
+  if (status==0) errno=0;
+  return(status);
+}
+
 void
 testCallin (oop)
      OOP	oop;
@@ -455,6 +486,7 @@ initCFuncs ()
   defineCFunc("putenv", my_putenv);
 
   defineCFunc("errno", getErrno);
+  defineCFunc("strerror", strerror);
   defineCFunc("stat", my_stat);
   defineCFunc("lstat", my_lstat);
 
@@ -467,9 +499,13 @@ initCFuncs ()
   defineCFunc("unlink", unlink);
   defineCFunc("rename", rename);
   defineCFunc("rmdir", rmdir);
-  defineCFunc("chdir", chdir);
+  defineCFunc("chdir", my_chdir);
+  defineCFunc("mkdir", mkdir);
   defineCFunc("getCurDirName", getCurDirName);
 
+  defineCFunc("fileIsReadable", fileIsReadable);
+  defineCFunc("fileIsWriteable", fileIsWriteable);
+
   /* Define stubs, at least */
   initDldLib();
 
@@ -551,7 +587,7 @@ defineCFunc (funcName, funcAddr)
 
 
 /*
- *	void (*lookupFunction(funcName))()
+ *	pVoidFunc lookupFunction(funcName)
  *
  * Description
  *
@@ -566,10 +602,10 @@ defineCFunc (funcName, funcAddr)
  *
  * Outputs
  *
- *	Address of a C function, or nil if non was registered with "funcName"
+ *	Address of a C function, or nil if none was registered as "funcName"
  */
-void
-(*lookupFunction (funcName))()
+pVoidFunc
+lookupFunction (funcName)
      char	*funcName;
 {
   CFuncInfo	*fip;
@@ -679,9 +715,10 @@ CFuncDescriptor
 getCFuncDescriptor (methodOOP)
      OOP	methodOOP;
 {
-  OOP		associationOOP, descOOP;
+  register OOP		associationOOP, descOOP, literalsOOP;
 
-  associationOOP = oopToObj(((Method)oopToObj(methodOOP))->literals)->data[0];
+  literalsOOP = ((Method)oopToObj(methodOOP))->literals;
+  associationOOP = oopToObj(literalsOOP)->data[0];
   descOOP = associationValue(associationOOP);
   return ((CFuncDescriptor)oopToObj(descOOP));
 }
@@ -733,13 +770,7 @@ pushSmalltalkObj (oop, cType)
     return;
   }
 
-  if (isInt(oop)) {
-    class = integerClass;
-  } else if (oop == trueOOP || oop == falseOOP) {
-    class = booleanClass;
-  } else {
-    class = oopClass(oop);
-  }
+  class = isInt(oop) ? smallIntegerClass : oopClass(oop);
 
   if (cType == smalltalkType) {
     u.ptrVal = (voidPtr)oop;
@@ -756,17 +787,17 @@ pushSmalltalkObj (oop, cType)
     } else {
       errorf("Attempted to pass a non-CObject as a cObjectPtr");
     }
-  } else if (class == integerClass) {
+  } else if (isAKindOf(class, integerClass)) {
     if (cType == longType || cType == unknownType) {
-      u.longVal = toInt(oop);
+      u.longVal = toCLong(oop);
       pushObj(&u, longAlign);
     } else if (cType == intType || cType == charType) {
-      u.intVal = toInt(oop);
+      u.intVal = toCInt(oop);
       pushObj(&u, intAlign);
     } else {
-      badType("Integer", cType);
+      badType(class, cType);
     }
-  } else if (class == booleanClass) {
+  } else if (oop == trueOOP || oop == falseOOP) {
     if (cType == intType || cType == booleanType || cType == charType || cType == unknownType) {
       u.intVal = (oop == trueOOP);
       pushObj(&u, intAlign);
@@ -774,31 +805,36 @@ pushSmalltalkObj (oop, cType)
       u.longVal = (oop == trueOOP);
       pushObj(&u, longAlign);
     } else {
-      badType("Boolean", cType);
+      badType(class, cType);
     }
   } else if (class == charClass) {
     if (cType == charType || cType == unknownType) {
       u.intVal = charOOPValue(oop);
       pushObj(&u, intAlign);
     } else {
-      badType("Character", cType);
+      badType(class, cType);
     }
   } else if (class == stringClass) {
     if (cType == stringType || cType == stringOutType
+        || cType == byteArrayType || cType == byteArrayOutType
 	|| cType == unknownType) {
       if (sip - stringInfoBase >= ARG_VEC_SIZE) {
 	errorf("Too many string arguments, max is %d.  Extras ignored.",
 	       ARG_VEC_SIZE);
 	return;
       }
-      sip->cString = (Byte *)toCString(oop);
+      if (cType == byteArrayType || cType == byteArrayOutType) {
+	sip->cString = toByteArray(oop);
+      } else {
+        sip->cString = (Byte *)toCString(oop);
+      }
       u.ptrVal = (voidPtr)sip->cString;
       sip->stringOOP = oop;
       sip->returnedType = cType;
       sip++;
       pushObj(&u, ptrAlign);
     } else {
-      badType("String", cType);
+      badType(class, cType);
     }
   } else if (class == symbolClass) {
     if (cType == symbolType || cType == stringType || cType == unknownType) {
@@ -814,24 +850,29 @@ pushSmalltalkObj (oop, cType)
       sip++;
       pushObj(&u, ptrAlign);
     } else {
-      badType("Symbol", cType);
+      badType(class, cType);
     }
   } else if (class == byteArrayClass) {
-    if (cType == byteArrayType || cType == byteArrayOutType
+    if (cType == stringType || cType == stringOutType
+        || cType == byteArrayType || cType == byteArrayOutType
 	|| cType == unknownType) {
       if (sip - stringInfoBase >= ARG_VEC_SIZE) {
 	errorf("Too many string arguments, max is %d.  Extras ignored.",
 	       ARG_VEC_SIZE);
 	return;
       }
-      sip->cString = toByteArray(oop);
+      if (cType == stringType || cType == stringOutType) {
+        sip->cString = (Byte *)toCString(oop);
+      } else {
+	sip->cString = toByteArray(oop);
+      }
       u.ptrVal = (voidPtr)sip->cString;
       sip->stringOOP = oop;
       sip->returnedType = cType;
       sip++;
       pushObj(&u, ptrAlign);
     } else {
-      badType("ByteArray", cType);
+      badType(class, cType);
     }
   } else if (class == floatClass) {
     if (cType == doubleType || cType == unknownType) {
@@ -841,14 +882,14 @@ pushSmalltalkObj (oop, cType)
       u.floatVal = (float)floatOOPValue(oop);
       pushObj(&u, floatAlign);
     } else {
-      badType("Float", cType);
+      badType(class, cType);
     }
   } else if (class == cObjectClass) { 
     if (cType == cObjectType || cType == unknownType) {
       u.ptrVal = cObjectValue(oop);
       pushObj(&u, ptrAlign);
     } else {
-      badType("CObject", cType);
+      badType(class, cType);
     }
   } else if ((cType == cObjectType || cType == unknownType)
 	     && isAKindOf(class, cObjectClass)) {
@@ -857,7 +898,10 @@ pushSmalltalkObj (oop, cType)
   } else if (class == undefinedObjectClass) { /* how to encode nil */
     switch (cType) {
     case cObjectType:
+    case byteArrayType:
+    case byteArrayOutType:
     case stringType:
+    case stringOutType:
     case symbolType:
     case unknownType:
       u.ptrVal = nil;
@@ -865,7 +909,7 @@ pushSmalltalkObj (oop, cType)
       break;
 
     default:
-      badType("UndefinedObject", cType);
+      badType(class, cType);
     }
   } else if (cType == unknownType) {
     u.ptrVal = (voidPtr)oop;
@@ -881,10 +925,10 @@ pushSmalltalkObj (oop, cType)
         pushSmalltalkObj(arrayAt(oop, i), unknownType);
       }
     } else {
-      badType("Array", cType);
+      badType(class, cType);
     }
   } else {
-    badType("Unrecognized class", cType);
+    badType(class, cType);
   }
 }
 
@@ -920,13 +964,13 @@ callCFunction (desc)
   int		intResult;
   long		longResult;
   double	doubleResult;
-  int		(*cFunction)();
+  pVoidFunc	cFunction;
   CDataType	returnType;
   OOP		returnTypeOOP;
 
   errno = 0;
 
-#define callCFunction(type)		\
+#define doTheCall(type)							\
 	(* (type (*)()) cFunction)(					\
 	      cArgVec[0],  cArgVec[1],  cArgVec[2],  cArgVec[3],	\
 	      cArgVec[4],  cArgVec[5],  cArgVec[6],  cArgVec[7],	\
@@ -934,7 +978,7 @@ callCFunction (desc)
 	      cArgVec[12], cArgVec[13], cArgVec[14], cArgVec[15],	\
 	      cArgVec[16], cArgVec[17], cArgVec[18], cArgVec[19]);
 
-  cFunction = (int (*)())cObjectValue(desc->cFunction);
+  cFunction = (pVoidFunc)cObjectValue(desc->cFunction);
   if (!cFunction) {
     char *funcName;
     funcName = toCString(desc->cFunctionName);
@@ -954,26 +998,26 @@ callCFunction (desc)
     
   switch (returnType) {
   case voidType:
-    callCFunction(void);
+    doTheCall(void);
     break;
 
   case charType:
-    intResult = callCFunction(int);
+    intResult = doTheCall(int);
     setStackTop(charOOPAt((Byte)intResult));
     break;
 
   case booleanType: 
-    intResult = callCFunction(int);
+    intResult = doTheCall(int);
     setStackTopBoolean(intResult);
     break;
 
   case intType:
-    intResult = callCFunction(int);
+    intResult = doTheCall(int);
     setStackTopInt((long)intResult);
     break;
 
   case longType:
-    longResult = callCFunction(long);
+    longResult = doTheCall(long);
     setStackTop(fromInt(longResult));
     break;
 
@@ -982,7 +1026,7 @@ callCFunction (desc)
   case symbolType:
   case cObjectType:
   case smalltalkType:
-    longResult = callCFunction(long);
+    longResult = doTheCall(long);
     if (longResult == 0) {
       setStackTop(nilOOP);
       break;
@@ -1010,12 +1054,12 @@ callCFunction (desc)
     break;
 
   case doubleType:
-    doubleResult = callCFunction(double);
+    doubleResult = doTheCall(double);
     setStackTop(floatNew(doubleResult));
     break;
 
   case floatType:
-    doubleResult = (double) callCFunction(float);
+    doubleResult = (double) doTheCall(float);
     setStackTop(floatNew(doubleResult));
     break;
 
@@ -1036,12 +1080,33 @@ callCFunction (desc)
 }
 
 void
-badType (smalltalkTypeName, cType)
-     char	*smalltalkTypeName;
-     CDataType cType;
+badType (classOOP, cType)
+     OOP	classOOP;
+     CDataType	cType;
 {
-  errorf("Attempt to pass a %s as a %s", smalltalkTypeName,
+  char	smalltalkTypeName[256], *className;
+  Class	classObject;
+
+  if (isAMetaclass(classOOP)) {
+    classOOP = findAnInstance(classOOP);
+    classObject = (Class) oopToObj(classOOP);
+    if (isNil(classOOP)) {
+      strcpy(smalltalkTypeName, "an unnamed class object");	/* we're a nameless class */
+    } else {
+      className = toCString(classObject->name);
+      sprintf(smalltalkTypeName, "the %s class object", className);
+      xfree(className);
+    }
+  } else {
+    classObject = (Class) oopToObj(classOOP);
+    className = toCString(classObject->name);
+    sprintf(smalltalkTypeName, "an instance of %s", className);
+    xfree(className);
+  }
+
+  errorf("Attempt to pass %s as a %s", smalltalkTypeName,
 	 cTypeName[ENUM_INT(cType)]);
+  showBacktrace();
 }
 
 
@@ -1076,7 +1141,7 @@ makeDescriptor (funcNameOOP, returnTypeOOP, argsOOP)
      OOP	funcNameOOP, returnTypeOOP, argsOOP;
 {
   char		*funcName;
-  void		(*funcAddr)();
+  pVoidFunc	funcAddr;
   int		numArgs, i;
   CFuncDescriptor desc;
   IncPtr	incPtr;
@@ -1174,7 +1239,7 @@ restoreCFuncDescriptor (cFuncDescOOP)
      OOP	cFuncDescOOP;
 {
   CFuncDescriptor desc;
-  void		(*funcAddr)();
+  pVoidFunc	funcAddr;
   char		*funcName;
 
   desc = (CFuncDescriptor)oopToObj(cFuncDescOOP);
