@@ -1,16 +1,16 @@
-/***********************************************************************
+/******************************** -*- C -*- ****************************
  *
  *	Dictionary Support Module.
  *
- *	$Revision: 1.6.2$
- *	$Date: 1999/08/31 11:23:18$
+ *	$Revision: 1.8.5$
+ *	$Date: 2000/12/27 10:45:49$
  *	$Author: pb$
  *
  ***********************************************************************/
 
 /***********************************************************************
  *
- * Copyright 1990, 91, 92, 94, 95, 99 Free Software Foundation, Inc.
+ * Copyright 1988-92, 1994-95, 1999, 2000 Free Software Foundation, Inc.
  * Written by Steve Byrne.
  *
  * This file is part of GNU Smalltalk.
@@ -36,6 +36,7 @@
 #include "alloc.h"
 #include "dict.h"
 #include "oop.h"
+#include "cint.h"
 #include "interp.h"
 #include "str.h"
 #include "lib.h"
@@ -74,22 +75,24 @@ typedef struct ClassInfoStruct {
 OOP			mst_objectClass, magnitudeClass, charClass, timeClass,
 			dateClass, fractionClass, contextPartClass,
 			numberClass, floatClass, integerClass,
+			smallIntegerClass,
 			largeIntegerClass, largeNegativeIntegerClass,
 			largeZeroIntegerClass, largePositiveIntegerClass,
 			associationClass, linkClass, processClass,
-			symLinkClass, collectionClass,
+			symLinkClass, collectionClass, lookupKeyClass,
 			sequenceableCollectionClass, linkedListClass,
-			semaphoreClass,
+			semaphoreClass, characterArrayClass,
 			arrayedCollectionClass, arrayClass, stringClass,
 			symbolClass, byteArrayClass, compiledMethodClass,
+			compiledCodeClass, compiledBlockClass,
 			intervalClass, orderedCollectionClass,
 			sortedCollectionClass, bagClass, mappedCollectionClass,
-			setClass, dictionaryClass, 
+			setClass, dictionaryClass, namespaceClass,
 			systemDictionaryClass, identitySetClass,
 			identityDictionaryClass, undefinedObjectClass,
-			booleanClass, falseClass, trueClass, 
+			booleanClass, falseClass, trueClass, lookupTableClass,
 			processorSchedulerClass, delayClass, sharedQueueClass,
-			behaviorClass,
+			behaviorClass, rootNamespaceClass, methodDictionaryClass,
 			classDescriptionClass, classClass, metaclassClass,
 			smalltalkDictionary, messageClass, methodContextClass,
 			blockContextClass, blockClosureClass,
@@ -102,12 +105,16 @@ OOP			mst_objectClass, magnitudeClass, charClass, timeClass,
 
 			processorOOP = nil;
 
-static OOP		identityDictionaryNew(), systemDictionaryNew(),
-			newClass(), newMetaclass();
+
+static Dictionary	systemDictionaryNewInstance();
+
+static OOP		identityDictionaryNew(), newClass(), newMetaclass();
+
 static void		initSmalltalkDictionary(), addSmalltalk(),
 			printOOPClassName(), printClassName(),
 			createClassesPass1(), createClassesPass2(),
-			addSubClass(), addSTDIOObject();
+			addSubClass(), addSTDIOObject(),
+			initRuntimeObjects();
 
 /* static int		identityDictionaryFindKeyOrNil(),
 			findKeyOrNil(), findKey(), instanceSize(); */
@@ -134,7 +141,7 @@ static char *featureStrings[] = {
 static ClassInfo classInfo[] = {
   { &mst_objectClass,		nil,
       true,	false,	false,	0,
-      "Object",		nil,	nil,    "Smalltalk CFunctionDescs",
+      "Object",		nil,	nil,    "CFunctionDescs",
       nil /*"I am the root of the Smalltalk class system. \n\
 All classes in the system are subclasses of me."*/ },
 
@@ -149,7 +156,7 @@ All classes in the system are subclasses of me."*/ },
       nil },
 
   { &charClass,			&magnitudeClass,
-      false,	true,	true,	0, /* really has 1 indexed var */
+      false,	true,	true,	0, /* has 1 indexed var */
       "Character",	nil,	"Table",	nil,
       nil /*"My instances represent the 256 characters of the character set.  I provide\n\
 messages to translate between integers and character objects, and provide \n\
@@ -158,7 +165,7 @@ names for some of the common unprintable characters."*/ },
   { &timeClass,			&magnitudeClass,
       true,	false,	false,	1,
       "Time",		"seconds",	
-      "SecondClockAdjustment SecondClockOnStartup",	nil,
+      "SecondClockAdjustment ClockOnStartup",	nil,
       nil },
 
   { &dateClass,			&magnitudeClass,
@@ -183,10 +190,15 @@ names for some of the common unprintable characters."*/ },
       nil },
 
   { &integerClass,	       	&numberClass,
-      false,	true,	false,	0,
+      false,	false,	false,	0,
       "Integer",	nil,	nil,	nil,
       nil },
 
+  { &smallIntegerClass,	       	&integerClass,
+      false,	false,	false,	0,
+      "SmallInteger",	nil,	nil,	nil,
+      nil },
+
   { &largeIntegerClass,	       	&integerClass,  /*these four classes added by*/
       false,	false,	true,	0,              /*pb Sep 10 18:06:49 1998 */
       "LargeInteger",	nil,
@@ -208,9 +220,14 @@ names for some of the common unprintable characters."*/ },
       "LargeNegativeInteger",	nil,    nil,	nil,
       nil },
 
-  { &associationClass,		&magnitudeClass,
-      true,	false,	false,	2,
-      "Association",	"key value",	nil,	nil,
+  { &lookupKeyClass,		&magnitudeClass,
+      true,	false,	false,	1,
+      "LookupKey",	"key",	nil,	nil,
+      nil },
+
+  { &associationClass,		&lookupKeyClass,
+      true,	false,	false,	1,
+      "Association",	"value",	nil,	nil,
       nil },
 
   { &linkClass,			&mst_objectClass,
@@ -258,7 +275,12 @@ names for some of the common unprintable characters."*/ },
       "Array",  	nil,	nil,	nil,
       nil },
 
-  { &stringClass,		&arrayedCollectionClass,
+  { &characterArrayClass,	&arrayedCollectionClass,
+      false,	true,	true,	0,
+      "CharacterArray",	nil,	nil,	nil,
+      nil },
+
+  { &stringClass,		&characterArrayClass,
       false,	false,	true,	0,
       "String",		nil,	nil,	nil,
       nil },
@@ -273,9 +295,22 @@ names for some of the common unprintable characters."*/ },
       "ByteArray",	nil,	nil,	nil,
       nil },
 
-  { &compiledMethodClass,	&arrayedCollectionClass,
-      false,	false,	true,	3,
-      "CompiledMethod",	"descriptor methodHeader literals",	nil,	nil, 
+  { &compiledCodeClass,		&arrayedCollectionClass,
+      false,	false,	true,	2,
+      "CompiledCode",	"literals header",
+      nil,	nil, 
+      nil },
+
+  { &compiledBlockClass,	&compiledCodeClass,
+      false,	false,	true,	1,
+      "CompiledBlock",	"method",
+      nil,	nil, 
+      nil },
+
+  { &compiledMethodClass,	&compiledCodeClass,
+      false,	false,	true,	1,
+      "CompiledMethod",	"descriptor ",
+      nil,	nil, 
       nil /*"I represent methods that have been compiled.  I can recompile \n\
 methods from their source code, I can invoke Emacs to edit the source code \n\
 for one of my instances, and I know how to access components of my \n\
@@ -294,8 +329,8 @@ members that my instance represents."*/ },
       nil },
 
   { &sortedCollectionClass,	&orderedCollectionClass,
-      true,	false,	true,	1,
-      "SortedCollection",	"sortBlock",	nil,	nil,
+      true,	false,	true,	3,
+      "SortedCollection",	"lastOrdered sorted sortBlock",	"DefaultSortBlock",	nil,
       nil /*"I am a collection of objects, stored and accessed according to some\n\
 sorting criteria.  I store things using a bubble sort.  My instances have a \n\
 comparison block associated with them; this block takes two arguments and\n\
@@ -334,16 +369,37 @@ unique objects (typcially instances of Symbol), and associates another \n\
 object with that index.  I use the equality operator = to determine \n\
 equality of indices."*/ },
 
-  { &identityDictionaryClass,		&dictionaryClass,
+  { &lookupTableClass,			&dictionaryClass,
       true,	false,	true,	1,
-      "IdentityDictionary",	"values",	nil,	nil,
+      "LookupTable",	"values",	nil,	nil,
+      nil /*"I am similar to dictionary, except that my representation is \n\
+different."*/ },
+
+  { &identityDictionaryClass,		&lookupTableClass,
+      true,	false,	true,	0,
+      "IdentityDictionary",	nil,	nil,	nil,
       nil /*"I am similar to dictionary, except that my representation is \n\
 different, and I use the object identity comparision message == to \n\
 determine equivalence of indices."*/ },
 
-  /* MUST have the same structure as dictionary; they're used interchangeably
-   * within the C portion of the system */
-  { &systemDictionaryClass,		&dictionaryClass,
+  { &methodDictionaryClass,		&lookupTableClass,
+      true,	false,	true,	0,
+      "MethodDictionary",	nil,	nil,	nil,
+      nil },
+
+  /* These three MUST have the same structure as dictionary; they're used
+   * interchangeably within the C portion of the system */
+  { &rootNamespaceClass,		&dictionaryClass,
+      true,	false,	true,	0, 
+      "RootNamespace",	nil,	nil,	nil,
+      nil },
+
+  { &namespaceClass,			&rootNamespaceClass,
+      true,	false,	true,	0, 
+      "Namespace",	nil,	"Current",	nil,
+      nil },
+
+  { &systemDictionaryClass,		&rootNamespaceClass,
       true,	false,	true,	0, 
       "SystemDictionary",	nil,	nil,	nil,
       nil },
@@ -420,8 +476,8 @@ messages when errors occur, but currently I am not."*/ },
       nil },
 
   { &processorSchedulerClass,	&mst_objectClass,
-      true,	false,	false,	2,
-      "ProcessorScheduler",	"processLists activeProcess",	nil,	nil,
+      true,	false,	false,	3,
+      "ProcessorScheduler",	"processLists activeProcess idleTasks",	nil,	nil,
       nil },
 
   { &delayClass,	&mst_objectClass,
@@ -438,20 +494,19 @@ messages when errors occur, but currently I am not."*/ },
   /* Change this, classDescription, or Class, and you must change 
    * the implementaion of newMetaclass some */
   { &behaviorClass,		&mst_objectClass,
-      true,	false,	false,	4,
-      "Behavior",	"superClass subClasses methodDictionary instanceSpec",
+      true,	false,	false,	5,
+      "Behavior",	"superClass subClasses methodDictionary instanceSpec instanceVariables",
       nil,	nil,
       nil },
 
   { &classDescriptionClass,		&behaviorClass,
-      true,	false,	false,	4,
-      "ClassDescription",	"name comment instanceVariables category",
-      nil,	nil,
+      true,	false,	false,	0,
+      "ClassDescription",	nil,	nil,	nil,
       nil },
 
   { &classClass,		&classDescriptionClass,
-      true,	false,	false,	2,
-      "Class",	"classVariables sharedPools",	nil,	nil,
+      true,	false,	false,	6,
+      "Class",	"name comment category environment classVariables sharedPools",	nil,	nil,
       nil },
 
   { &metaclassClass,		&classDescriptionClass,
@@ -460,24 +515,24 @@ messages when errors occur, but currently I am not."*/ },
       nil },
 
   { &contextPartClass,		&mst_objectClass,
-      true,	false,	true,	4,
-      "ContextPart",	"parent ip sp hasBlock",	"UnwindPoints",	nil,
+      true,	false,	true,	5,
+      "ContextPart",	"parent ip sp receiver method ",
+      "UnwindPoints",	nil,
       nil },
 
   { &methodContextClass,		&contextPartClass,
-      true,	false,	true,	4,
-      "MethodContext",	"method methodClass selector receiver",	nil,	nil,
+      true,	false,	true,	1,
+      "MethodContext",	"flags ",	nil,	nil,
       nil },
 
   { &blockContextClass,		&contextPartClass,
-      true,	false,	true,	4,
-      "BlockContext",	"numArgs numTemps outerContext home",	nil,
-     nil,
+      true,	false,	true,	1,
+      "BlockContext",	"outerContext ",	nil,	nil,
       nil },
 
   { &blockClosureClass,		&mst_objectClass,
-      false,	false,	true,	1,
-      "BlockClosure",	"outerContext",	nil,	nil,
+      true,	false,	false,	3,
+      "BlockClosure",	"outerContext block receiver",	nil,	nil,
       nil },
 
 
@@ -526,8 +581,8 @@ external C data items." },
       nil },
 
   { &methodInfoClass,		&mst_objectClass,
-      true,	false,	false,	2,
-      "MethodInfo",	"sourceCode category",	nil,	nil,
+      true,	false,	false,	4,
+      "MethodInfo",	"sourceCode category class selector",	nil,	nil,
       nil },
 
   { &fileSegmentClass,		&mst_objectClass,
@@ -539,7 +594,7 @@ external C data items." },
 
 /* Smalltalk classes not defined:
    SmallInteger (integer has its function now)
-   (others like Point/Rectangle/RunArray are defined after the kernel starts
+   (others like Point/Rectangle/RunArray are defined after the kernel starts)
  */
 
 };
@@ -561,6 +616,11 @@ external C data items." },
 void
 initDictionary ()
 {
+  /* The order of this must match the indices defined in oop.h!! */
+  smalltalkDictionary	= allocOOP(nil);
+  processorOOP		= allocOOP(nil);
+  symbolTable		= allocOOP(nil);
+
   createClassesPass1();
 
   initBuiltinObjectsClasses();	/* we can do this now that classes are def'd */
@@ -645,6 +705,7 @@ createClassesPass2 ()
       }
     }
     class->objClass = newMetaclass(classOOP, index);
+    class->environment = smalltalkDictionary;
     class->instanceVariables =
       makeInstanceVariableArray(class->superClass, ci->instVarNames);
     class->classVariables = makeClassVariableDictionary(class->superClass,
@@ -685,15 +746,12 @@ int	numSubClasses;
      variable names, the right number of instance variables, etc.  We could
      take three passes, and use the instance variable spec for classes once
      it's established, but it's easier to create them here by hand */
-  metaclass->name = nilOOP;
-  metaclass->comment = nilOOP;
   metaclass->instanceVariables = 
       makeInstanceVariableArray(nilOOP, 
 "superClass subClasses methodDictionary instanceSpec \
-name comment instanceVariables category \
+instanceVariables name comment category environment \
 classVariables sharedPools");
 
-  metaclass->category = nilOOP;
   metaclass->subClasses = arrayNew(numSubClasses);
   if (numSubClasses > 0) {
     arrayAtPut(metaclass->subClasses, 1, fromInt(numSubClasses));
@@ -738,36 +796,42 @@ addSubClass (superClassOOP, subClassOOP)
 void
 initSmalltalkDictionary ()
 {
-  OOP		cFunctionDescsDictionary, featuresArray;
+  OOP		featuresArray;
   char	    	fullVersionString[200];
   int		i, numFeatures;
 
-  smalltalkDictionary 	   = systemDictionaryNew();
-  symbolTable 		   = arrayNew(INITIAL_SYMBOL_TABLE_SIZE);
-  cFunctionDescsDictionary = dictionaryNew();
+  setOOPObject(smalltalkDictionary, systemDictionaryNewInstance());
+  setOOPObject(symbolTable,	    instantiateWith(arrayClass, SYMBOL_TABLE_SIZE));
+  setOOPObject(processorOOP,	    instantiate(processorSchedulerClass));
+
+  currentNamespace	   = smalltalkDictionary;
   cObjectTypeCType 	   = cTypeNew(cObjectClass);
 
-  sprintf(fullVersionString, "GNU Smalltalk version %s", versionString);
+  for (numFeatures = 0; featureStrings[numFeatures]; numFeatures++) ;
 
-  initProcessSystem();
+  featuresArray = arrayNew(numFeatures);
+  for (i = 0; i < numFeatures; i++) {
+    arrayAtPut(featuresArray, i + 1, internString(featureStrings[i]));
+  }
+
+  sprintf(fullVersionString, "GNU Smalltalk version %s", VERSION);
 
   addSmalltalk("Smalltalk",		smalltalkDictionary);
-  addSmalltalk("CFunctionDescs",	cFunctionDescsDictionary);
+  addSmalltalk("CFunctionDescs",	dictionaryNew());
   addSmalltalk("Version", 		stringNew(fullVersionString));
   addSmalltalk("CObjectType",		cObjectTypeCType);
   addSmalltalk("KernelInitialized",	falseOOP);
   addSmalltalk("SymbolTable", 		symbolTable);
   addSmalltalk("Processor",		processorOOP);
+  addSmalltalk("Undeclared",		dictionaryNew());
+  addSmalltalk("InitBlocks", 		nilOOP);
 
-  for (numFeatures = 0; featureStrings[numFeatures]; numFeatures++) ;
-
-  featuresArray = arrayNew(numFeatures);
-
-  for (i = 0; i < numFeatures; i++) {
-    arrayAtPut(featuresArray, i + 1, internString(featureStrings[i]));
-  }
+  /* Smalltalk must be a proper namespace! */
+  addSmalltalk("Super",			nilOOP);
 
   addSmalltalk("Features", featuresArray);
+
+  initProcessSystem();
 }
 
 /* runs before GC turned on */
@@ -782,7 +846,13 @@ static void addSmalltalk (globalName, globalValue)
 void
 initRuntimeObjects ()
 {
-  addSmalltalk("KernelFilePath", stringNew(kernelFileDefaultPath));
+  SymbolInfo *si;
+  for (si = symbolInfo; si->symbolVar; si++) {
+    *si->symbolVar = internString(si->value);
+  }
+
+  addSmalltalk("KernelFilePath", stringNew(KERNEL_PATH));
+  addSmalltalk("KernelFileLocalPath", stringNew(kernelFileDefaultPath));
   addSmalltalk("ImageFilePath", stringNew(imageFileDefaultPath));
   addSmalltalk("ImageFileName", stringNew(binaryImageName));
 
@@ -857,6 +927,49 @@ newClass (superClassOOP, isPointers, isWords, isIndexable, numFixedFields)
 }
 
 
+/* Get everything needed from the Smalltalk dictionary; runs before
+   GC is turned on. */
+mst_Boolean
+initDictionaryOnImageLoad()
+{
+  ClassInfo	*info;
+  OOP		oop;
+
+  smalltalkDictionary	= oopAt(smalltalkOOPIndex);
+  processorOOP		= oopAt(processorOOPIndex);
+  symbolTable		= oopAt(symbolTableOOPIndex);
+
+  if (isNil(processorOOP) || isNil(symbolTable) || isNil(smalltalkDictionary)) {
+    return(false);
+  }
+
+  for (info = classInfo; info->classVar; info++) {
+    *info->classVar = dictionaryAt(smalltalkDictionary,
+      internString(info->name));
+    if (isNil(*info->classVar)) {
+      return(false);
+    }
+  }
+
+  currentNamespace = dictionaryAt(
+    classVariableDictionary(namespaceClass),
+    internString("Current"));
+
+  cObjectTypeCType = dictionaryAt(smalltalkDictionary,
+    internString("CObjectType"));
+
+  initBuiltinObjectsClasses();
+  initRuntimeObjects();
+    
+  for (oop = allOopsTable; oop < &oopTable[oopTableSize]; oop++) {
+    if (oopValid(oop) && oopClass(oop) == cFuncDescriptorClass) {
+      restoreCFuncDescriptor(oop); /* in cint.c */
+    }
+  }
+
+  return(true);
+}
+
 void
 setComment (classDescOOP, commentOOP)
      OOP	classDescOOP, commentOOP;
@@ -878,7 +991,7 @@ printOOPConstructor (oop)
   if (isAMetaclass(oop)) {
     classOOP = findAnInstance(oop);
     if (isNil(classOOP)) {
-      printf("<name unknown>");		/* we're a nameless class */
+      printf("<no name>");		/* we're a nameless class */
     } else {
       printClassName(classOOP);
     }
@@ -989,7 +1102,10 @@ validClassMethodDictionary (classOOP)
   class = (Class)oopToObj(classOOP);
   if (isNil(class->methodDictionary)) {
     OOP identDict;
+    mst_Object obj;
     identDict = identityDictionaryNew();
+    obj = oopToObj(identDict);
+    obj->objClass = methodDictionaryClass;
     class = (Class)oopToObj(classOOP);
     class->methodDictionary = identDict;
   }
@@ -997,6 +1113,27 @@ validClassMethodDictionary (classOOP)
   return (class->methodDictionary);
 }
 
+OOP
+findClassMethod (classOOP, selector)
+     OOP	classOOP, selector;
+{
+  Class			class;
+  IdentityDictionary	methodDictionary;
+  register OOP		methodDictionaryOOP;
+  register int		index;
+
+  class = (Class)oopToObj(classOOP);
+  methodDictionaryOOP = class->methodDictionary;
+  if (isNil(methodDictionaryOOP)) {
+    return (nilOOP);
+  }
+
+  index = identityDictionaryFindKeyOrNil(methodDictionaryOOP, selector);
+  methodDictionary = (IdentityDictionary)oopToObj(methodDictionaryOOP);
+
+  return (arrayAt(methodDictionary->values, index+1));
+}
+
 OOP
 classVariableDictionary (classOOP)
      OOP	classOOP;
@@ -1032,28 +1169,27 @@ sharedPoolDictionary (classOOP)
 
 
 OOP
-findSharedPoolVariable (classOOP, symbol)
-     OOP	classOOP, symbol;
+findSharedPoolVariable (poolOOP, symbol)
+     OOP	poolOOP, symbol;
 {
-  OOP		assocOOP, poolDictionaryOOP;
-  Class		class;
-  int		numPools, i;
+  OOP		assocOOP;
 
-  /* ??? check for non-class objects */
-  class = (Class)oopToObj(classOOP);
-
-  /* ??? shared pools are currently represented as arrays, from the book
-     I conjecture that their shared pools are implemented as sets. */
-  numPools = numOOPs(oopToObj(class->sharedPools));
-  for (i = 0; i < numPools; i++) {
-    poolDictionaryOOP = arrayAt(class->sharedPools, i+1);
-    assocOOP = dictionaryAssociationAt(poolDictionaryOOP, symbol);
-    if (!isNil(assocOOP)) {
-      return (assocOOP);
+  assocOOP = dictionaryAssociationAt(poolOOP, symbol);
+  while (isNil(assocOOP)) {
+    /* Try to find a super-namespace */
+    assocOOP = dictionaryAssociationAt(poolOOP, uppercaseSuperSymbol);
+    if (isNil(assocOOP)) {
+      return (nilOOP);
+    }
+    poolOOP = associationValue(assocOOP);
+    if (isNil(poolOOP)) {
+      return (nilOOP);
     }
-  }
 
-  return (nilOOP);
+    /* Bring on the search in the super-namespace */
+    assocOOP = dictionaryAssociationAt(poolOOP, symbol);
+  }
+  return (assocOOP);
 }
 
 
@@ -1135,7 +1271,7 @@ growIdentityDictionary (identityDictionaryOOP)
   incAddOOP(valuesOOP);
 
   identityDictionary =
-    (IdentityDictionary)instantiateWith(identityDictionaryClass, numFields);
+    (IdentityDictionary)instantiateWith(oopClass(identityDictionaryOOP), numFields);
 
   oldIdentityDictionary = (IdentityDictionary)oopToObj(identityDictionaryOOP);
   identityDictionary->tally = oldIdentityDictionary->tally;
@@ -1207,8 +1343,8 @@ identityDictionaryAtPut (identityDictionaryOOP, keyOOP, valueOOP)
   return (keyOOP);
 }
 
-OOP
-systemDictionaryNew()
+Dictionary
+systemDictionaryNewInstance()
 {
   Dictionary	dictionary;
 
@@ -1216,7 +1352,7 @@ systemDictionaryNew()
   dictionary = (Dictionary)instantiateWith(systemDictionaryClass,
 					   INITIAL_SMALLTALK_SIZE);
   dictionary->tally = fromInt(0);
-  return (allocOOP(dictionary));
+  return (dictionary);
 }
 
 OOP
@@ -1294,19 +1430,6 @@ printAssociationKey (associationOOP)
 }
 
 
-OOP nilVec[100];
-
-void
-dictInit ()
-{
-  int i;
-  for (i = 0; i < 100; i++) {
-    nilVec[i] = nilOOP;
-  }
-}
-
-
-
 OOP
 newString (len)
      int	len;
@@ -1519,3 +1642,17 @@ cTypeNew (cObjectSubclassOOP)
   cType->cObjectType = cObjectSubclassOOP;
   return (allocOOP(cType));
 }
+
+void
+setFileStreamFile(fileStreamOOP, fileOOP, fileNameOOP, isPipe)
+OOP	fileStreamOOP, fileOOP, fileNameOOP;
+mst_Boolean	isPipe;
+{
+  FileStream	fileStream;
+
+  fileStream = (FileStream)oopToObj(fileStreamOOP);
+  fileStream->file = fileOOP;
+  fileStream->name = fileNameOOP;
+  fileStream->isPipe = isPipe ? trueOOP : falseOOP;
+}
+
