@@ -1,16 +1,16 @@
-/***********************************************************************
+/******************************** -*- C -*- ****************************
  *
  *	Lexer Module.
  *
- *	$Revision: 1.6.2$
- *	$Date: 1999/08/31 11:23:18$
+ *	$Revision: 1.8.5$
+ *	$Date: 2000/12/27 10:45:49$
  *	$Author: pb$
  *
  ***********************************************************************/
 
 /***********************************************************************
  *
- * Copyright 1990, 91, 92, 94, 95, 99 Free Software Foundation, Inc.
+ * Copyright 1988-92, 1994-95, 1999, 2000 Free Software Foundation, Inc.
  * Written by Steve Byrne.
  *
  * This file is part of GNU Smalltalk.
@@ -34,29 +34,30 @@
 
 #include "gst.h"
 #include "alloc.h"
+#include "obstack.h"
 #include "lex.h"
 #include "lib.h"
-#if defined(MSDOS) | defined (__OS2__)
-#include "gst_tab.h"
-#else
-#include "gst.tab.h"
-#endif
 #include "str.h"
 #include "dict.h"
 #include "comp.h"
 #include "tree.h"
 #include "sysdep.h"
+#if defined(MSDOS) || defined (__OS2__)
+#include "gst_tab.h"
+#else
+#include "gst.tab.h"
+#endif
+
 #include <stdio.h>
 #include <math.h>
+
 #ifdef HAVE_READLINE
 #include <readline/readline.h>
 #include <readline/history.h>
 #endif /* HAVE_READLINE */
-
 #if STDC_HEADERS
 #  include <string.h>
 #endif /* STDC_HEADERS */
-
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
@@ -67,12 +68,12 @@
 /* Define this if you want the lexer to print all the tokens that it scans,
  * before passing them to the parser.
  */
-#undef LEXDEBUG
+/* #define LEXDEBUG */
 
 /* Define this if you're debugging the lexer and you don't want the parser
  * to be ran -- only lexical scanning will take place.
  */
-#undef NO_PARSE
+/* #define NO_PARSE */
 
 
 
@@ -83,7 +84,7 @@
 
 #define EMACS_PROCESS_MARKER	'\001' /* ^A as marker -- random choice */
 
-#if defined(_WIN32) || defined(__OS2__) || defined(MSDOS)
+#if defined(WIN32) || defined(__OS2__) || defined(MSDOS)
 #define HAVE_DOS_STYLE_PATH_NAMES
 #endif
 
@@ -92,22 +93,22 @@ typedef struct StringStreamStruct {
   char		*str;		/* pointer into asciz string */
 } StringStream;
 
-struct StreamStruct {
-  StreamType	type;
-  int		pushedBackChars[2]; /* holds the 2 characters of buffering */
-  int		pushedBackCount; /* number of chars pushed back */
-  int		line;
-  int		column;
-  char		*fileName;
-  int		fileOffset;
-  OOP		fileNameOOP;	/* holds full path name for file as an oop */
-  mst_Boolean	prompt;
+typedef struct StreamStruct {
+  StreamType	      type;
+  int		      pushedBackChars[2]; /* the 2 buffered characters */
+  int		      pushedBackCount;  /* number of chars pushed back */
+  int		      line;
+  int		      column;
+  char		      *fileName;
+  int		      fileOffset;
+  OOP		      fileNameOOP;	/* the full path name for file */
+  mst_Boolean	      prompt;
   union {
     FILE	*u_st_file;
     StringStream u_st_str;
   } st;
-  Stream	prevStream;
-};
+  struct StreamStruct *prevStream;
+} *Stream;
 
 #define st_file		st.u_st_file
 #define st_str		st.u_st_str
@@ -117,31 +118,33 @@ struct StreamStruct {
 
 mst_Boolean			compileCode = false;
 
-Stream	inStream = nil;
-char	*processingFile = nil;
+struct obstack			*compilationObstack;
 
-mst_Boolean		reportErrors = true;
-char		*firstErrorStr = nil;
-char		*firstErrorFile = nil;
-long		firstErrorLine;
+Stream				inStream = nil;
+char				*processingFile = nil;
+
+mst_Boolean			reportErrors = true;
+char				*firstErrorStr = nil;
+char				*firstErrorFile = nil;
+long				firstErrorLine;
 
 /* Controls the use of the changes file, for recording source text.
    If nil, no recording */
-char		*changeFileName = nil;
+char				*changeFileName = nil;
 
-static FILE	*changeStr = nil;
+static FILE			*changeStr = nil;
 
 static int			methodStartPos = 0;
 
 static mst_Boolean		isDigit(), isBaseDigit(), parsePrimitive(),
-				parseDigits(), parseFraction();
+				parseFraction();
+static double			parseDigits();
+static ByteObject		parseLargeInteger();
 static int			invalid(), comment(), charLiteral(),
 				parseBinOP(), stringLiteral(), parseNumber(),
 				parseIdent(), myGetC(), parseColon(),
   				nextChar(), digitToInt();
-static char			*scanStringoid();
-static void 			unreadChar(), lineStamp(), /* myUngetC(), */
-				myClose();
+static void 			lineStamp(), /* myUngetC(), */ myClose();
 static Stream			pushNewStream();
 
 #ifdef LEXDEBUG
@@ -156,7 +159,10 @@ typedef struct {
   int		charClass;
 } LexTabElt;
 
-static LexTabElt charTab[128] = {
+/* This macro is needed to properly handle 8-bit characters */
+#define charTab(x)		((x) < 128 ? charTable + (x) : charTable)
+
+static LexTabElt charTable[128] = {
 /*   0 */  { invalid,		0,		0 },
 /*   1 */  { invalid,		0,		0 },
 /*   2 */  { invalid,		0,		0 },
@@ -280,40 +286,37 @@ static LexTabElt charTab[128] = {
 /*   x */  { parseIdent,	0,		ID_CHAR },
 /*   y */  { parseIdent,	0,		ID_CHAR },
 /*   z */  { parseIdent,	0,		ID_CHAR },
-/*   { */  { invalid,		0,		0 },
+/*   { */  { 0,			OPEN_BRACE,	0 },
 /*   | */  { parseBinOP,	0,		SPECIAL_CHAR },
-/*   } */  { invalid,		0,		0 },
+/*   } */  { 0,			CLOSE_BRACE,	0 },
 /*   ~ */  { parseBinOP,	0,		SPECIAL_CHAR },
 /*  ^? */  { invalid,		0,		0 }
 };
 
 
+#if defined(LEXDEBUG)
+static inline int	yylexInternal();
 
 int
 yylex(lvalp)
      YYSTYPE *lvalp;
 {
-#if defined(USE_MONCONTROL) || defined(LEXDEBUG)
   int result;
 
-#ifdef USE_MONCONTROL
-  moncontrol(1);
-  result = myyylex(lvalp);
-  moncontrol(0); 
-#else
-  result = myyylex(lvalp);
+  result = yylexInternal(lvalp);
   printToken(result, lvalp);
-#endif
-
   return (result);
 }
 
+#define yylex yylexInternal
+#endif /* LEXDEBUG */
+
 int
-myyylex(lvalp, llocp)
+yylex(lvalp)
      YYSTYPE *lvalp;
 {
-#endif /* USE_MONCONTROL */
   int		ic, result, tokenStartPos = 0;
+  LexTabElt	*ct;
 
   if (compileCode) {
     /* The internal token is a trick to make the grammar be "conditional".
@@ -331,21 +334,22 @@ myyylex(lvalp, llocp)
   }
 
   while ((ic = nextChar()) != EOF) {
-    if ((charTab[ic].charClass & WHITE_SPACE) == 0) {
+    ct = charTab(ic);
+    if ((ct->charClass & WHITE_SPACE) == 0) {
       if (methodStartPos < 0) {
 	tokenStartPos = getCurFilePos();
       }
 
-      if (charTab[ic].lexFunc) {
-	result = (*charTab[ic].lexFunc)(ic, lvalp);
+      if (ct->lexFunc) {
+	result = (*ct->lexFunc)(ic, lvalp);
 	if (result) {
 	  if (methodStartPos < 0) { /* only do this here to ignore comments */
 	    methodStartPos = tokenStartPos - 1;
 	  }
 	  return (result);
 	}
-      } else if (charTab[ic].retToken) {
-	return (charTab[ic].retToken);
+      } else if (ct->retToken) {
+	return (ct->retToken);
       } else {
 	errorf("Unknown character %d in input stream, ignoring", ic);
 	hadError = true;
@@ -413,8 +417,16 @@ comment(c, lvalp)
      char	c;
      YYSTYPE *lvalp;
 {
-  scanStringoid(c, "comment");
+  int		ic;
 
+  do {
+    ic = nextChar();
+    if (ic == EOF) {
+      errorf("Unterminated comment, attempting recovery");
+      hadError = true;
+      break;
+    }
+  } while (ic != c);
   return (0);
 }
 
@@ -491,7 +503,7 @@ parseBinOP(c, lvalp)
     }
   }
 
-  if (ic != EOF && (charTab[ic].charClass & SPECIAL_CHAR)) {
+  if (ic != EOF && (charTab(ic)->charClass & SPECIAL_CHAR)) {
     *bp++ = ic;			/* accumulate next char */
   } else {
     unreadChar(ic);
@@ -510,7 +522,7 @@ parseBinOP(c, lvalp)
 /**/      }
 /**/    }
 /**/  } else {
-/**/    if (ic != EOF && (charTab[ic].charClass & SPECIAL_CHAR)) {
+/**/    if (ic != EOF && (charTab(ic)->charClass & SPECIAL_CHAR)) {
 /**/      *bp++ = ic;
 /**/    } else {
 /**/      unreadChar(ic);
@@ -545,7 +557,28 @@ stringLiteral(c, lvalp)
      char	c;
      YYSTYPE *lvalp;
 {
-  lvalp->sval = strdup(scanStringoid(c, "string"));
+  int		ic;
+
+  resetBuffer();
+
+  for (;;) {
+    ic = nextChar();
+    if (ic == EOF) {
+      errorf("Unterminated string, attempting recovery");
+      hadError = true;
+      break;
+    }
+    if (ic == c) {
+      /* check for doubled delimiters */
+      ic = nextChar();
+      if (ic != c) {
+	unreadChar(ic);
+	break;
+      }
+    }
+    addStrBufChar(ic);
+  }
+  lvalp->sval = obstack_curStrBuf(compilationObstack);
   return (STRING_LITERAL);
 }
 
@@ -558,7 +591,9 @@ parsePrimitive(c, lvalp)
 
   parseIdent(c, lvalp);
   result = (strcmp(lvalp->sval, "primitive:") == 0);
-  xfree(lvalp->sval);
+#ifdef prev_obstacks
+/**/  xfree(lvalp->sval);
+#endif
   return (result);
 }
 
@@ -569,13 +604,13 @@ parseIdent(c, lvalp)
 {
   int		ic, identType;
   
-  initStrBuf();
+  resetBuffer();
   addStrBufChar(c);
   
   identType = IDENTIFIER;
   
   for (;;) {
-    while (((ic = nextChar()) != EOF) && (charTab[ic].charClass & ID_CHAR)) {
+    while (((ic = nextChar()) != EOF) && (charTab(ic)->charClass & ID_CHAR)) {
       addStrBufChar(ic);
     }
     
@@ -591,8 +626,8 @@ parseIdent(c, lvalp)
 	break;
       }
       addStrBufChar(':');
-      if (ic == EOF || (charTab[ic].charClass & ID_CHAR) == 0
-	  || (charTab[ic].charClass & DIGIT) != 0) {
+      if (ic == EOF || (charTab(ic)->charClass & ID_CHAR) == 0
+	  || (charTab(ic)->charClass & DIGIT) != 0) {
 	if (identType == IDENTIFIER) {
 	  /* first time through */
 	  identType = KEYWORD;
@@ -606,11 +641,12 @@ parseIdent(c, lvalp)
     }
   }
   
-  lvalp->sval = strdup(curStrBuf());
+  lvalp->sval = obstack_curStrBuf(compilationObstack);
   
   return (identType);
 }
 
+
 int
 parseNumber(c, lvalp)
      char	c;
@@ -618,7 +654,8 @@ parseNumber(c, lvalp)
 {
   int		base, exponent, ic;
   double	num, floatExponent;
-  mst_Boolean	mantissaParsed = false, isNegative = false, dotSeen = false;
+  mst_Boolean	mantissaParsed = false, isNegative = false, isFloat = false,
+  		largeInteger = false;
   double	scale;
   
   base = 10;
@@ -627,10 +664,14 @@ parseNumber(c, lvalp)
   ic = c;
 
   if (ic != '-') {
-    parseDigits(ic, 10, &num);
+    num = parseDigits(ic, isNegative, 10, &largeInteger);
     ic = nextChar();
     if (ic == 'r') {
       base = (int)num;
+      if (base > 36 || largeInteger) {
+	errorf("Numeric base too large %d", base);
+	hadError = true;
+      }
       ic = nextChar();
     } else {
       mantissaParsed = true;
@@ -650,7 +691,7 @@ parseNumber(c, lvalp)
       ic = nextChar();
     }
     
-    parseDigits(ic, base, &num);
+    num = parseDigits(ic, isNegative, base, &largeInteger);
     ic = nextChar();
   }
 
@@ -663,19 +704,19 @@ parseNumber(c, lvalp)
       unreadChar(ic);
       ic = '.';
     } else {
-      dotSeen = true;
+      isFloat = true;
       parseFraction(ic, base, &num, &scale);
       ic = nextChar();
     }
   }
 
-  if (ic == 'e') {
+  if (ic == 'e' || ic == 'd' || ic == 'q') {
     ic = nextChar();
     if (ic == '-') {
-      parseDigits(nextChar(), 10, &floatExponent);
+      floatExponent = parseDigits(nextChar(), true, 10, nil);
       exponent -= (int)floatExponent;
     } else {
-      parseDigits(ic, 10, &floatExponent);
+      floatExponent = parseDigits(ic, false, 10, nil);
       exponent += (int)floatExponent;
     }
   } else {
@@ -683,48 +724,55 @@ parseNumber(c, lvalp)
   }
 
   if (scale != 0.0) {
-    num *= scale/*pow((double)base, (double)exponent)*/;
+    num *= scale;
   }
 
   if (exponent != 0) {
     num *= pow((double)base, (double)exponent);
+    isFloat = true;
   }
 
   if (isNegative) {
     num = -num;
   }
     
-  if (dotSeen) {
+  if (isFloat) {
     lvalp->fval = num;
     return (FLOATING_LITERAL);
+  } else if (largeInteger || num < MIN_ST_INT || num > MAX_ST_INT) {
+    lvalp->boval = parseLargeInteger(isNegative, base);
+    return (LARGE_INTEGER_LITERAL);
+  } else if (num >= 0 & num <= 255) {
+    lvalp->ival = (long)num;
+    return (BYTE_LITERAL);
   } else {
     lvalp->ival = (long)num;
-
-    /* I don't use 1L because I am not certain all C compilers handle it */
-    if (num < -(((long)1) << ((SIZEOF_LONG * 8) - 2))
-	|| num >= ( ((long)1) << ((SIZEOF_LONG * 8) - 2))) {
-      /* at least warn the guy... */
-      errorf("Integer literal too large to be represented in Smalltalk", num);
-      hadError = true;
-    }
-      
     return (INTEGER_LITERAL);
   }
 }
 
-mst_Boolean
-parseDigits(c, base, numPtr)
-     char	c;
-     int	base;
-     double	*numPtr;
+double
+parseDigits(c, negative, base, largeInteger)
+     char	 c;
+     mst_Boolean negative;
+     int	 base;
+     mst_Boolean *largeInteger;
 {
   double	result;
   mst_Boolean	oneDigit = false;
 
+  resetBuffer();
+
   for (result = 0.0; isBaseDigit(c, base); c = nextChar()) {
     result *= base;
     oneDigit = true;
     result += digitToInt(c, base);
+    if (largeInteger) {
+      addStrBufChar(digitToInt(c, base));
+      if (result > -MIN_ST_INT || (!negative && result > MAX_ST_INT)) {
+        *largeInteger = true;
+      }
+    }
   }
 
   if (!oneDigit) {
@@ -734,9 +782,7 @@ parseDigits(c, base, numPtr)
 
   unreadChar(c);
 
-  *numPtr = result;
-
-  return (true);
+  return (result);
 }
 
 mst_Boolean
@@ -814,37 +860,68 @@ mst_Boolean
 isDigit(ic)
      int	ic;
 {
-  return ((charTab[ic].charClass & DIGIT) != 0);
+  return ((charTab(ic)->charClass & DIGIT) != 0);
 }
 
-char *
-scanStringoid(startChar, typeName)
-     char	startChar, *typeName;
+ByteObject
+parseLargeInteger(negative, base)
+     mst_Boolean	negative;
+     int		base;
 {
-  int		ic;
+  register int	i;
+  int		size, digitsLeft;
+  Byte		*digits, *result;
+  ByteObject	bo;
 
-  initStrBuf();
-
-  for (;;) {
-    ic = nextChar();
-    if (ic == EOF) {
-      errorf("Unterminated %s, attempting recovery", typeName);
-      hadError = true;
-      return (curStrBuf());
-    }
-    if (ic == startChar) {
-      /* check for doubled delimiters */
-      ic = nextChar();
-      if (ic != startChar) {
-	unreadChar(ic);
-	return (curStrBuf());
-      }
-    }
-    addStrBufChar(ic);
-  }
+  size = bufferSize();
+  digits = (char *) alloca(size);
+  copyBuffer(digits);
   
-}
+  bo = (ByteObject) obstack_alloc(
+    compilationObstack, sizeof(ByteObjectStruct) + size);
 
+  bo->class = negative ? largeNegativeIntegerClass : largePositiveIntegerClass;
+  result = bo->body;
+  memzero(result, size);
+
+  /* On each pass, multiply the previous partial result by the base,
+   * and sum each of the digits as they were retrieved by parseDigits. */
+  for (digitsLeft = size; digitsLeft--; ) {
+    register int total, carry;
+
+    total = result[0] * base + *digits++;
+    carry = total >> 8;
+    result[0] = (Byte) total;
+    for (i = 1; i < size; i++) {
+      total = result[i] * base + carry;
+      carry = total >> 8;
+      result[i] = (Byte) total;
+    }
+  }
+
+  if (negative) {
+    /* Do two's complement -- first invert, then increment with carry */
+    for (i = 0; i < size; i++) {
+      result[i] ^= 255;
+    }
+    for (i = 0; (++result[i]) == 0; i++);
+    
+    /* Search where the number really ends -- discard trailing 111... bytes */
+    for (; size > 0 && result[size-1] == 255; size--);
+  } else {
+    /* Search where the number really ends -- discard trailing 000... bytes
+     * but remember, the most significant bit of the last digit must be 0! */
+    for (; size > 0 && result[size-1] == 0; size--);
+    if (result[size-1] > 127) {
+      size++;
+    }
+  }
+
+  /* Only now can we set the size! */
+  bo->size = size;
+  return (bo);
+}
+
 /*
  * Report an error to the user.  ### Will show position in the text of
  * the error at some point.
@@ -875,8 +952,6 @@ errorf(str, ERROR_ARGS)
 }
 #undef ERROR_ARGS
 
-
-
 /* ??? is void the right type */
 void
 yyerror(s)
@@ -1116,13 +1191,27 @@ int	ic;
 void
 parseStream()
 {
+  struct obstack thisObstack, *oldObstack;
+
+  /* Allow re-entrancy by allocating a different obstack every time
+   * parseStream is called */
+  oldObstack = compilationObstack;
+  compilationObstack = &thisObstack;
+  obstack_init (&thisObstack);
+
+  {
 #ifdef NO_PARSE
-  YYSTYPE yylval;
-  while(yylex(&yylval)) ;
-#else /* debugging the lexer */
-  extern int yyparse();
-  yyparse();
-#endif /* NO_PARSE */
+    YYSTYPE yylval;
+    while(yylex(&yylval)) ;
+#else /* !NO_PARSE */
+    extern int yyparse();
+    hadError = false;
+    yyparse();
+#endif /* !NO_PARSE */
+  }
+
+  obstack_free (&thisObstack, NULL);
+  compilationObstack = oldObstack;
 }
 
 
@@ -1389,13 +1478,13 @@ isKernelFile()
   }
 
   fullFileName = fullFileNameTest = getFullFileName(inStream->fileName);
-#if defined(MSDOS) || defined(_WIN32) || defined(__OS2__)
+#if defined(MSDOS) || defined(WIN32) || defined(__OS2__)
   if (fullFileName[1] == ':') {
     fullFileName += 2;		/* skip over "<drive>:"  */
   }
 #endif
   result = (strstr(fullFileNameTest, KERNEL_PATH)  == fullFileNameTest);
-  xfree(fullFileName);
+  xfree(fullFileNameTest);
 
   return (result);
 }
@@ -1410,7 +1499,7 @@ getMethodSourceFromCurFile()
   startPos = getMethodStartPos();
   endPos = getCurFilePos(); 
   /* The second -1 removes the terminating '!' */
-  length = endPos  - startPos - 1 - 1;
+  length = endPos  - startPos - 1;
   buf = (char *)xmalloc(length + 1);
   fseek(inStream->st_file, startPos, 0);
   fread(buf, length, 1, inStream->st_file);
@@ -1475,15 +1564,29 @@ printToken(token, yylval)
       break;
     case IDENTIFIER:
       printf("IDENTIFIER: %s\n", yylval->sval);
+#ifdef prev_obstacks
+/**/      free(yylval->sval);
+#endif
       break;
     case KEYWORD:
       printf("KEYWORD: %s\n", yylval->sval);
+#ifdef prev_obstacks
+/**/      free(yylval->sval);
+#endif
       break;
     case SYMBOL_KEYWORD:
       printf("SYMBOL_KEYWORD: %s\n", yylval->sval);
+#ifdef prev_obstacks
+/**/      free(yylval->sval);
+#endif
       break;
+    case LARGE_INTEGER_LITERAL;
+      printf("LARGE_INTEGER_LITERAL\n");
+#ifdef prev_obstacks
+/**/      free(yylval->boval);
+#endif
     case INTEGER_LITERAL:
-      printf("INTEGER_LITERAL: %d\n", yylval->ival);
+      printf("INTEGER_LITERAL: %ld\n", yylval->ival);
       break;
     case FLOATING_LITERAL:
       printf("FLOATING_LITERAL: %g\n", yylval->fval);
@@ -1493,10 +1596,44 @@ printToken(token, yylval)
       break;
     case STRING_LITERAL:
       printf("STRING_LITERAL: %s\n", yylval->sval);
+#ifdef prev_obstacks
+/**/      free(yylval->sval);
+#endif
       break;
     case BINOP:
       printf("BINOP: %s\n", yylval->sval);
+#ifdef prev_obstacks
+/**/      free(yylval->sval);
+#endif
       break;
   }
 }
 #endif
+
+void
+yyprint(file, token, yylval)
+     FILE *file;
+     int token;
+     YYSTYPE *yylval;
+{
+  switch(token) {
+    case IDENTIFIER:
+    case KEYWORD:
+    case SYMBOL_KEYWORD:
+    case STRING_LITERAL:
+    case BINOP:
+      fprintf(file, ": `%s'", yylval->sval);
+      break;
+    case INTEGER_LITERAL:
+      fprintf(file, ": %ld", yylval->ival);
+      break;
+    case FLOATING_LITERAL:
+      fprintf(file, ": %g", yylval->fval);
+      break;
+    case CHAR_LITERAL:
+      fprintf(file, ": %c", yylval->cval);
+      break;
+    default:
+      break;
+  }
+}
