@@ -1,16 +1,16 @@
-/***********************************************************************
+/******************************** -*- C -*- ****************************
  *
  *	Main (library) Module
  *
- *	$Revision: 1.6.2$
- *	$Date: 1999/08/31 11:23:18$
+ *	$Revision: 1.8.5$
+ *	$Date: 2000/12/27 10:45:49$
  *	$Author: pb$
  *
  ***********************************************************************/
 
 /***********************************************************************
  *
- * Copyright 1990, 91, 92, 94, 95, 99 Free Software Foundation, Inc.
+ * Copyright 1988-92, 1994-95, 1999, 2000 Free Software Foundation, Inc.
  * Written by Steve Byrne.
  *
  * This file is part of GNU Smalltalk.
@@ -43,7 +43,6 @@
 #include "dld_gst.h" 
 #include "lex.h"
 #include "dict.h"
-#include "str.h"
 #include "sysdep.h" 
 #include <stdio.h>
 #include <sys/types.h>
@@ -103,12 +102,15 @@ static char		helpText[] =
    "\n   -c --core-dump\t\t Dump core on fatal signal"
    "\n   -d --declaration-trace-user\t Trace compilation of files on the command line"
    "\n   -D --declaration-trace-all\t Trace compilation of all loaded files"
+#ifndef USE_DYNAMIC_TRANSLATION
    "\n   -e --execution-trace-user\t Trace execution of files on the command line"
    "\n   -E --execution-trace-all\t Trace execution of all loaded files"
+#endif
    "\n   -g --no-gc-message\t\t Do not print garbage collection messages"
    "\n   -H -h -? --help\t\t Print this message and exit"
    "\n   -i --rebuild-image\t\t Ignore the image file; rebuild it from scratch"
    "\n   -I --image-file file\t\t Instead of `gst.im', use `file' as the image\n\t\t\t\t file, and ignore the kernel files' timestamps\n"
+   "\n   -K --kernel-file file\t Make file's path relative to the image path."
    "\n   -l --log-changes\t\t Log changes in `change-log.st'"
    "\n   -L --log-file file\t\t Log changes in file `file'"
    "\n   -p --emacs-mode\t\t Execute as a `process' (from within Emacs)"
@@ -120,44 +122,46 @@ static char		helpText[] =
    "\n   -v --version\t\t\t Print the Smalltalk version number and exit"
    "\n   -V --verbose\t\t\t Print names of loaded files"
    "\n   -y --yacc-debug\t\t Turn on debugging in the parser"
-   "\n   -  --\t\t\t Read input from standard input explicitly"
+   "\n   -\t\t\t\t Read input from standard input explicitly"
    "\n"
    "\nFiles are loaded one after the other.  After the last one is loaded,"
    "\nSmalltalk will exit.  If no files are specified, Smalltalk reads from"
    "\nthe terminal, with prompts."
    "\n"
-   "\nReport bugs to bug-gnu-smalltalk@gnu.ai.mit.edu,"
-   "\nwith a Cc: to bonzini@gnu.org";
+   "\nReport bugs to help-smalltalk@gnu.org";
 
 static char		copyrightAndLegalStuffText[] =
-   "Copyright 1990, 91, 92, 94, 95, 99 Free Software Foundation, Inc."
+   "GNU Smalltalk version %s"
+   "\nwritten by Steve Byrne (sbb@gnu.org) and Paolo Bonzini (bonzini@gnu.org)"
+   "\nCopyright 1988-92, 1994-95, 1999, 2000 Free Software Foundation, Inc."
    "\n"
    "\nGNU Smalltalk comes with NO WARRANTY, to the extent permitted by law."
    "\nYou may redistribute copies of GNU Smalltalk under the terms of the"
-   "\nGNU General Public License.  For more information, see the file named COPYING."
+   "\nGNU General Public License.  For more information, see the file named"
+   "\nCOPYING."
    "\n"
+   "\nUsing default kernel path: %s"
+   "\nUsing default image path: %s"
    "\n";
 
 static char		*longOptions[] = {
    "-a\0--smalltalk",			"-c\0--core-dump",
    "-d\0--declaration-trace-user",	"-D\0--declaration-trace-all",
+#ifndef USE_DYNAMIC_TRANSLATION
    "-e\0--execution-trace-user",	"-E\0--execution-trace-all",
+#endif
    "-g\0--no-gc-message",		"-H\0--help",
    "-i\0--rebuild-image",		"-I\0--image-file",
-   "-l\0--log-changes",			"-L\0--log-file",
-   "-p\0--emacs-mode",			"-q\0--quiet",
-   "-Q\0--no-messages",			"-r\0--regression-test",
-   "-q\0--silent",			"-s\0--store-no-source",
-   "-S\0--snapshot",			"-v\0--version",
-   "-V\0--verbose",			"-y\0--yacc-debug",
-   nil};
+   "-K\0--kernel-file",			"-l\0--log-changes",
+   "-L\0--log-file",			"-p\0--emacs-mode",
+   "-q\0--quiet",			"-Q\0--no-messages",
+   "-q\0--silent",			"-r\0--regression-test",
+   "-s\0--store-no-source",		"-S\0--snapshot",	
+   "-v\0--version",			"-V\0--verbose",
+   "-y\0--yacc-debug",			nil};
 
 
 
-#ifdef unused
-/**/extern int		adds, reused, reprobes, hitsOn[];
-#endif /* unused */
-
 /* When true, this flag suppresses the printing of execution-related
  * messages, such as the number of byte codes executed by the
  * last expression, etc.
@@ -167,11 +171,6 @@ mst_Boolean		quietExecution;
 
 char			*kernelFileDefaultPath, *imageFileDefaultPath;
 
-/* This string contains the printed representation of the Smalltalk version
- * number
- */
-char			versionString[50];
-
 /* This is the name of the binary image to load.  If it is not nil after the
  * command line is parsed, the checking of the dates of the kernel source files
  * against the image file date is overridden.  If it is nil, it is set to
@@ -202,7 +201,7 @@ mst_Boolean			storeNoSource = false;
 static char *		parseLongOption();
 static mst_Boolean 	processFile(), okToLoadBinary(), findKernelFile();
 static void		loadStandardFiles(), loadUserInitFile(), initPaths(),
-			parseArgs(), loadUserPreImageFile(), makeVersionString();
+			parseArgs(), loadUserPreImageFile();
 
 /* Set by command line flag.  When true, Smalltalk saves a snapshot after
  * loading the files on the command line, before exiting.
@@ -210,7 +209,7 @@ static void		loadStandardFiles(), loadUserInitFile(), initPaths(),
 
 static mst_Boolean		snapshotAfterLoad = false;
 
-/* Whether SMALLTALK_IMAGE is set.  Only if it is set, .stpre is loaded. */
+/* Whether SMALLTALK_IMAGE is set (only if it is set, .stpre is loaded) */
 static mst_Boolean		isLocalImage;
 
 /* The imageFileDefaultPath followed by /gst.im */
@@ -244,7 +243,8 @@ static mst_Boolean		traceKernelExecution;
  */
 static mst_Boolean		ignoreImage;
 
-/* If true, make printed messages constant.
+/* This is TRUE if we are doing regression testing, and causes whatever sources
+ * of variance to be suppressed (such as printing out execution statistics).
  */
 mst_Boolean			regressionTesting;
 
@@ -276,20 +276,25 @@ static char		*standardFiles[] = {
   "Float.st",
   "Fraction.st",
   "LargeInt.st",
+  "SmallInt.st",
   MAP_FILE("Charactr.st", "Character.st"),
+  MAP_FILE("LookupK.st", "LookupKey.st"),
   MAP_FILE("Assoc.st", "Association.st"),
   "Link.st",
   "Process.st",
+  MAP_FILE("CompCode.st", "CompildCode.st"),
+  MAP_FILE("CompMeth.st", "CompildMeth.st"),
+  MAP_FILE("CompBlk.st", "CompiledBlk.st"),
   MAP_FILE("Collectn.st", "Collection.st"),
   MAP_FILE("SeqColl.st", "SeqCollect.st"),
   MAP_FILE("LnkList.st", "LinkedList.st"),
   MAP_FILE("Semaphor.st", "Semaphore.st"),
   MAP_FILE("ArrdColl.st", "ArrayColl.st"),
   "Array.st",
+  MAP_FILE("ByteArr.st", "ByteArray.st"),
+  MAP_FILE("CharArr.st", "CharArray.st"),
   "String.st",
   "Symbol.st",
-  MAP_FILE("ByteArr.st", "ByteArray.st"),
-  MAP_FILE("CompMeth.st", "CompildMeth.st"),
   "Interval.st",
   MAP_FILE("OrdColl.st", "OrderColl.st"),
   MAP_FILE("SortColl.st", "SortCollect.st"),
@@ -298,7 +303,11 @@ static char		*standardFiles[] = {
   "Set.st",
   MAP_FILE("IdentSet.st", "IdentitySet.st"),
   MAP_FILE("Dict.st", "Dictionary.st"),
+  MAP_FILE("LookupT.st", "LookupTable.st"),
   MAP_FILE("IdDict.st", "IdentDict.st"),
+  MAP_FILE("MthDict.st", "MethodDict.st"),
+  MAP_FILE("RootNS.st", "RootNamespc.st"),
+  MAP_FILE("Namespc.st", "Namespace.st"),
   MAP_FILE("SysDict.st", "SysDict.st"),
   "Stream.st",
   MAP_FILE("PosStrm.st", "PosStream.st"),
@@ -328,22 +337,30 @@ static char		*standardFiles[] = {
   MAP_FILE("FileSeg.st", "FileSegment.st"),
   "SymLink.st",
   MAP_FILE("init.st", "initialize.st"),
+
+  /* C call-out facilities */
   "CObject.st",
   "CType.st",
   "CFuncs.st",
   "CStruct.st",
-  "Autoload.st",
-  MAP_FILE("Trnscrpt.st", "Transcript.st"),	/* yuck! no vocals! */
-  "File.st",
-  MAP_FILE("Dir.st", "Directory.st"),
-  "Browser.st",
-  "RunArray.st",
-  MAP_FILE("DirMsg.st", "DirMessage.st"),
-  MAP_FILE("ExcHandl.st", "ExcHandling.st"),
-  MAP_FILE("Adaptors.st", "ValueAdapt.st"),
-  MAP_FILE("ObjDump.st", "ObjDumper.st"),
+
+  /* More core classes */
+  MAP_FILE("WeakObjs.st", "WeakObjects.st"),
   "Point.st",
   MAP_FILE("Rect.st", "Rectangle.st"),
+  "RunArray.st",
+  MAP_FILE("DirMsg.st", "DirMessage.st"),
+  "File.st",
+  MAP_FILE("Dir.st", "Directory.st"),
+  MAP_FILE("ExcHandl.st", "ExcHandling.st"),
+  MAP_FILE("AnsiExc.st", "AnsiExcept.st"),
+  MAP_FILE("Trnscrpt.st", "Transcript.st"),	/* yuck! no vocals! */
+
+  /* Goodies */
+  "Autoload.st",
+  "Browser.st",
+  MAP_FILE("Adaptors.st", "ValueAdapt.st"),
+  MAP_FILE("ObjDump.st", "ObjDumper.st"),
   MAP_FILE("PkgLoad.st", "PkgLoader.st"),
   "DLD.st",
   nil
@@ -394,7 +411,6 @@ initSmalltalk()
   smalltalkInitialized = true;
 
   initPaths();
-  makeVersionString();
 #ifdef HAVE_READLINE
   initializeReadline();
 #endif /* HAVE_READLINE */
@@ -406,6 +422,12 @@ initSmalltalk()
   initCFuncs();
   initOOPRegistry();
 
+  if (regressionTesting) {
+    traceKernelDeclarations = declareTracing = false;
+    traceKernelExecution = executionTracing = false;
+    verbose = false;
+  }
+
   if (binaryImageName) {
     loadBinary = true;
   } else {
@@ -415,14 +437,11 @@ initSmalltalk()
 
   if (loadBinary && loadFromFile(binaryImageName)) {
     initChangesStream();
-    initDefaultCompilationEnvironment();
-    initRuntimeObjects();
     initInterpreter();
     gcOn();
   } else {
-    initOOPTable();
+    initOOPTable(INITIAL_OOP_TABLE_SIZE);
     initDictionary();
-    initSymbols();
     initInterpreter();
 
     resetChangesFile();
@@ -458,23 +477,6 @@ initSmalltalk()
 
   invokeInitBlocks();
   loadUserInitFile();
-
-#ifdef unused
-/**/#ifdef symbol_table_profiling
-/**/  {
-/**/    int i;
-/**/    printf("%d adds, %d reused %d reprobes\n", adds, reused, reprobes);
-/**/    for (i = 0; i < OOP_TABLE_SIZE; i++) {
-/**/      if (hitsOn[i]) {
-/**/	    printf("hitsOn[%d] = %4d\t%d\t%4x", i, hitsOn[i], i, i);
-/**/	    printObject(oopAt(i));
-/**/	    printf("\n");
-/**/      }
-/**/    }
-/**/  }
-/**/#endif /* symbol_table_profiling */
-#endif /* unused */
-
 }
 
 void
@@ -488,11 +490,13 @@ topLevelLoop()
   quietExecution = runQuietly || emacsProcess;
 
   for (filesProcessed = 0; *++smalltalkArgv; ) {
-    if (smalltalkArgv[0][0] != '-') {
-      processFile(smalltalkArgv[0], quietExecution);
-      filesProcessed++;
-    } else if (smalltalkArgv[0][1] == '-' || smalltalkArgv[0][1] == '\0') {
-      /* either - by itself or -- indicates standard input */
+    if (smalltalkArgv[0][0] == '\0') {
+      /* Option parsing replaced option arguments with empty strings */
+      continue;
+    }
+    
+    if (!strcmp(smalltalkArgv[0], "-")) {
+      /* - by itself indicates standard input */
       nonInteractive = false;
 #ifdef HAVE_READLINE
       pushReadlineString();
@@ -502,6 +506,9 @@ topLevelLoop()
       parseStream();
       popStream(true);
       nonInteractive = true;
+    } else {
+      processFile(smalltalkArgv[0], quietExecution);
+      filesProcessed++;
     }
   }
 
@@ -513,14 +520,16 @@ topLevelLoop()
     pushUNIXFile(stdin, "stdin");
 #endif /* HAVE_READLINE */
     parseStream();
-  } else if (snapshotAfterLoad) {
-    saveToFile(binaryImageName);
   }
-
+  
 #ifdef USE_MONCONTROL
    moncontrol(0);
 #endif /* USE_MONCONTROL */
 
+  if (snapshotAfterLoad) {
+    saveToFile(binaryImageName);
+  }
+
 }
 
 
@@ -541,38 +550,36 @@ initPaths()
 {
   char *currentDirectory = getCurDirName();
 
-  if ((kernelFileDefaultPath = (char *)getenv("SMALLTALK_KERNEL")) == nil) {
-    kernelFileDefaultPath = KERNEL_PATH;
-  }
-  if (!fileIsReadable(kernelFileDefaultPath)) {
-    kernelFileDefaultPath = xmalloc(1024);
-    strcpy(kernelFileDefaultPath, currentDirectory);
-    strcat(kernelFileDefaultPath, "/kernel");
-    resizeString(&kernelFileDefaultPath);
-  }
-
+  kernelFileDefaultPath = (char *)getenv("SMALLTALK_KERNEL");
   imageFileDefaultPath = (char *)getenv("SMALLTALK_IMAGE");
   isLocalImage = true;
-  
-  if (imageFileDefaultPath == nil || !fileIsReadable(imageFileDefaultPath)) {
-    if (fileIsReadable(basicImageName)) {
-      imageFileDefaultPath = xmalloc(1024);
-      strcpy(imageFileDefaultPath, currentDirectory);
-      resizeString(&imageFileDefaultPath);
-    } else {
-      imageFileDefaultPath = IMAGE_PATH;
 
-      /* This is a kludge so that OSes such as Windows and MS-DOS which have
-       * no concept of home directories always load the .stpre file.
+  if (!kernelFileDefaultPath
+      || !fileIsReadable(kernelFileDefaultPath)) {
+    kernelFileDefaultPath = xmalloc(strlen(currentDirectory) + 8);
+    sprintf(kernelFileDefaultPath, "%s/kernel", currentDirectory);
+  }
+
+  if (!imageFileDefaultPath
+      || !fileIsReadable(imageFileDefaultPath)
+      || !fileIsWriteable(imageFileDefaultPath)) {
+    imageFileDefaultPath = IMAGE_PATH;
+
+    if (!fileIsReadable(imageFileDefaultPath)
+        || !fileIsWriteable(imageFileDefaultPath)) {
+      /* Resort to the current directory -- isLocalImage is already true. */
+      imageFileDefaultPath = strdup(currentDirectory);
+    } else {
+      /* Found in the standard image path.  Apply this kludge so that OSes
+       * such as Windows and MS-DOS which have no concept of home directories
+       * always load the .stpre file.
        */
       isLocalImage = ( ((char *)getenv("HOME")) == nil);
     }
   }
 
-  defaultImageName = xmalloc(1024);
-  strcpy(defaultImageName, imageFileDefaultPath);
-  strcat(defaultImageName, "/" basicImageName);
-  resizeString(&defaultImageName);
+  defaultImageName = xmalloc(strlen(imageFileDefaultPath) + 8);
+  sprintf(defaultImageName, "%s/gst.im", imageFileDefaultPath);
 
   free(currentDirectory);
 }
@@ -640,9 +647,9 @@ loadStandardFiles()
   for (fileNames = standardFiles; *fileNames; fileNames++) {
     findKernelFile(*fileNames, fullFileName);
     if (!processFile(fullFileName, true)) {
-      fprintf(stderr,
-	      "Can't find system file '%s', proceeding without it.\n",
-	      fullFileName);
+      errorf("can't find system file '%s'", fullFileName);
+      errorf("image bootstrap failed.");	      
+      exit(1);
     }
   }
 
@@ -678,8 +685,7 @@ loadStandardFiles()
  * Outputs
  *
  *	Returns true if the kernel file is found in the local directory
- *	or its `kernel' subdirector, and false if the file was found using
- *	the default path.
+ *	or and false if the file was found using the default path.
  */
 mst_Boolean
 findKernelFile(fileName, fullFileName)
@@ -690,13 +696,17 @@ findKernelFile(fileName, fullFileName)
     return (true);
   }
 
-  strcpy(fullFileName, "kernel/");
-  strcat(fullFileName, fileName);
+  sprintf(fullFileName, "%s/%s", kernelFileDefaultPath, fileName);
   if (fileIsReadable(fullFileName)) {
-    return (true);
+    char systemFileName[256];
+    sprintf(systemFileName, KERNEL_PATH "/%s", fileName);
+    /* If this file and the system file are the same, consider the file as
+     * a system file instead. */
+    return (!fileIsReadable(systemFileName) ||
+      getFileModifyTime(fullFileName) != getFileModifyTime(systemFileName));
   }
 
-  sprintf(fullFileName, "%s/%s", kernelFileDefaultPath, fileName);
+  sprintf(fullFileName, KERNEL_PATH "/%s", fileName);
   return (false);
 }
 
@@ -844,27 +854,35 @@ parseArgs(argc, argv)
 {
   char		**av;
   char		*flags;
+  static char	empty[] = "";
 
   smalltalkPassedArgc = 0;
-  while (*++argv) {
-    if (argv[0][0] != '-') {
+  for(smalltalkPassedArgc = 0, argv++; *argv; ) {
+    if (argv[0][0] != '-' || argv[0][1] == '\0') {
+      argv++;
       continue;
     }
-    if (argv[0][1] == '-' && argv[0][2] != '\0') {
-      /* Try to convert a long option to a short one */
-      *argv = parseLongOption(longOptions, argv[0], "-h");
+    if (argv[0][1] == '-') {
+      if (argv[0][2] == '\0') {
+        /* -- stops interpreting of options.  But -- itself is an option! */
+        *argv = empty;
+        break;
+      } else {
+        /* Try to convert a long option to a short one */
+        *argv = parseLongOption(longOptions, argv[0], "-h");
+      }
     }
 
     /* Parse short options */
-    flags = &argv[0][1];
-
-    for (; *flags; flags++) {
+    for (flags = &argv[0][1]; *flags; flags++) {
       switch (*flags) {
 	case 'c': makeCoreFile = true;			break;
 	case 'D': traceKernelDeclarations = true; 	/* fall thru */
 	case 'd': declareTracing = true;		break;
+#ifndef USE_DYNAMIC_TRANSLATION
 	case 'E': traceKernelExecution = true;		/* fall thru */
 	case 'e': executionTracing = true;		break;
+#endif
 	case 'g': gcMessage = false;			break;
 	case 'i': ignoreImage = true;			break;
 	case 'p': emacsProcess = true;			break;
@@ -886,31 +904,48 @@ parseArgs(argc, argv)
 	 * allow to override with the -L option. */
 	case 'l': changeFileName = "change-log.st";     break;
 
-	case '-':
-	  /* stdin is just another file to process, so it matters later */
-	  break;
-
 	case 'a':
 	  /* "Officially", the C command line ends here.  The Smalltalk command
-	      line, instead, starts right after the parameter containing -a */
-	  smalltalkPassedArgv = argv + 1;
-          *argv = nil;
-          while (*++argv) {
+	   *  line, instead, starts right after the parameter containing -a */
+          *argv++ = nil;
+	  smalltalkPassedArgv = argv;
+          while (*argv++) {
             smalltalkPassedArgc++;
           }
-          /* Get back to the last valid argv. ++argv in the while statement
-             will get to the zero pointer. */
-          argv--;
+          argv = smalltalkPassedArgv - 2;
           break;
 
 	case 'I':
+	  if (!argv[1]) {
+	    printf("Missing parameter for -%c.\n", *flags);
+	    printf(helpText);
+            exit(0);
+          }
 	  binaryImageName = argv[1];
 	  for (av = argv+1; *av; av++) {	/* remove this argument */
 	    *av = *(av+1);
 	  }
 	  break;
 
+	case 'K': {
+	  char *oldFileName;
+	  if (!argv[1]) {
+	    printf("Missing parameter for -%c.\n", *flags);
+	    printf(helpText);
+            exit(0);
+          }
+          oldFileName = argv[1];
+	  argv[1] = xmalloc(strlen(imageFileDefaultPath) + strlen(argv[1])+2);
+	  sprintf(argv[1], "%s/%s", imageFileDefaultPath, oldFileName);
+	  break;
+	}
+
 	case 'L':
+	  if (!argv[1]) {
+	    printf("Missing parameter for -%c.\n", *flags);
+	    printf(helpText);
+            exit(0);
+          }
 	  changeFileName = argv[1];
 	  for (av = argv+1; *av; av++) {	/* remove this argument */
 	    *av = *(av+1);
@@ -918,11 +953,7 @@ parseArgs(argc, argv)
 	  break;
 
 	case 'v':
-	  printf("GNU Smalltalk version %s\n", versionString);
-	  printf("written by Steve Byrne (sbb@gnu.org) and Paolo Bonzini (bonzini@gnu.org)\n");
-	  printf(copyrightAndLegalStuffText);
-	  printf("Using default kernel path: %s\n", KERNEL_PATH);
-	  printf("Using default image path : %s\n", IMAGE_PATH);
+	  printf(copyrightAndLegalStuffText, VERSION, KERNEL_PATH, IMAGE_PATH);
 	  exit(0);
 
 	default:
@@ -934,22 +965,13 @@ parseArgs(argc, argv)
           exit(0);
       }
     }
-  }
+    if(smalltalkPassedArgc) {
+      break;
+    }
 
-  if (regressionTesting) {
-    traceKernelDeclarations = declareTracing = false;
-    traceKernelExecution = executionTracing = false;
-    verbose = false;
-  }
-}
+    /* Mark the argument as an option -- searching the command line for
+     * files to be loaded will discard this. */
+    *argv++ = empty;
 
-void
-makeVersionString()
-{
-  if (ST_EDIT_VERSION != 0) {
-    sprintf(versionString, "%d.%d.%s%d", ST_MAJOR_VERSION, ST_MINOR_VERSION,
-	    ST_EDIT_PREFIX, ST_EDIT_VERSION);
-  } else {
-    sprintf(versionString, "%d.%d", ST_MAJOR_VERSION, ST_MINOR_VERSION);
   }
 }
