@@ -1,16 +1,16 @@
-/***********************************************************************
+/******************************** -*- C -*- ****************************
  *
  *	Object Table maintenance module.
  *
- *	$Revision: 1.6.2$
- *	$Date: 1999/08/31 11:23:18$
+ *	$Revision: 1.8.5$
+ *	$Date: 2000/12/27 10:45:49$
  *	$Author: pb$
  *
  ***********************************************************************/
 
 /***********************************************************************
  *
- * Copyright 1990, 91, 92, 94, 95, 99 Free Software Foundation, Inc.
+ * Copyright 1988-92, 1994-95, 1999, 2000 Free Software Foundation, Inc.
  * Written by Steve Byrne.
  *
  * This file is part of GNU Smalltalk.
@@ -37,12 +37,16 @@
 #include "alloc.h"
 #include "oop.h"
 #include "dict.h"
+#include "heap.h"
 #include "save.h"
 #include "comp.h"
 #include "callin.h"
 #include "lex.h"
 #include "lib.h"
+#include "str.h"
 #include "sym.h"
+#include "sysdep.h"
+
 #if STDC_HEADERS
 #include <stdlib.h>
 #include <string.h>
@@ -65,6 +69,9 @@
 /* Define this flag to turn on debugging code for OOP table management */
 /* #define OOP_DEBUGGING */
 
+/* Define this flag to force a GC on every object allocation. */
+/* #define GC_TORTURE */
+
 
 #define alignSize(size) \
  ( ((size) + sizeof(OOP) - 1) & ~(sizeof(OOP) - 1) )
@@ -84,19 +91,21 @@
 
 
 
+/* These are the secondary heaps within which we mantain object data. */
+heap			oopHeap, objectHeap;
 
 /* These are the real OOPS for nil, true, and false */
 OOP			nilOOP, trueOOP, falseOOP;
 
 /* The OOP table.  This contains a pointer to the object, and some flag
-   bits indicating whether the object is read-only, reachable and/or fake.
+   bits indicating whether the object is read-only, reachable and/or pooled.
    Some of the bits indicate the difference between the allocated length
    (stored in the object itself), and the real length, because variable
    byte objects may not be an even multiple of sizeof(voidPtr). */
-struct OOPStruct	*oopTable;
+struct OOPStruct	*oopTable = nil, *allOopsTable;
 OOP                     firstFreeOOP, lastUsedOOP;
 
-int			numFreeOOPs;
+int			numFreeOOPs, oopTableSize;
 
 mst_Boolean		gcState = true, gcMessage = true;
 static mst_Boolean	gcFlipping = false;
@@ -133,12 +142,6 @@ struct NilObjectStruct	nilObject;
    false == &booleanObjects[0], true == &booleanObjects[1] */
 struct BooleanObjectStruct booleanObjects[2];
 
-/* This contains the maximum size of the heap.  It is checked at every
- * allocObj; if the space is too small, it is  brought up to spec before
- * being used
- */
-unsigned long	maxSpaceSize;
-
 /* This variable represents information about the memory space.  memSpace
    holds the required information: basically the pointer to the base and
    top of the space, and the pointers into it for allocation and copying.
@@ -151,20 +154,18 @@ OOP 			*incOOPBasePtr, *incOOPPtr, *incOOPEndPtr;
 
 
 /* static mst_Object	moveObject(); */
-static mst_Boolean	growMemory();
-static void		initSpace(), displayObject(), 
-			displayOOP();
+static void		displayObject(), allocOOPTable(), displayOOP();
 			
-static inline void	markOOPs(), sweepOOPs(), markGlobalOOPs(),
+static inline void	markOOPs(), sweepOOPs(), /* markGlobalOOPs(), */
 			markIncubatorOOPs(), markFinalizableOOPReferences(),
 			checkObjectsAfterSweep(),
-			checkWeakRefs(), realizeOOPs();
+			checkWeakRefs(), sweepPooledContexts();
 
 static unsigned long	prepareForSweep();
 
 
 /*
- *	void initOOPTable()
+ *	void initOOPTable(size)
  *
  * Description
  *
@@ -174,15 +175,23 @@ static unsigned long	prepareForSweep();
  *	defined, so things that have definite classes must wait until
  *	the classes are defined.
  *
+ * Input
+ *
+ *	size of the table to be created
  */
 void
-initOOPTable()
+initOOPTable(size)
+     long	size;
 {
   int i;
 
-  allocOOPTable();
+  oopHeap = NULL;
+  for(i = MAX_OOP_TABLE_SIZE; i && !oopHeap; i >>= 1) {
+    oopHeap = heap_create(i * sizeof(struct OOPStruct));
+  }
+  if (!oopHeap) nomemory(true);
 
-  numFreeOOPs = OOP_TABLE_SIZE;
+  allocOOPTable(size);
 
   nilOOP->flags		= F_READONLY | F_REACHABLE;
   nilOOP->object	= (mst_Object)&nilObject;
@@ -203,28 +212,33 @@ initOOPTable()
     oopTable[i + CHAR_OBJECT_BASE].object = (mst_Object)&charObjectTable[i];
     oopTable[i + CHAR_OBJECT_BASE].flags = F_READONLY | F_REACHABLE;
   }
-
-  dictInit();			/* ### TEMP HACK ### */
 }
 
 
 void
-allocOOPTable()
+allocOOPTable(size)
+     long	size;
 {
   register long i;
+  long		bytes;
 
-  oopTable = (struct OOPStruct *)xmalloc(
-    sizeof(struct OOPStruct) * TOTAL_OOP_TABLE_SLOTS);
+  oopTableSize  = size;
+  bytes		= (size - FIRST_OOP_INDEX) * sizeof(struct OOPStruct);
+  allOopsTable  = (struct OOPStruct *)heap_sbrk(oopHeap, bytes);
+  if (!allOopsTable) nomemory(true);
 
+  oopTable	= &allOopsTable[-FIRST_OOP_INDEX];
   nilOOP	= &oopTable[nilOOPIndex];
   trueOOP	= &oopTable[trueOOPIndex];
   falseOOP	= &oopTable[falseOOPIndex];
 
-  firstFreeOOP  = &oopTable[OOP_TABLE_SIZE - 1];
-  lastUsedOOP   = oopTable;
-  firstFreeOOP->object = NULL;
-  firstFreeOOP->flags = F_FREE;
-  i = OOP_TABLE_SIZE - 1;
+  firstFreeOOP  	= &oopTable[size - 1];
+  lastUsedOOP   	= oopTable;
+  numFreeOOPs		= size;
+  firstFreeOOP->object	= NULL;
+  firstFreeOOP->flags	= F_FREE;
+  
+  i = size - 1;
   while(firstFreeOOP != &oopTable[0]) {
     firstFreeOOP = &oopTable[--i];
     firstFreeOOP->object = (mst_Object) &oopTable[i + 1];
@@ -232,6 +246,35 @@ allocOOPTable()
   }
 }
 
+mst_Boolean
+reallocOOPTable(newSize)
+     long	newSize;
+{
+  long			bytes;
+  OOP			oop;
+
+  bytes		= (newSize - oopTableSize) * sizeof(struct OOPStruct);
+  if (bytes < 0) {
+    return (true);
+  }
+
+  if (!heap_sbrk(oopHeap, bytes)) {
+    /* try to recover.  Note that we cannot move the OOP table like
+     * we do with the object data. */
+    nomemory(false);
+    return (false);
+  }
+
+  /* mark the new OOPs as available */
+  for (oop = &oopTable[oopTableSize]; oop < &oopTable[newSize]; oop++) {
+    oop->flags = F_FREE;
+  }
+
+  firstFreeOOP = &oopTable[newSize - 1];
+  oopTableSize = newSize;
+  refreshOOPFreeList();
+  return (true);
+}
 
 /*
  *	void initBuiltinObjectsClasses()
@@ -317,87 +360,6 @@ findAnInstance(classOOP)
 }
 
 
-#ifdef ACCESSOR_DEBUGGING
-/*
- *	long oopIndex(oop)
- *
- * Description
- *
- *	Returns the index within the OOP table of the given OOP.
- *
- * Inputs
- *
- *	oop   : OOP to return index of
- *
- * Outputs
- *
- *	Returned index in the OOP table, in range 0..TOTAL_OOP_TABLE_SLOTS.
- */
-long
-oopIndex(oop)
-     OOP	oop;
-{
-  return (oopIndexMac(oop));
-}
-#endif /* ACCESSOR_DEBUGGING */
-
-/*
- *	mst_Boolean oopIndexValid(index)
- *
- * Description
- *
- *	Checks to see if index represents a valid OOP.
- *
- * Inputs
- *
- *	index : a long index into the OOP table, apparently 1 based due to
- *		being called from Smalltalk via a primitive.
- *
- * Outputs
- *
- *	True if the index represents a valid OOP table element, false
- *	otherwise.
- */
-mst_Boolean
-oopIndexValid(index)
-     long	index;
-{
-  return (index >= 1 && index <= TOTAL_OOP_TABLE_SLOTS);
-}
-
-#ifdef ACCESSOR_DEBUGGING
-
-OOP
-oopAt(index)
-     long	index;
-{
-  return (oopAtMac(index));
-}
-
-#endif  /* ACCESSOR_DEBUGGING */
-
-
-#ifdef ACCESSOR_DEBUGGING
-
-mst_Boolean
-oopValid(oop)
-     OOP	oop;
-{
-  return (oopValidMac(index));
-}
-
-
-mst_Boolean
-oopAvailable(index)
-     long	index;
-{
-  return (oopAvailableMac(index));
-}
-
-#endif /* ACCESSOR_DEBUGGING */
-
-
-
 void
 swapObjects(oop1, oop2)
      OOP	oop1, oop2;
@@ -422,6 +384,30 @@ charOOPValue(charOOP)
 {
   return (charOOP - &oopTable[CHAR_OBJECT_BASE]);
 }
+
+/*
+ *	mst_Boolean oopIndexValid(index)
+ *
+ * Description
+ *
+ *	Checks to see if index represents a valid OOP.
+ *
+ * Inputs
+ *
+ *	index : a long index into the OOP table, apparently 1 based due to
+ *		being called from Smalltalk via a primitive.
+ *
+ * Outputs
+ *
+ *	True if the index represents a valid OOP table element, false
+ *	otherwise.
+ */
+mst_Boolean
+oopIndexValid(index)
+     long	index;
+{
+  return (index >= FIRST_OOP_INDEX && index <= oopTableSize);
+}
 
 
 void
@@ -462,7 +448,7 @@ classifyAddr(addr)
   } else if (isObjAddr(addr)) {
     displayObject(addr);
   } else if isInt(addr) {
-    printf("Smalltalk Integer %ld\n", toInt(addr));
+    printf("Smalltalk SmallInteger %ld\n", toInt(addr));
   } else {
     printf("Address %p is not a Smalltalk entity\n", addr);
   }
@@ -480,7 +466,7 @@ displayOOP(oop)
     return;
   }
 
-  isBuiltin = (oop >= &oopTable[OOP_TABLE_SIZE]) ? true : false;
+  isBuiltin = (oop < oopTable);
 
   if (!isBuiltin) {
     printf ("OOP %p [%ld]\n", oop, (unsigned long) (oop - oopTable));
@@ -565,7 +551,7 @@ allocObj(size)
       percent = (memSpace.newAllocPtr - memSpace.space)
         * 100 / memSpace.totalSize;
 
-      if (percent > growThresholdPercent) {
+      if (percent >= growThresholdPercent) {
         gcFlip();
       } else {
         minorGCFlip();
@@ -579,9 +565,12 @@ allocObj(size)
       unsigned long spaceNeeded;
 
       spaceInUse = memSpace.allocPtr - memSpace.space;
-      spaceNeeded = size + spaceInUse; /* we need what we already are using,
-					plus some  */
-      spaceNeeded = (unsigned long)(spaceNeeded * (1.0 + spaceGrowRate/100.0));	/* allow breathing room */
+      
+      /* we need what we already are using, plus some breathing room */
+      spaceNeeded = size + spaceInUse;
+      if (spaceGrowRate > 0) {
+        spaceNeeded += spaceNeeded / 100 * spaceGrowRate;
+      }
       if (!growMemoryTo(spaceNeeded)) {
 	/* !!! do something more reasonable in the future */
 	errorf("Cannot recover, exiting...");
@@ -610,7 +599,7 @@ allocObj(size)
  *	Previous state of the garbage collector (on or off).
  */
 mst_Boolean
-     gcOff()
+gcOff()
 {
   mst_Boolean	oldGCState;
 
@@ -675,8 +664,23 @@ setGCState(state)
 void
 initMem()
 {
-  maxSpaceSize = INIT_MEM_SPACE_SIZE;
-  initSpace(&memSpace);
+  int i;
+  
+  objectHeap = NULL;
+  for(i = MAX_OBJECT_DATA_SIZE; i && !objectHeap; i >>= 1) {
+    objectHeap = heap_create(i);
+  }
+  if (!objectHeap) nomemory(true);
+
+  memSpace.totalSize = INIT_MEM_SPACE_SIZE;
+  memSpace.space = (char *)heap_sbrk(objectHeap,
+    memSpace.totalSize + NEW_GENERATION_SIZE);
+  
+  if (!memSpace.space) nomemory(true);
+
+  memSpace.allocPtr = memSpace.space;
+  memSpace.newAllocPtr = memSpace.space;
+  memSpace.maxPtr = memSpace.space + memSpace.totalSize;
 
   gcOff();
   incInitRegistry();
@@ -703,30 +707,24 @@ setSpaceInfo(size)
  * Description
  *
  *	Grows the allocated memory to the given size in bytes, if it's not
- *	there already.  In the future, the gc should give some indication that
- *	it was not able to acquire the required number of bytes, and this
- *	function will return false, causing the system primitive to fail.
+ *	there already.
  *
  * Inputs
  *
  *	spaceSize: 
- *		Size in bytes to grow the spaces to.
+ *		Size in bytes to grow the memory space to.
  *
  * Outputs
  *
  *	True if the operation was successful, False if the memory could not be
- *	allocated (this is not currently implemented).
+ *	allocated.
  */
 mst_Boolean
 growTo(spaceSize)
      unsigned long spaceSize;
 {
-  if (spaceSize > maxSpaceSize) {
-    maxSpaceSize = spaceSize;
-    gcFlip();
-  }
-  
-  return true;
+  gcFlip();
+  return growMemoryTo(spaceSize);
 }
 
 
@@ -735,7 +733,7 @@ growTo(spaceSize)
  *
  * Description
  *
- *	Grows the memory segment to sapceSize.  Shoudl be called after the
+ *	Grows the memory segment to sapceSize.  Should be called after the
  *	sweep has occurred so that things are contiguous.  Ensures that the OOP
  *	table pointers are fixed up to point to the new objects.
  *
@@ -754,57 +752,72 @@ mst_Boolean
 growMemoryTo(spaceSize)
      unsigned long spaceSize;
 {
-  if (spaceSize > maxSpaceSize) {
-    maxSpaceSize = spaceSize;
-    return growMemory();
-  }
-  
-  return true;
-}
-  
-
-
-/*
- *	static mst_Boolean growMemory()
- *
- * Description
- *
- *	Grows the memory segment to maxSpaceSize.  Should be called after the
- *	sweep has occurred so that things are contiguous.  Ensures that the OOP
- *	table pointers (the live ones) are fixed up to point to the new
- *	objects. 
- *
- */
-mst_Boolean
-growMemory()
-{
-  unsigned long spaceDelta;
-  char		*oldSpacePtr;
+  long		spaceDelta;
+  long		bytes;
+  char		*newSpacePtr;
   OOP		oop;
+  heap		newHeap;
+
+  if (spaceSize <= memSpace.totalSize) {
+    return (true);
+  }
+
+  if (spaceSize > MAX_OBJECT_DATA_SIZE - NEW_GENERATION_SIZE) {
+    if (memSpace.totalSize >= MAX_OBJECT_DATA_SIZE - NEW_GENERATION_SIZE) {
+      return (false);
+    } else {
+      spaceSize = MAX_OBJECT_DATA_SIZE - NEW_GENERATION_SIZE;
+    }
+  }
+
+  /* Do the real work... */
+  bytes = spaceSize + NEW_GENERATION_SIZE;
+  if (heap_sbrk(objectHeap, spaceSize - memSpace.totalSize)) {
+    newSpacePtr = memSpace.space;
+    newHeap = objectHeap;
+  } else {
+    int i;
+
+    /* Try to move to a newly allocated heap */  
+    newHeap = NULL;
+    for(i = MAX_OBJECT_DATA_SIZE; i > bytes && !newHeap; i >>= 1) {
+      newHeap = heap_create(i);
+    }
+
+    if (!newHeap ||
+        !(newSpacePtr = heap_sbrk(newHeap, bytes))) {
+      nomemory(false);
+      spaceSize = memSpace.totalSize;
+      if (newHeap) {
+	heap_destroy(newHeap);
+      }
+      return (false);
+    }
+  }
+
+#ifdef GROW_DEBUG
+  printf("old = %8x, new = %8x, delta = %8x\n", memSpace.space, newSpacePtr, spaceDelta);
+#endif
 
   fixupObjectPointers();
-  oldSpacePtr = memSpace.space;
-  memSpace.space = (char *)xrealloc(memSpace.space, maxSpaceSize + NEW_GENERATION_SIZE);
-  if (!memSpace.space) {
-    printf("\n\n[Memory allocation failure]\nCan't allocate enough memory to continue.\n");
-    /* try to recover */
-    maxSpaceSize = memSpace.totalSize;
-    memSpace.space = oldSpacePtr;
-    return (false);
+
+  spaceDelta = newSpacePtr - memSpace.space;
+  if (spaceDelta) {
+    memcpy(newSpacePtr, memSpace.space, spaceSize + NEW_GENERATION_SIZE);
+    heap_destroy(objectHeap);
+    objectHeap = newHeap;
   }
-  spaceDelta = memSpace.space - oldSpacePtr;
-  memSpace.totalSize = maxSpaceSize;
+
+  memSpace.space = newSpacePtr;
+  memSpace.totalSize = spaceSize;
   memSpace.allocPtr += spaceDelta;
   memSpace.newAllocPtr += spaceDelta;
   memSpace.maxPtr = memSpace.space + memSpace.totalSize;
-#ifdef GROW_DEBUG
-  printf("old = %8x, new = %8x, delta = %8x\n", oldSpacePtr, memSpace.space, spaceDelta);
-#endif
 
-  if (oopTable) {
+  if (spaceDelta && oopTable) {
     /* Fix up the OOP table pointers to objects */
     for (oop = oopTable; oop <= lastUsedOOP; oop++) {
-      if (!(oop->flags & (F_FREE | F_FAKE))) {
+      if (!(oop->flags & (F_FREE | F_POOLED))) {
 #ifdef GROW_DEBUG
         voidPtr obj;
         printf("old = %8x, ", obj = (voidPtr) oop->object);
@@ -816,11 +829,11 @@ growMemory()
       }
     }
   }
+
   restoreObjectPointers();
-
-  return true;
+  return (true);
 }
-
+  
 
 
 /*
@@ -856,6 +869,11 @@ gcFlip()
     fflush(stderr);
   }
 
+  /* In allocOOP, we don't want to worry about going below the low-water
+   * threshold, so we massage numFreeOOPs.  The real number of free OOPs
+   * will be computed in prepareForSweep. */
+  numFreeOOPs += LOW_WATER_OOP_THRESHOLD;
+
   gcFlipping = true;
   fixupObjectPointers();
   markOOPs();
@@ -872,12 +890,14 @@ gcFlip()
   
   if (lastPercent > growThresholdPercent) {
     /* with bad grow rates, can undergrow.  Takes care of it */
-    maxSpaceSize += maxSpaceSize / 100 * spaceGrowRate;
-    maxSpaceSize &= ~(sizeof(long)-1);		/* round to word boundary */
-    if (maxSpaceSize > memSpace.totalSize) {
-      growMemory();
+    unsigned long newSpaceSize;
+
+    debug();
+    newSpaceSize = memSpace.totalSize / 100 * (100 + spaceGrowRate);
+    newSpaceSize &= ~(sizeof(long)-1);		/* round to word boundary */
+    if (newSpaceSize > memSpace.totalSize) {
+      growMemoryTo(newSpaceSize);
     }
-    maxSpaceSize = memSpace.totalSize;
   }
 
   if (gcMessage && !regressionTesting) {
@@ -893,9 +913,9 @@ gcFlip()
  *
  * Description
  *
- *	Mark & copy the fake objects.  Starting from the root set, recursively
- *	mark objects as reachable, then copy the actually reachable ones to the
- *	old generation.
+ *	Mark & sweep the new objects.  Starting from the root set, recursively
+ *	mark objects as reachable, then include the actually reachable ones to
+ *	the old generation.
  *
  */
 void
@@ -920,6 +940,11 @@ minorGCFlip()
     fflush(stderr);
   }
 
+  /* In allocOOP, we don't want to worry about going below the low-water
+   * threshold, so we massage numFreeOOPs.  The real number of free OOPs
+   * will be computed in prepareForSweep. */
+  numFreeOOPs += LOW_WATER_OOP_THRESHOLD;
+
   gcFlipping = true;
   fixupObjectPointers();
   markOOPs();
@@ -935,6 +960,7 @@ minorGCFlip()
     fflush(stderr);
   }
 
+  checkObjectsAfterSweep();
 }
 
 
@@ -948,26 +974,24 @@ minorGCFlip()
  *	other references, the slot of the weak object is replaced by
  *	a nil.
  *	When a finalizable object has no references outside weak objects,
- *	prepareForSweep() marks it as F_FINALIZING.  This routine (which
- *	is executed after the sweep) unmarks it and calls back the fina-
- *	lize method.
+ *	prepareForSweep() marks it in a buffer.  This routine (which
+ *	is executed after the sweep) calls back the finalize method.
  *
  */
 void
 checkObjectsAfterSweep()
 {
-  register OOP  oop;
+  register OOP	*pOOP;
+  register long	size;
 
-  for (oop = oopTable; oop <= lastUsedOOP; oop++) {
-    if (oop->flags & F_FINALIZING) {
-      /* What happens when a GC occurs during finalization?!? Good question!
-         The inner occurrence of checkObjectsAfterSweep will proceed finalizing
-         things, unmarking them on the fly.  This way the outer call to this
-         function won't ever know of the objects finalized by the inner call.
-         Eh eh! */
-      oop->flags ^= F_FINALIZING;
-      msgSend(oop, finalizeSymbol, nil);
-    }
+  /* Make a local copy of the buffer */
+  size = bufferSize();
+  pOOP = alloca(size);
+  copyBuffer(pOOP);
+  size /= SIZEOF_CHAR_P;
+
+  while (size--) {
+    msgSend(*pOOP++, finalizeSymbol, nil);
   }
 }
 
@@ -986,12 +1010,12 @@ checkWeakRefs(oop)
       continue;
     }
     if (*field <= oop) {
-      /* Not yet scanned */
+      /* Not yet scanned by prepareForSweep */
       if (!isOOPMarked(*field)) {
 	*field = nilOOP;
       }
     } else {
-      /* Already scanned */
+      /* Already scanned by prepareForSweep */
       if (isOOPFree(*field)) {
 	*field = nilOOP;
       }
@@ -1001,7 +1025,7 @@ checkWeakRefs(oop)
 
 void markOOPs()
 {
-  markGlobalOOPs();
+  markAnOOPInternal(smalltalkDictionary, nil, nil);
   markProcessorRegisters();
   markRegisteredOOPs();
   markIncubatorOOPs();
@@ -1066,31 +1090,30 @@ markAnOOPInternal(oop, curOOP, atEndOOP)
 {
   for (;;) {
     if (!oop) {						/* in the loop! */
+#ifndef OPTIMIZE
+      mst_Object obj = (mst_Object) (curOOP - 1);	/* for debugging */
+#endif
     iterationLoop:
       /* in a loop, do next iteration */
       oop = *curOOP;
       curOOP++;
+      if (isOOP(oop)) {
 #ifndef OPTIMIZE
-      if (!isInt(oop) && !isOOPAddr(oop)) {
-	debug();
-	printf("Error! Bad OOP %p is being marked!", oop);
-	printf("\n");
-	if (curOOP < atEndOOP) {
-	  goto iterationLoop;
-	} else {
-	  break;
-	}
-      }
+	if (!isOOPAddr(oop)) {
+	  printf("Error! Invalid OOP %p was found inside %p!\n", oop, obj);
+	  debug();
+        } else
 #endif
-      if (isOOP(oop) && !isOOPMarked(oop)) {
-	if (curOOP < atEndOOP) {
-	  markAnOOPInternal(oop, nil, nil);
-	  goto iterationLoop;
-	} else {
-	  /* On the last object in the set, reuse the current invocation.
-	     oop is valid, so we go to the single-object case */
-	  continue;
-	}
+        if (!isOOPMarked(oop)) {
+	  if (curOOP < atEndOOP) {
+	    markAnOOPInternal(oop, nil, nil);
+	    goto iterationLoop;
+	  } else {
+	    /* On the last object in the set, reuse the current invocation.
+	     * oop is valid, so we go to the single-object case */
+	    continue;
+	  }
+        }
       }
       /* We reach this point if the object isn't to be marked.  The code above
          contains a continue to tail recurse, so we cannot put the loop in a
@@ -1105,10 +1128,14 @@ markAnOOPInternal(oop, curOOP, atEndOOP)
       unsigned long	size;
 
 #ifndef OPTIMIZE
-      if (!isFake(oop) && !isObjAddr(oopToObj(oop))) {
+      if (isOOPFree(oop)) {
+        printf("Error! Free OOP %p is being marked!\n", oop);
+        debug();
+        break;
+      }
+      if (!isPooled(oop) && !isObjAddr(oopToObj(oop))) {
+        printf("Error! OOP at %p points to invalid object data at %p!\n", oop, oop->object);
         debug();
-        printf("Error! Bad OOP at %p (%p) is being marked!", oop, oop->object);
-        printf("\n");
         break;
       }
 #endif
@@ -1159,13 +1186,13 @@ refreshOOPFreeList()
 
   /* The free list will be reconstructed */
   newFirstFree = NULL;
-  numFreeOOPs = OOP_TABLE_SIZE;
+  numFreeOOPs = oopTableSize;
 
   /* Proceed backwards so that the first free OOPs are at the head of the
      free list.  This minimizes the amount of space used by the OOP table in
      a saved image. */
 
-  for (oop = &oopTable[OOP_TABLE_SIZE - 1]; oop >= oopTable; oop--) {
+  for (oop = &oopTable[oopTableSize - 1]; oop >= oopTable; oop--) {
     if (oop->flags & F_FREE) {
       oop->object  = (mst_Object) newFirstFree;
       newFirstFree = oop;
@@ -1203,7 +1230,7 @@ refreshOOPFreeList()
  *
  * Outputs
  *
- *	The number of bytes to be allocated for reachable fake objects.
+ *	The number of bytes to be allocated for reachable pooled objects.
  */
 unsigned long
 prepareForSweep(bottom)
@@ -1211,12 +1238,14 @@ prepareForSweep(bottom)
 {
   register OOP  	newFirstFree, oop;
   register mst_Object	object;
-  register unsigned long size;
+  register unsigned long pooledSize;
 
   /* The free list will be reconstructed, but the unused OOPs are unchanged */
   newFirstFree = firstFreeOOP;
-  numFreeOOPs = OOP_TABLE_SIZE;
-  size = 0;
+  numFreeOOPs = oopTableSize;
+  pooledSize = 0;
+
+  resetBuffer();
 
   /* Proceed backwards so that the first free OOPs are at the head of the
      free list.  This minimizes the amount of space used by the OOP table in
@@ -1228,9 +1257,9 @@ prepareForSweep(bottom)
     }
 
     if (oop->flags & (F_FINALIZE | F_REACHABLE)) {
-      if (oop->flags & F_FAKE) {
-        /* Reachable fake object.  Realize it. */
-        size += size2Bytes(oopToObj(oop)->objSize);
+      if (oop->flags & F_POOLED) {
+        /* Reachable pooled object.  Get it to the heap. */
+        pooledSize += size2Bytes(oopToObj(oop)->objSize);
 
       } else if (((voidPtr) oop->object) >= bottom) {
         /* Object reachable after mark phase.  Reverse the pointer */
@@ -1244,7 +1273,8 @@ prepareForSweep(bottom)
          * weak reference to it, so we mark it so that finalization will occur
          * soon after the end of the sweep pass.  In the meanwhile, we let it
          * survive and decide the object's fate at the next GC pass */
-        oop->flags ^= F_FINALIZE | F_FINALIZING;
+        oop->flags ^= F_FINALIZE;
+        addBufPointer(oop);
       }
       oop->flags &= ~F_REACHABLE;
       numFreeOOPs--;
@@ -1259,37 +1289,45 @@ prepareForSweep(bottom)
       }
     }
   }
+  
   firstFreeOOP = newFirstFree;
 
-  return (size);
+#ifdef USE_JIT_TRANSLATION
+  /* Go and really free the blocks associated to garbage collected
+   * native code. */
+  freeNativeCode();
+#endif
+
+  return (pooledSize);
 }
 
 /*
- *	static void realizeOOPs(size)
+ *	static void sweepPooledContexts(dest)
  *
  * Description
  *
- *	Iterate through the OOP table.  Copy fake, marked objects
- *	to the main heap space, free the others.
+ *	Iterate through the OOP table.  Copy pooled, marked objects
+ *	to the main heap space (starting at address dest), free the
+ *	others.
  *
  */
 void
-realizeOOPs(object)
-     register char *object;
+sweepPooledContexts(dest)
+     register char *dest;
 {
   register OOP		 oop;
   register unsigned long size;
 
   for (oop = lastUsedOOP; oop >= oopTable; oop--) {
-    if (oop->flags & F_FAKE) {
+    if (oop->flags & F_POOLED) {
       size = size2Bytes(oopToObj(oop)->objSize);
-      memcpy(object, oop->object, size);
-      oop->object = (mst_Object) object;
-      oop->flags &= ~(F_FAKE | F_REACHABLE);
-      object += size;
+      memcpy(dest, oop->object, size);
+      oop->object = (mst_Object) dest;
+      oop->flags &= ~(F_POOLED | F_REACHABLE);
+      dest += size;
     }
   }
-  resetFakeContexts();
+  emptyContextPool();
 }
 
 /* #define SWEEP_DEBUG */
@@ -1298,11 +1336,11 @@ sweepOOPs(bottom)
      char *bottom;
 {
   register char		*from, *fromStart, *to;
-  unsigned long		chunkSize, realizedSize;
+  unsigned long		chunkSize, pooledSize;
   register mst_Object	object;
   OOP			curClass, oop;
 
-  realizedSize = prepareForSweep(bottom);
+  pooledSize = prepareForSweep(bottom);
 
   /* Algorithm:
      initialize:
@@ -1405,8 +1443,8 @@ sweepOOPs(bottom)
 
   memSpace.allocPtr = to;
   
-  object = allocObj(realizedSize);
-  realizeOOPs((char *) object);
+  object = allocObj(pooledSize);
+  sweepPooledContexts((char *) object);
   memSpace.newAllocPtr = memSpace.allocPtr;
 }
 
@@ -1427,67 +1465,28 @@ markFinalizableOOPReferences()
   }
 }
 
-void
-markGlobalOOPs()
-{
-  OOP		**oopPtr, oop;
-
-  /* first mark them as reachable to work on the global objects */
-  /* one at a time and have more locality of reference */
-  for (oopPtr = globalOOPs; *oopPtr; oopPtr++) {
-    oop = **oopPtr;
-    oop->flags |= F_REACHABLE;
-  }
-  /* then do the real marking */
-  for (oopPtr = globalOOPs; *oopPtr; oopPtr++) {
-    markAnOOPInternal(**oopPtr, nil, nil);
-  }
-}
-
-
-/*
- *	static void initSpace(space)
- *
- * Description
- *
- *	Initializes the allocation and copying pointers for space SPACE.
- *
- * Inputs
- *
- *	space : Pointer to the struct memorySpaceStruct to be initialized.
- *
- */
-void
-initSpace(space)
-     struct memorySpaceStruct *space;
-{
-  space->space = (char *)xmalloc(maxSpaceSize + NEW_GENERATION_SIZE);
-  space->totalSize = maxSpaceSize;
-  space->allocPtr = space->space;
-  space->newAllocPtr = space->space;
-  space->maxPtr = space->space + space->totalSize;
-}
+#ifdef not_needed
+/**/void
+/**/markGlobalOOPs()
+/**/{
+/**/  OOP		**oopPtr, oop;
+/**/
+/**/  /* first mark them as reachable to work on the global objects */
+/**/  /* one at a time and have more locality of reference */
+/**/  for (oopPtr = globalOOPs; *oopPtr; oopPtr++) {
+/**/    oop = **oopPtr;
+/**/    oop->flags |= F_REACHABLE;
+/**/  }
+/**/  /* then do the real marking */
+/**/  for (oopPtr = globalOOPs; *oopPtr; oopPtr++) {
+/**/    markAnOOPInternal(**oopPtr, nil, nil);
+/**/  }
+/**/}
+#endif
 
 
 
 
-/*
- *	void debug()
- *
- * Description
- *
- *	Used for debugging.  You set a breakpoint in the debug routine in the
- *	debugger, and have code call it when you want it to stop.  Performs no
- *	action normally.
- *
- */
-void
-debug()
-{
-/* getchar(); */
-}
-
-
 #ifdef GC_TORTURE
 
 /*
@@ -1537,15 +1536,11 @@ stopGCTorture()
 
 
 
-/**********************************************************************
-***********************************************************************
-***********************************************************************
-*
-* Incubator support routines
-*
-***********************************************************************
-***********************************************************************
-***********************************************************************/
+/***********************************************************************
+ *
+ * Incubator support routines
+ *
+ ***********************************************************************/
 
 void
 incInitRegistry()
