@@ -1,16 +1,16 @@
-/***********************************************************************
+/******************************** -*- C -*- ****************************
  *
  *	Binary image save/restore.
  *
- *	$Revision: 1.6.2$
- *	$Date: 1999/08/31 11:23:18$
+ *	$Revision: 1.8.5$
+ *	$Date: 2000/12/27 10:45:49$
  *	$Author: pb$
  *
  ***********************************************************************/
 
 /***********************************************************************
  *
- * Copyright 1990, 91, 92, 94, 95, 99 Free Software Foundation, Inc.
+ * Copyright 1988-92, 1994-95, 1999, 2000 Free Software Foundation, Inc.
  * Written by Steve Byrne.
  *
  * This file is part of GNU Smalltalk.
@@ -40,10 +40,14 @@
 #include "sym.h"
 #include "oop.h"		/* defines oopTable */
 #include "lex.h"
-#include "cint.h"
 #include "sysdep.h"
 #include <stdio.h>
 
+#if STDC_HEADERS
+#include <string.h> /* for memcpy */
+#include <stdlib.h>
+#endif /* STDC_HEADERS */
+
 #ifndef MAXPATHLEN
 #define MAXPATHLEN		1024 /* max length of a file and path */
 #endif
@@ -54,21 +58,21 @@
 #define MASK_LONG_SIZE_FLAG 2
 
 #ifdef WORDS_BIGENDIAN
-#  define LOCAL_ENDIANNESS_FLAG MASK_ENDIANNESS_FLAG
+# define LOCAL_ENDIANNESS_FLAG MASK_ENDIANNESS_FLAG
 #else
-#  define LOCAL_ENDIANNESS_FLAG 0
+# define LOCAL_ENDIANNESS_FLAG 0
 #endif
 
 #if SIZEOF_LONG == 4
-#  define LOCAL_LONG_SIZE_FLAG 0
-#  define byteInvert(x) \
+# define LOCAL_LONG_SIZE_FLAG 0
+# define byteInvert(x) \
         ((unsigned long int)((((unsigned long int)(x) & 0x000000ffU) << 24) | \
                              (((unsigned long int)(x) & 0x0000ff00U) <<  8) | \
                              (((unsigned long int)(x) & 0x00ff0000U) >>  8) | \
                              (((unsigned long int)(x) & 0xff000000U) >> 24)))
 #else
-#  define LOCAL_LONG_SIZE_FLAG MASK_LONG_SIZE_FLAG
-#  define byteInvert(x) \
+# define LOCAL_LONG_SIZE_FLAG MASK_LONG_SIZE_FLAG
+# define byteInvert(x) \
         ((unsigned long int)((((unsigned long int)(x) & 0x00000000000000ffU) << 56) | \
                              (((unsigned long int)(x) & 0x000000000000ff00U) << 40) | \
                              (((unsigned long int)(x) & 0x0000000000ff0000U) << 24) | \
@@ -100,183 +104,32 @@
 
 typedef struct SaveFileHeaderStruct {
   unsigned char signature[7];	/* 7+1=8 should be enough to align version! */
-  unsigned char flags;		/* misc size */
+  unsigned char flags;		/* flags for endianness and sizeof(long) */
   unsigned long version;	/* the Smalltalk version that made this dump */
   unsigned long	objectDataSize;	/* size of object data section in bytes */
   unsigned long	oopTableSize;	/* size of the oop table at dump */
-  unsigned long	semiSpaceSize;	/* size of the semi spaces at dump time */
+  unsigned long	memSpaceSize;	/* size of the semi spaces at dump time */
 } SaveFileHeader;
 
-
-OOP		*globalOOPs[] = {
-  &mst_objectClass,
-  &magnitudeClass,
-  &charClass,
-  &timeClass,
-  &numberClass,
-  &floatClass,
-  &integerClass,
-  &largeIntegerClass,
-  &largeNegativeIntegerClass,
-  &largePositiveIntegerClass,
-  &largeZeroIntegerClass,
-  &fractionClass,
-  &associationClass,
-  &linkClass,
-  &processClass,
-  &symLinkClass,
-  &collectionClass,
-  &sequenceableCollectionClass,
-  &linkedListClass,
-  &semaphoreClass,
-  &arrayedCollectionClass,
-  &arrayClass,
-  &stringClass,
-  &symbolClass,
-  &byteArrayClass,
-  &compiledMethodClass,
-  &intervalClass,
-  &orderedCollectionClass,
-  &sortedCollectionClass,
-  &bagClass,
-  &mappedCollectionClass,
-  &setClass,
-  &dictionaryClass,
-  &identityDictionaryClass,
-  &systemDictionaryClass,
-  &undefinedObjectClass,
-  &booleanClass,
-  &falseClass,
-  &trueClass,
-  &processorSchedulerClass,
-  &delayClass,
-  &sharedQueueClass,
-  &behaviorClass,
-  &classDescriptionClass,
-  &classClass,
-  &metaclassClass,
-  &messageClass,
-  &contextPartClass,
-  &methodContextClass,
-  &blockContextClass,
-  &blockClosureClass,
-  &streamClass,
-  &positionableStreamClass,
-  &readStreamClass,
-  &writeStreamClass,
-  &readWriteStreamClass,
-  &byteStreamClass,
-  &cObjectClass,
-  &cTypeClass,
-  &fileStreamClass,
-  &memoryClass,
-  &byteMemoryClass,
-  &wordMemoryClass,
-  &randomClass,
-  &cFuncDescriptorClass,
-  &tokenStreamClass,
-  &methodInfoClass,
-  &fileSegmentClass,
-  &andColonSymbol,
-  &atColonPutColonSymbol,
-  &atColonSymbol,
-  &atEndSymbol,
-  &bitAndColonSymbol,
-  &bitOrColonSymbol,
-  &bitShiftColonSymbol,
-  &blockCopyColonTemporariesColonSymbol,
-  &classSymbol,
-  &divideSymbol,
-  &doColonSymbol,
-  &equalSymbol,
-  &finalizeSymbol,
-  &greaterEqualSymbol,
-  &greaterThanSymbol,
-  &ifFalseColonIfTrueColonSymbol,
-  &ifFalseColonSymbol,
-  &ifTrueColonIfFalseColonSymbol,
-  &ifTrueColonSymbol,
-  &integerDivideSymbol,
-  &isNilSymbol,
-  &lessEqualSymbol,
-  &lessThanSymbol,
-  &minusSymbol,
-  &newColonSymbol,
-  &newSymbol,
-  &nextPutColonSymbol,
-  &nextSymbol,
-  &notEqualSymbol,
-  &notNilSymbol,
-  &notSameObjectSymbol,
-  &orColonSymbol,
-  &plusSymbol,
-  &remainderSymbol,
-  &sameObjectSymbol,
-  &sizeSymbol,
-  &thisContextSymbol,
-  &timesSymbol,
-  &timesRepeatColonSymbol,
-  &toColonDoColonSymbol,
-  &toColonByColonDoColonSymbol,
-  &valueColonSymbol,
-  &valueColonValueColonSymbol,
-  &valueColonValueColonValueColonSymbol,
-  &valueWithArgumentsColonSymbol,
-  &valueSymbol,
-  &whileFalseColonSymbol,
-  &whileTrueColonSymbol,
-  &yourselfSymbol,
-  &orSymbol,
-  &andSymbol,
-  &superSymbol,
-  &nilSymbol,
-  &trueSymbol,
-  &falseSymbol,
-  &selfSymbol,
-  &doesNotUnderstandColonSymbol,
-  &unknownSymbol,
-  &charSymbol,
-  &stringSymbol,
-  &stringOutSymbol, 
-  &symbolSymbol,
-  &intSymbol,
-  &longSymbol,
-  &floatSymbol,
-  &doubleSymbol,
-  &voidSymbol,
-  &variadicSymbol,
-  &cObjectSymbol,
-  &cObjectPtrSymbol,
-  &smalltalkSymbol,
-  &byteArraySymbol,
-  &byteArrayOutSymbol,
-  &booleanSymbol,
-  &cObjectTypeCType,
-  &processorOOP,
-  &symbolTable,
-  &smalltalkDictionary,
-  nil
-};
-
 
 
-static void	skipOverHeader(), saveObject(), fixupObject(),
-		restoreObject(), saveOOPTable(), fixupOOP(),
-		addByteFlagToOOP(), fixupByteOrder(),
-		restoreAllOOPs(), restoreInstanceVars(),
-		loadOOPTable(), loadNormalOOPs(),
-		/* loadCharOOPs(), loadSpecialOOPs(), */
-		saveGlobalOOPs(), loadGlobalOOPs(),
-		saveFileVersion(), loadFileVersion(),
-		fixupInstanceVars(), fixupOOPInstanceVars(),
-		skipToHeader();
+static void		skipOverHeader(), saveObject(),
+			saveOOPTable(), fixupOOP(),
+			addByteFlagToOOP(), fixupByteOrder(),
+			restoreAllOOPs(), restoreInstanceVars(),
+			loadOOPTable(), loadNormalOOPs(),
+			/* loadCharOOPs(), loadSpecialOOPs(), */
+			/* saveGlobalOOPs(), loadGlobalOOPs(), */
+			saveFileVersion(), loadFileVersion(),
+			fixupInstanceVars(), fixupOOPInstanceVars(),
+			skipToHeader();
 
-static int	saveAllObjects(), my_fread();
+static int		saveAllObjects(), my_fread();
 
 static mst_Boolean	wrongEndianness;
 
 /* This variable contains the OOP slot index of the highest non-free OOP,
- * excluding the built-in ones (i.e., it will always be < OOP_TABLE_SIZE).
+ * excluding the built-in ones (i.e., it will always be < oopTableSize).
  * This is used for optimizing the size of the saved image, and minimizing
  * the load time when restoring the system. */
 static int	maxUsedOOPSlot = 0;
@@ -316,8 +169,6 @@ saveToFile(fileName)
 printf("After saving oopt table: %d\n", ftell(imageFile));
 #endif /* OOP_TABLE_TRACE */
 
-  saveGlobalOOPs(imageFile);
-
 #ifdef OOP_TABLE_TRACE
 printf("After global oop table: %d\n", ftell(imageFile));
 #endif /* OOP_TABLE_TRACE */
@@ -338,7 +189,7 @@ printf("After saving all objects table: %d\n", ftell(imageFile));
 } 
 
 
-static void
+void
 skipOverHeader(imageFile)
      FILE	*imageFile;
 {
@@ -363,25 +214,24 @@ skipOverHeader(imageFile)
  *		correctly before this routine is called.
  *
  */
-static void
+void
 saveOOPTable(imageFile)
      FILE	*imageFile;
 {
-  int		i;
   OOP		oop;
 
   maxUsedOOPSlot = 0;
 
-  for (i = 0, oop = oopTable; i < TOTAL_OOP_TABLE_SLOTS; i++, oop++) {
+  for (oop = allOopsTable; oop < &oopTable[oopTableSize]; oop++) {
     addByteFlagToOOP(oop);
   }
-  for (i = 0, oop = oopTable; i < TOTAL_OOP_TABLE_SLOTS; i++, oop++) {
-    fixupOOP(i, oop);
+  for (oop = allOopsTable; oop < &oopTable[oopTableSize]; oop++) {
+    fixupOOP(oop);
   }
 
 #ifdef OOP_TABLE_TRACE
 printf("there are %d free oops out of %d oops, leaving %d\n",
-       numFreeOOPs, OOP_TABLE_SIZE, OOP_TABLE_SIZE - numFreeOOPs);
+       numFreeOOPs, oopTableSize, oopTableSize - numFreeOOPs);
 printf("max used is %d\n", maxUsedOOPSlot);
 #endif /* OOP_TABLE_TRACE */
 
@@ -390,8 +240,8 @@ printf("max used is %d\n", maxUsedOOPSlot);
 
 #ifdef actually_not_needed
 /**/  /* then save the constant ones at the end */
-/**/  fwrite(&oopTable[OOP_TABLE_SIZE], sizeof(struct OOPStruct),
-/**/	 TOTAL_OOP_TABLE_SLOTS - OOP_TABLE_SIZE, imageFile);
+/**/  fwrite(allOopsTable, sizeof(struct OOPStruct),
+/**/	 oopTable - allOopsTable, imageFile);
 #endif
 
   restoreAllOOPs();
@@ -399,7 +249,7 @@ printf("max used is %d\n", maxUsedOOPSlot);
 
 
 
-static void
+void
 addByteFlagToOOP(oop)
      OOP	oop;
 {
@@ -410,27 +260,23 @@ addByteFlagToOOP(oop)
   }
 }
 
-static void
-fixupOOP(i, oop)
-     int	i;
+void
+fixupOOP(oop)
      OOP	oop;
 {
   if (oopValid(oop)) {
-    if (i < OOP_TABLE_SIZE) {
-      maxUsedOOPSlot = i;
-    }
+    maxUsedOOPSlot = oopIndex(oop);
     oop->object = (mst_Object)toRelative(oop->object);
   }
 }
 
 
-static void
+void
 restoreAllOOPs()
 {
-  int		i;
   OOP		oop;
 
-  for (i = 0, oop = oopTable; i < TOTAL_OOP_TABLE_SLOTS; i++, oop++) {
+  for (oop = allOopsTable; oop < &oopTable[oopTableSize]; oop++) {
     if (oopValid(oop)) {
       oop->flags &= ~F_BYTE;
       oop->object = (mst_Object)fromRelative(oop->object);
@@ -438,29 +284,15 @@ restoreAllOOPs()
   }
 }
 
-static void
-saveGlobalOOPs(imageFile)
-     FILE	*imageFile;
-{
-  OOP		**oopPtr, oop;
-
-  for (oopPtr = globalOOPs; *oopPtr; oopPtr++) {
-    oop = toRelative(**oopPtr);
-    fwrite(&oop, sizeof(OOP), 1, imageFile);
-  }
-}
-
-
-static int
+int
 saveAllObjects(imageFile)
      FILE	*imageFile;
 {
   long		objectStart, objectEnd;
-  int		i;
   OOP		oop;
 
   objectStart = ftell(imageFile);
-  for (i = 0, oop = oopTable; i < OOP_TABLE_SIZE; i++, oop++) {
+  for (oop = oopTable; oop < &oopTable[oopTableSize]; oop++) {
     if (oopValid(oop)) {
       saveObject(imageFile, oop);
     }
@@ -470,7 +302,7 @@ saveAllObjects(imageFile)
 
 #ifdef actually_not_needed
 /**/  /* dump out the character objects, nil, true, and false */
-/**/  for (i = OOP_TABLE_SIZE; i < TOTAL_OOP_TABLE_SLOTS; i++) {
+/**/  for (i = FIRST_OOP_INDEX; i < 0; i++) {
 /**/    saveObject(imageFile, oopAt(i));
 /**/  }
 #endif
@@ -478,7 +310,7 @@ saveAllObjects(imageFile)
   return (objectEnd - objectStart);
 }
 
-static void
+void
 saveObject(imageFile, oop)
      FILE	*imageFile;
      OOP	oop;
@@ -489,78 +321,66 @@ saveObject(imageFile, oop)
   object = oopToObj(oop);
   numFixed = numOOPs(object);
 
-  fixupObject(oop, numFixed);
+  fixupObject(object, numFixed);
   fwrite(object, sizeof(OOP), object->objSize, imageFile);
-  restoreObject(oop, numFixed);
+  restoreObject(object, numFixed);
 }
 
 
 
-static void
-fixupObject(oop, numFixed)
-     OOP	oop;
+void
+fixupObject(obj, numFixed)
+     mst_Object	obj;
      int	numFixed;
 {
   OOP		instOOP, classOOP, *i;
 
-  classOOP = oopClass(oop);
-  oopToObj(oop)->objClass = toRelative(classOOP);
+  classOOP = obj->objClass;
+  obj->objClass = toRelative(classOOP);
 
-  i = oopToObj(oop)->data;
-  while(numFixed) {
+  for (i = obj->data; numFixed; i++, numFixed--) {
     instOOP = *i;
     if (isOOP(instOOP)) {
       *i = toRelative(instOOP);
     }
-    i++;
-    numFixed--;
   }
 }
 
-static void
-restoreObject(oop, numFixed)
-     OOP	oop;
+void
+restoreObject(object, numFixed)
+     mst_Object	object;
      int	numFixed;
 {
-  mst_Object	object;
-
-  object = oopToObj(oop);
   object->objClass = fromRelative(object->objClass);
-
-  restoreInstanceVars(oop, numFixed);
+  restoreInstanceVars(object, numFixed);
 }
 
-static void
-restoreInstanceVars(oop, numFixed)
-     OOP	oop;
+void
+restoreInstanceVars(obj, numFixed)
+     mst_Object	obj;
      int	numFixed;
 {
-  OOP		instOOP, classOOP, *i;
+  OOP		instOOP, *i;
 
-  classOOP = oopClass(oop);
-  i = oopToObj(oop)->data;
-  while(numFixed) {
+  i = obj->data;
+  while(numFixed--) {
     instOOP = *i;
     if (isOOP(instOOP)) {
       *i = fromRelative(instOOP);
     }
     i++;
-    numFixed--;
-  }
-  if (classOOP == cFuncDescriptorClass) {
-    restoreCFuncDescriptor(oop); /* in mstcint.c */
   }
 }
 
 
-static void
+void
 skipToHeader(imageFile)
      FILE	*imageFile;
 {
   rewind(imageFile);
 }
 
-static void
+void
 saveFileVersion(imageFile, objectDataSize)
      FILE	*imageFile;
      unsigned long objectDataSize;
@@ -572,7 +392,7 @@ saveFileVersion(imageFile, objectDataSize)
   header.version = VERSION_REQUIRED;
   header.objectDataSize = objectDataSize;
   header.oopTableSize = maxUsedOOPSlot + 1; /* n slots, numbered 0..n-1 */
-  header.semiSpaceSize = maxSpaceSize;
+  header.memSpaceSize = max(memSpace.totalSize, objectDataSize * 2);
 
   fwrite(&header, sizeof(SaveFileHeader), 1, imageFile);
 }
@@ -610,7 +430,7 @@ bad:
   if (wrongEndianness = flagChanged(header.flags, ENDIANNESS_FLAG)) {
     header.objectDataSize = byteInvert(header.objectDataSize);
     header.oopTableSize = byteInvert(header.oopTableSize);
-    header.semiSpaceSize = byteInvert(header.semiSpaceSize);
+    header.memSpaceSize = byteInvert(header.memSpaceSize);
     header.version = byteInvert(header.version);
   }
 
@@ -620,45 +440,39 @@ bad:
   /* check for version mismatch; if so this image file is invalid */
   if (header.version != VERSION_REQUIRED) goto bad;
   
-  if (header.semiSpaceSize > maxSpaceSize) {
-    growMemoryTo(header.semiSpaceSize);
+  if (header.memSpaceSize > memSpace.totalSize) {
+    growMemoryTo(header.memSpaceSize);
   }
 
-  initOOPTable();
+  initOOPTable(max(header.oopTableSize * 2, INITIAL_OOP_TABLE_SIZE));
   loadOOPTable(imageFile, header.oopTableSize);
 
 #ifdef OOP_TABLE_TRACE
-printf("After loading oopt table: %d\n", ftell(imageFile));
-#endif /* OOP_TABLE_TRACE */
-
-  loadGlobalOOPs(imageFile);
-
-#ifdef OOP_TABLE_TRACE
-printf("After loading global oop table: %d\n", ftell(imageFile));
+  printf("After loading oopt table: %d\n", ftell(imageFile));
 #endif /* OOP_TABLE_TRACE */
 
   loadNormalOOPs(imageFile, header.objectDataSize);
+
 #ifdef actually_not_needed
 /**/  loadCharOOPs(imageFile);
 /**/  loadSpecialOOPs(imageFile);
 #endif
 
 #ifdef OOP_TABLE_TRACE
-printf("After loading objects: %d\n", ftell(imageFile));
+  printf("After loading objects: %d\n", ftell(imageFile));
 #endif /* OOP_TABLE_TRACE */
 
   fixupInstanceVars();
-  initBuiltinObjectsClasses();
   refreshOOPFreeList();
+  if(!initDictionaryOnImageLoad()) goto bad;
 
   fclose(imageFile);
-
   setGCState(oldGCState);
 
   return (true);
 }
 
-static void
+void
 loadFileVersion(imageFile, headerp)
      FILE	*imageFile;
      SaveFileHeader *headerp;
@@ -666,12 +480,11 @@ loadFileVersion(imageFile, headerp)
   fread(headerp, sizeof(SaveFileHeader), 1, imageFile);
 }
 
-static void
+void
 loadOOPTable(imageFile, oldSlotsUsed)
      FILE	*imageFile;
      long 	oldSlotsUsed;
 {
-  long		i;
   OOP		oop;
 
   /* load in the valid OOP slots from previous dump */
@@ -679,76 +492,57 @@ loadOOPTable(imageFile, oldSlotsUsed)
   
   /* mark the remaining ones as available (not a for because of a probable
      bug in GNU C!) */
-  i = oldSlotsUsed;
-  oop = &oopTable[oldSlotsUsed];
-  while(i++ < OOP_TABLE_SIZE) {
+  for (oop = &oopTable[oldSlotsUsed]; oop < &oopTable[oopTableSize]; oop++) {
     oop->flags = F_FREE;
     oop++;
   }
 
 #ifdef actually_not_needed
 /**/  /* read in the constant stuff at the end */
-/**/  my_fread(&oopTable[OOP_TABLE_SIZE], sizeof(struct OOPStruct),
-/**/	TOTAL_OOP_TABLE_SLOTS - OOP_TABLE_SIZE, imageFile);
+/**/  my_fread(&allOopsTable, sizeof(struct OOPStruct),
+/**/	oopTable - allOopsTable, imageFile);
 #endif
 
   /* the fixup gets handled by load normal oops */
 }
 
 
-static void
-loadGlobalOOPs(imageFile)
-     FILE	*imageFile;
-{
-  OOP		**oopPtr, *oopVec;
-  long		numGlobalOOPs;
-
-  /* !!! a) this is clumsy, and b) this is incorrect -- if the number of
-   * global oops changes, I think we're hosed, and this code does nothing
-   * to detect or prevent this from happening. */
-  for (numGlobalOOPs = 0, oopPtr = globalOOPs; *oopPtr;
-       oopPtr++, numGlobalOOPs++);
-
-  /* using alloca, doing one fread, and then iterating through the array
-     speeds up the thing. */
-  oopVec = (OOP *)alloca(numGlobalOOPs * sizeof(OOP));
-  my_fread(oopVec, sizeof(OOP), numGlobalOOPs, imageFile);
-
-  for (oopPtr = globalOOPs; *oopPtr; oopPtr++, oopVec++) {
-    **oopPtr = fromRelative(*oopVec);
-  }
 #ifdef old_code /* Wed Apr 26 21:15:38 1989 */ /* <<<== what dedication...3 days before my wedding -- SBB */
+/**/mst_Boolean
+/**/loadGlobalOOPs(imageFile)
+/**/     FILE	*imageFile;
+/**/{
+/**/  OOP		**oopPtr, *oopVec;
+/**/  long		numGlobalOOPs;
+/**/
 /**/  for (oopPtr = globalOOPs; *oopPtr; oopPtr++) {
 /**/    my_fread(&oop, sizeof(OOP), 1, imageFile);
 /**/    **oopPtr = fromRelative(oop);
 /**/  }
+/**/}
 #endif /* old_code Wed Apr 26 21:15:38 1989 */
-}
 
-static void
+void
 loadNormalOOPs(imageFile, objectDataSize)
      FILE	*imageFile;
      unsigned long objectDataSize;
 {
-  register long	i;
-  OOP		oop, prevFreeOOP;
+  OOP		oop;
   mst_Object	object, objPtr;
 
   objPtr = curSpaceAddr();
 
-  prevFreeOOP = nil;
-
   /* Not everything must be byte-inverted in this part of the file,
-     so read everything through good old fread and fix later. */
+   * so read everything through good old fread and fix later. */
   fread(objPtr, 1, objectDataSize, imageFile);
 
   numFreeOOPs = 0;
-  for (i = 0, oop = oopTable; i < OOP_TABLE_SIZE; i++, oop++) {
+  for (oop = oopTable; oop < &oopTable[oopTableSize]; oop++) {
     if (oopValid(oop)) {
       object = objPtr;
       oop->object = object;
       if (wrongEndianness && !(oop->flags & F_BYTE)) {
-        fixupByteOrder(object, byteInvert(object->objSize));
+	fixupByteOrder(object, byteInvert(object->objSize));
       }
       object->objClass = fromRelative(object->objClass);
       objPtr = (mst_Object)((long *)object + object->objSize);
@@ -759,17 +553,19 @@ loadNormalOOPs(imageFile, objectDataSize)
   }
 
   /* numOOPs requires access to the instance spec in the class objects. So
-     we start by fixing the endianness of ALL the non-byte objects (including
-     classes!), for which we can do without numOOPs, then do another pass
-     and fix the byte objects using the now correct class objects. */
+   * we start by fixing the endianness of ALL the non-byte objects (including
+   * classes!), for which we can do without numOOPs, then do another pass
+   * here and fix the byte objects using the now correct class objects. */
   if (wrongEndianness) {
-    for (i = 0, oop = oopTable; i < OOP_TABLE_SIZE; i++, oop++) {
+    for (oop = oopTable; oop < &oopTable[oopTableSize]; oop++) {
       /* The cycle above ensures that free OOPs don't have F_BYTE set. */
       if (oop->flags & F_BYTE) {
-        object = oopToObj(oop);
-        fixupByteOrder(object, 2+numOOPs(object));
-        oop->flags &= ~F_BYTE;
+	object = oopToObj(oop);
+	fixupByteOrder(object, 2+numOOPs(object));
       }
+
+      /* Remove flags that get invalid when a session ends. */
+      oop->flags &= ~F_BYTE;
     }
   }
 
@@ -779,7 +575,7 @@ loadNormalOOPs(imageFile, objectDataSize)
 
 
 #ifdef actually_not_needed
-/**/static void loadCharOOPs(imageFile)
+/**/void loadCharOOPs(imageFile)
 /**/FILE	*imageFile;
 /**/{
 /**/  int		i;
@@ -796,7 +592,7 @@ loadNormalOOPs(imageFile, objectDataSize)
 /**/  }
 /**/}
 /**/
-/**/static void loadSpecialOOPs(imageFile)
+/**/void loadSpecialOOPs(imageFile)
 /**/FILE	*imageFile;
 /**/{
 /**/  fread(&nilObject, sizeof(struct NilObjectStruct), 1, imageFile);
@@ -814,20 +610,19 @@ loadNormalOOPs(imageFile, objectDataSize)
 /**/}
 #endif /* actually_not_needed */
 
-static void
+void
 fixupInstanceVars()
 {
-  int		i;
   OOP		oop;
 
-  for (i = 0, oop = oopTable; i < OOP_TABLE_SIZE; i++, oop++) {
+  for (oop = oopTable; oop < &oopTable[oopTableSize]; oop++) {
     if (oopValid(oop)) {
       fixupOOPInstanceVars(oop);
     }
   }
 }
 
-static void
+void
 fixupOOPInstanceVars(oop)
      OOP	oop;
 {
@@ -836,10 +631,10 @@ fixupOOPInstanceVars(oop)
 
   object = oopToObj(oop);
   numFixed = numOOPs(object);
-  restoreInstanceVars(oop, numFixed);
+  restoreInstanceVars(object, numFixed);
 }
 
-static void
+void
 fixupByteOrder(p, size)
      unsigned long int	*p;
      int		size;
@@ -849,7 +644,7 @@ fixupByteOrder(p, size)
   }
 }
 
-static int
+int
 my_fread(p, size, n, file)
      unsigned long int	*p;
      int		size;
