@@ -1,16 +1,16 @@
-/***********************************************************************
+/******************************** -*- C -*- ****************************
  *
  *	Public interface for binary save/restore module.
  *
- *	$Revision: 1.6.2$
- *	$Date: 1999/08/31 11:23:18$
+ *	$Revision: 1.8.5$
+ *	$Date: 2000/12/27 10:45:49$
  *	$Author: pb$
  *
  ***********************************************************************/
 
 /***********************************************************************
  *
- * Copyright 1990, 91, 92, 94, 95, 99 Free Software Foundation, Inc.
+ * Copyright 1988-92, 1994-95, 1999, 2000 Free Software Foundation, Inc.
  * Written by Steve Byrne.
  *
  * This file is part of GNU Smalltalk.
@@ -35,8 +35,12 @@
 #ifndef __GSTSAVE__
 #define __GSTSAVE__
 
-extern OOP			*globalOOPs[];
+#ifdef not_anymore
+/**/extern OOP			*globalOOPs[];
+#endif
 
 extern mst_Boolean		saveToFile(), loadFromFile();
 
+extern void			fixupObject(), restoreObject();
+
 #endif /* __GSTSAVE__ */
