@@ -1,9 +1,9 @@
-/***********************************************************************
+/******************************** -*- C -*- ****************************
  *
  *	Simple string support
  *
- *	$Revision: 1.6.2$
- *	$Date: 1999/08/31 11:23:18$
+ *	$Revision: 1.8.5$
+ *	$Date: 2000/12/27 10:45:49$
  *	$Author: pb$
  *
  ***********************************************************************/
@@ -11,7 +11,7 @@
 
 /***********************************************************************
  *
- * Copyright 1990, 91, 92, 94, 95, 99 Free Software Foundation, Inc.
+ * Copyright 1988-92, 1994-95, 1999, 2000 Free Software Foundation, Inc.
  * Written by Steve Byrne.
  *
  * This file is part of GNU Smalltalk.
@@ -36,9 +36,9 @@
 #ifndef __GSTSTR__
 #define _GSTSTR__
 
-extern char		*curStrBuf();
-extern void		initStrBuf();
-extern void		addStrBufChar();
-extern void		resizeString();
+extern char		*curStrBuf(), *obstack_curStrBuf();
+extern voidPtr		copyBuffer();
+extern long		bufferSize();
+extern void		addStrBufChar(), resetBuffer(), addBufPointer();
 
 #endif /* __GSTSTR__ */
