@@ -1,16 +1,16 @@
-/***********************************************************************
+/******************************** -*- C -*- ****************************
  *
  *	Symbol Table module.
  *
- *	$Revision: 1.6.2$
- *	$Date: 1999/08/31 11:23:18$
+ *	$Revision: 1.8.5$
+ *	$Date: 2000/12/27 10:45:49$
  *	$Author: pb$
  *
  ***********************************************************************/
 
 /***********************************************************************
  *
- * Copyright 1990, 91, 92, 94, 95, 99 Free Software Foundation, Inc.
+ * Copyright 1988-92, 1994-95, 1999, 2000 Free Software Foundation, Inc.
  * Written by Steve Byrne.
  *
  * This file is part of GNU Smalltalk.
@@ -40,6 +40,7 @@
 #include "dict.h"
 #include "tree.h"
 #include "lex.h"
+#include "sysdep.h"
 #include <stdio.h>
 #include <ctype.h>
 #if STDC_HEADERS
@@ -47,14 +48,6 @@
 #include <string.h>
 #endif /* STDC_HEADERS */
 
-#if !defined(max)
-#define	max(x, y) \
-  ( ((x) > (y)) ? (x) : (y) )
-#endif
-
-#define isSymbol(oop) \
-   ( !isNil(oop) && (oopClass(oop) ==  symbolClass) )
-
 typedef struct {
   OBJ_HEADER;
   OOP		nextLink;
@@ -69,7 +62,6 @@ struct SymbolListStruct {
   int		index;
   SymbolList	prevSymbol;
 };
-  
 
 /* Represents all the identifiers, both arguments and temporaries, which are
  * declared in a given scope.  Nested scopes result in nested instances of
@@ -88,7 +80,7 @@ struct ScopeStruct {
 
 OOP			andColonSymbol, atColonPutColonSymbol, atColonSymbol,
 			atEndSymbol, bitAndColonSymbol, bitOrColonSymbol,
-			bitShiftColonSymbol, blockCopyColonTemporariesColonSymbol, 
+			bitShiftColonSymbol, blockCopyColonSymbol, 
 			classSymbol, isNilSymbol, notNilSymbol,
 			divideSymbol, doColonSymbol, equalSymbol,
 			greaterEqualSymbol, greaterThanSymbol,
@@ -101,26 +93,25 @@ OOP			andColonSymbol, atColonPutColonSymbol, atColonSymbol,
 			remainderSymbol, sameObjectSymbol, sizeSymbol,
 			thisContextSymbol, finalizeSymbol, timesSymbol, 
 			toColonDoColonSymbol, timesRepeatColonSymbol,
-			toColonByColonDoColonSymbol,
+			toColonByColonDoColonSymbol, uppercaseSuperSymbol,
 			valueColonSymbol, valueColonValueColonSymbol,
 			valueColonValueColonValueColonSymbol,
- 			valueWithArgumentsColonSymbol,
-			valueSymbol, yourselfSymbol,
-			whileFalseColonSymbol, whileTrueColonSymbol, orSymbol,
-			andSymbol, superSymbol, nilSymbol, trueSymbol,
-			falseSymbol, selfSymbol,
+ 			valueWithArgumentsColonSymbol, whileFalseSymbol, 
+			valueSymbol, yourselfSymbol, whileTrueSymbol,
+			whileFalseColonSymbol, whileTrueColonSymbol,
+			superSymbol, nilSymbol, trueSymbol,
+			falseSymbol, selfSymbol, repeatSymbol,
 			doesNotUnderstandColonSymbol, unknownSymbol,
 			charSymbol, stringSymbol, stringOutSymbol, 
 			symbolSymbol, intSymbol, longSymbol, floatSymbol, 
-			doubleSymbol, booleanSymbol,
+			doubleSymbol, booleanSymbol, undeclaredSymbol,
 			voidSymbol, variadicSymbol, cObjectSymbol,
 			cObjectPtrSymbol,  smalltalkSymbol, symbolTable,
-			byteArraySymbol, byteArrayOutSymbol;
+			byteArraySymbol, byteArrayOutSymbol, terminateSymbol,
+			
+			currentNamespace;
 
-#ifdef unused
-/**/int			adds = 0, reused = 0, reprobes = 0,
-/**/			hitsOn[OOP_TABLE_SIZE];
-#endif /* unused */
+long			useUndeclared;
 
 void 			printString();
 
@@ -132,6 +123,89 @@ static OOP		scanName(), internCountedString();
 static int		instanceVariableIndex(), localVarIndex();
 
 static Scope		curScope = nil;
+
+SymbolInfo		symbolInfo[] = {
+  { &andColonSymbol,				"and:" },
+  { &atColonPutColonSymbol,			"at:put:" },
+  { &atColonSymbol,				"at:" },
+  { &atEndSymbol,				"atEnd" },
+  { &bitAndColonSymbol,				"bitAnd:" },
+  { &bitOrColonSymbol,				"bitOr:" },
+  { &bitShiftColonSymbol,			"bitShift:" },
+  { &blockCopyColonSymbol,			"blockCopy:" },
+  { &byteArraySymbol,				"byteArray" },
+  { &byteArrayOutSymbol,			"byteArrayOut" },
+  { &booleanSymbol,				"boolean" },
+  { &charSymbol,				"char" },
+  { &classSymbol,				"class" },
+  { &cObjectSymbol,				"cObject" },
+  { &cObjectPtrSymbol,				"cObjectPtr" },
+  { &divideSymbol,				"/" },
+  { &doColonSymbol,				"do:" },
+  { &doesNotUnderstandColonSymbol,		"doesNotUnderstand:" },
+  { &floatSymbol,				"float" },
+  { &doubleSymbol,				"double" },
+  { &equalSymbol,				"=" },
+  { &falseSymbol,				"false" },
+  { &finalizeSymbol,				"finalize" },
+  { &greaterEqualSymbol,			">=" },
+  { &greaterThanSymbol,				">" },
+  { &ifFalseColonIfTrueColonSymbol,		"ifFalse:ifTrue:" },
+  { &ifFalseColonSymbol,			"ifFalse:" },
+  { &ifTrueColonIfFalseColonSymbol,		"ifTrue:ifFalse:" },
+  { &ifTrueColonSymbol,				"ifTrue:" },
+  { &integerDivideSymbol,			"//" },
+  { &intSymbol,					"int" },
+  { &isNilSymbol,				"isNil" },
+  { &lessEqualSymbol,				"<=" },
+  { &lessThanSymbol,				"<" },
+  { &longSymbol,				"long" },
+  { &minusSymbol,				"-" },
+  { &newColonSymbol,				"new:" },
+  { &newSymbol,					"new" },
+  { &nextPutColonSymbol,			"nextPut:" },
+  { &nextSymbol,				"next" },
+  { &nilSymbol,					"nil" },
+  { &notEqualSymbol,				"~=" },
+  { &notNilSymbol,				"notNil" },
+  { &notSameObjectSymbol,			"~~" },
+  { &orColonSymbol,				"or:" },
+  { &plusSymbol,				"+" },
+  { &remainderSymbol,				"\\" },
+  { &repeatSymbol,				"repeat" },
+  { &sameObjectSymbol,				"==" },
+  { &selfSymbol,				"self" },
+  { &sizeSymbol,				"size" },
+  { &smalltalkSymbol,				"smalltalk" },
+  { &stringOutSymbol,				"stringOut" },
+  { &stringSymbol,				"string" },
+  { &superSymbol,				"super" },
+  { &symbolSymbol,				"symbol" },
+  { &terminateSymbol,				"__terminate" },
+  { &thisContextSymbol,				"thisContext" },
+  { &timesSymbol,				"*" },
+  { &timesRepeatColonSymbol,			"timesRepeat:" },
+  { &toColonByColonDoColonSymbol,		"to:by:do:" },
+  { &toColonDoColonSymbol,			"to:do:" },
+  { &trueSymbol,				"true" },
+  { &undeclaredSymbol,				"Undeclared" },
+  { &unknownSymbol,				"unknown" },
+  { &uppercaseSuperSymbol,			"Super" },
+  { &valueColonSymbol,				"value:" },
+  { &valueColonValueColonSymbol,		"value:value:" },
+  { &valueColonValueColonValueColonSymbol,	"value:value:value:" },
+  { &valueSymbol,				"value" },
+  { &valueWithArgumentsColonSymbol,		"valueWithArguments:" },
+  { &variadicSymbol,				"variadic" },
+  { &voidSymbol,				"void" },
+  { &whileFalseColonSymbol,			"whileFalse:" },
+  { &whileFalseSymbol,				"whileFalse" },
+  { &whileTrueColonSymbol,			"whileTrue:" },
+  { &whileTrueSymbol,				"whileTrue" },
+  { &yourselfSymbol,				"yourself" },
+  { nil }
+};
+
 
 
 
@@ -166,6 +240,7 @@ popOldScope()
 
   oldScope = curScope;
   curScope = curScope->prevScope;
+
   freeScopeSymbols(oldScope);
   xfree(oldScope);
 }
@@ -264,33 +339,51 @@ OOP
 findClassVariable(varName)
      OOP	varName;
 {
-  OOP		classOOP, assocOOP, classVariableOOP;
+  OOP		classOOP, assocOOP, poolDictionaryOOP;
+  OOP		myClass;
+  int		numPools, i;
+  Class		class;
 
-  if (oopClass(thisClass) == behaviorClass
-      || oopClass(thisClass) == classDescriptionClass) {
-    /* classDescriptions and above don't have class or pool variables */
-    /* ### this isn't quite the right test; we probably should be testing
-       for if we have a classClass or some derivative of that */
-    return (nilOOP);
+  myClass = thisClass;
+  if (oopClass(myClass) == metaclassClass) {
+    /* pretend that metaclasses have the class variables and shared
+     * pools that their instance classes do */
+    myClass = metaclassInstance(myClass);
+  }
+  while (oopClass(myClass) == behaviorClass
+      || oopClass(myClass) == classDescriptionClass) {
+    /* pretend that lightweight classes have the class variables
+     * and shared pools that their superclasses have */
+    myClass = superClass(myClass);
   }
 
-  for (classOOP = thisClass; !isNil(classOOP);
-       classOOP = superClass(classOOP)) {
-    if (oopClass(classOOP) == metaclassClass) {
-      /* pretend that metaclasses have the class variables and shared
-	 pools that their instance classes do */
-      classVariableOOP = metaclassInstance(classOOP);
-    } else {
-      classVariableOOP = classOOP;
-    }
-    assocOOP =
-      dictionaryAssociationAt(classVariableDictionary(classVariableOOP),
-			      varName);
+  /* Now search in the class pools */
+  for (classOOP = myClass; !isNil(classOOP); classOOP = superClass(classOOP)) {
+    assocOOP = dictionaryAssociationAt(
+      classVariableDictionary(classOOP), varName);
+
     if (!isNil(assocOOP)) {
       return (assocOOP);
     }
+  }
 
-    assocOOP = findSharedPoolVariable(classVariableOOP, varName);
+  /* and in the shared pools */
+  for (classOOP = myClass; !isNil(classOOP); classOOP = superClass(classOOP)) {
+    class = (Class) oopToObj(classOOP);
+    numPools = numOOPs(oopToObj(class->sharedPools));
+    for (i = 0; i < numPools; i++) {
+      poolDictionaryOOP = arrayAt(class->sharedPools, i+1);
+      assocOOP = findSharedPoolVariable(poolDictionaryOOP, varName);
+      if (!isNil(assocOOP)) {
+        return (assocOOP);
+      }
+    }
+  }
+
+  /* Now search in the `environments' */
+  for (classOOP = myClass; !isNil(classOOP); classOOP = superClass(classOOP)) {
+    class = (Class) oopToObj(classOOP);
+    assocOOP = findSharedPoolVariable(class->environment, varName);
     if (!isNil(assocOOP)) {
       return (assocOOP);
     }
@@ -301,17 +394,15 @@ findClassVariable(varName)
 
 
 SymbolEntry
-findVariable(varName, mustStore)
-     char		*varName;
+findVariable(symbol, mustStore)
+     OOP		symbol;
      mst_Boolean	mustStore;
 {
-  OOP		varAssoc, symbol;
+  OOP		varAssoc, undeclaredDictionary;
   int		index;
   unsigned int	scopeDistance;
   Scope		scope;
 
-  symbol = internString(varName);
-
   for (scope = curScope, scopeDistance = 0; scope != nil;
        scope = scope->prevScope, scopeDistance++) {
     index = localVarIndex(scope, symbol, mustStore);
@@ -326,8 +417,21 @@ findVariable(varName, mustStore)
   }
 
   varAssoc = findClassVariable(symbol);
-  if (isNil(varAssoc)) {
-    return (nil);
+  if(isNil(varAssoc)) {
+    Byte	*varName;
+    
+    varName = stringOOPChars(symbol);
+    if (!isupper(*varName) || !useUndeclared) {
+      return (nil);
+    }
+
+    undeclaredDictionary =
+      dictionaryAt(smalltalkDictionary, undeclaredSymbol);
+
+    varAssoc = dictionaryAssociationAt(undeclaredDictionary, symbol);
+    if (isNil(varAssoc)) {
+      varAssoc = dictionaryAtPut(undeclaredDictionary, symbol, nilOOP);
+    }
   }
 
   index = addForcedObject(varAssoc);
@@ -690,7 +794,7 @@ internCountedString(str, len)
   OOP		symbolOOP, linkOOP;
   IncPtr	incPtr;
 
-  index = (hashString(str, len) % INITIAL_SYMBOL_TABLE_SIZE) + 1;
+  index = (hashString(str, len) % SYMBOL_TABLE_SIZE) + 1;
   for (linkOOP = arrayAt(symbolTable, index); !isNil(linkOOP);
        linkOOP = link->nextLink) {
     link = (SymLink)oopToObj(linkOOP);
@@ -748,8 +852,8 @@ hashString(str, len)
      char	*str;
      int	len;
 {
-  unsigned long hashVal = 0;
-  mst_Boolean multiply = false;
+  unsigned long hashVal = 1;
+  mst_Boolean multiply = true;
 
   for ( ; len > 0; multiply ^= 1, str++, len--) {
     if (multiply) {
@@ -834,7 +938,7 @@ checkSymbolChain()
 {
   unsigned long	i;
 
-  for (i = 1; i <= INITIAL_SYMBOL_TABLE_SIZE; i++) {
+  for (i = 1; i <= SYMBOL_TABLE_SIZE; i++) {
     SymLink	link;
     OOP		linkOOP;
     for (linkOOP = arrayAt(symbolTable, i); !isNil(linkOOP);
@@ -878,7 +982,7 @@ printSymbols()
 {
   unsigned long	i;
   
-  for (i = 1; i <= INITIAL_SYMBOL_TABLE_SIZE; i++) {
+  for (i = 1; i <= SYMBOL_TABLE_SIZE; i++) {
     printf("Table[%ld]:\n", i);
     printSymbolChain(i);
     printf("\n");
@@ -897,84 +1001,25 @@ printSymLink(symLinkOOP)
   printObject(link->symbol);
 }
 
-
-
-void
-initSymbols()
+int
+selectorNumArgs(symbolOOP)
+     OOP	symbolOOP;
 {
-  andColonSymbol 	        = internString("and:");
-  andSymbol			= internString("&");
-  atColonPutColonSymbol         = internString("at:put:");
-  atColonSymbol 	        = internString("at:");
-  atEndSymbol 		        = internString("atEnd");
-  bitAndColonSymbol 	        = internString("bitAnd:");
-  bitOrColonSymbol 	        = internString("bitOr:");
-  bitShiftColonSymbol 	        = internString("bitShift:");
-  blockCopyColonTemporariesColonSymbol
-		 	        = internString("blockCopy:temporaries:");
-  byteArraySymbol		= internString("byteArray");
-  byteArrayOutSymbol		= internString("byteArrayOut");
-  booleanSymbol			= internString("boolean");
-  charSymbol                    = internString("char");
-  classSymbol 		        = internString("class");
-  cObjectSymbol                 = internString("cObject");
-  cObjectPtrSymbol              = internString("cObjectPtr");
-  divideSymbol 		        = internString("/");
-  doColonSymbol 	        = internString("do:");
-  doesNotUnderstandColonSymbol  = internString("doesNotUnderstand:");
-  floatSymbol			= internString("float");
-  doubleSymbol                  = internString("double");
-  equalSymbol 		        = internString("=");
-  falseSymbol			= internString("false");
-  finalizeSymbol		= internString("finalize");
-  greaterEqualSymbol 	        = internString(">=");
-  greaterThanSymbol 	        = internString(">");
-  ifFalseColonIfTrueColonSymbol = internString("ifFalse:ifTrue:");
-  ifFalseColonSymbol 	        = internString("ifFalse:");
-  ifTrueColonIfFalseColonSymbol = internString("ifTrue:ifFalse:");
-  ifTrueColonSymbol 	        = internString("ifTrue:");
-  integerDivideSymbol 	        = internString("//");
-  intSymbol                     = internString("int");
-  isNilSymbol 	                = internString("isNil");
-  lessEqualSymbol 	        = internString("<=");
-  lessThanSymbol 	        = internString("<");
-  longSymbol                    = internString("long");
-  minusSymbol 		        = internString("-");
-  newColonSymbol 	        = internString("new:");
-  newSymbol 		        = internString("new");
-  nextPutColonSymbol 	        = internString("nextPut:");
-  nextSymbol 		        = internString("next");
-  nilSymbol			= internString("nil");
-  notEqualSymbol 	        = internString("~=");
-  notNilSymbol 	                = internString("notNil");
-  notSameObjectSymbol		= internString("~~");
-  orColonSymbol 	        = internString("or:");
-  orSymbol			= internString("|");
-  plusSymbol 		        = internString("+");
-  remainderSymbol 	        = internString("\\");
-  sameObjectSymbol 	        = internString("==");
-  selfSymbol			= internString("self");
-  sizeSymbol 		        = internString("size");
-  smalltalkSymbol		= internString("smalltalk");
-  stringOutSymbol               = internString("stringOut");
-  stringSymbol                  = internString("string");
-  superSymbol			= internString("super");
-  symbolSymbol                  = internString("symbol");
-  thisContextSymbol		= internString("thisContext");
-  timesSymbol 		        = internString("*");
-  timesRepeatColonSymbol	= internString("timesRepeat:");
-  toColonByColonDoColonSymbol	= internString("to:by:do:");
-  toColonDoColonSymbol		= internString("to:do:");
-  trueSymbol			= internString("true");
-  unknownSymbol                 = internString("unknown");
-  valueColonSymbol 	        = internString("value:");
-  valueColonValueColonSymbol	= internString("value:value:");
-  valueColonValueColonValueColonSymbol = internString("value:value:value:");
-  valueSymbol 		        = internString("value");
-  valueWithArgumentsColonSymbol = internString("valueWithArguments:");
-  variadicSymbol                = internString("variadic");
-  voidSymbol                    = internString("void");
-  whileFalseColonSymbol	        = internString("whileFalse:");
-  whileTrueColonSymbol 	        = internString("whileTrue:");
-  yourselfSymbol 	        = internString("yourself");
-}
+  Byte *bytes;
+  int   numArgs, len;
+
+  len = stringOOPLen(symbolOOP);
+  bytes = (Byte *) (oopToObj(symbolOOP)->data);
+  if (bytes[0] < 'A' || bytes[0] > 'z') {
+    return (1);
+  }
+  if (bytes[0] > 'Z' && bytes[0] < 'a') {
+    return (1);
+  }
+  for (numArgs = 0; len; ) {
+    if (bytes[--len] == ':') {
+      numArgs++;
+    }
+  }
+  return (numArgs);
+}
\ No newline at end of file
