@@ -1,20 +1,20 @@
-/***********************************************************************
+/******************************** -*- C -*- ****************************
  *
  * System specific implementation module.
  *
  * This module contains implementations of various operating system
- * specific routines.  This module should encapsulate most of these OS
- * specific calls so that the rest of the code is portable.
+ * specific routines.  This module should encapsulate most (or all) of
+ * these calls so that the rest of the code is portable.
  *
- * $Revision: 1.6.2$
- * $Date: 1999/08/31 11:23:18$
+ * $Revision: 1.8.5$
+ * $Date: 2000/12/27 10:45:49$
  * $Author: pb$
  *
  ***********************************************************************/
 
 /***********************************************************************
  *
- * Copyright 1990, 91, 92, 94, 95, 99 Free Software Foundation, Inc.
+ * Copyright 1988-92, 1994-95, 1999, 2000 Free Software Foundation, Inc.
  * Written by Steve Byrne.
  *
  * This file is part of GNU Smalltalk.
@@ -40,10 +40,6 @@
 #include "sysdep.h"
 #include "alloc.h"
 
-#ifdef sgi
-# define _BSD_SIGNALS
-#endif
-
 #include <errno.h>
 #include <stdio.h>
 #include <signal.h>
@@ -55,6 +51,10 @@
 # include <unistd.h>
 #endif 
 
+#ifdef HAVE_GETRUSAGE
+# include <sys/resource.h>
+#endif
+
 #if defined(HAVE_SYS_TIME_H) || defined(TIME_WITH_SYS_TIME)
 # include <sys/time.h>
 #endif
@@ -86,7 +86,7 @@
 #if !defined(__OS2__) && !defined(_MSC_VER)
 #  include <sys/param.h>
 #endif
-#ifdef _WIN32
+#ifdef WIN32
 #  include <windows.h>
 #endif
 
@@ -111,17 +111,16 @@
 #endif
 
 #ifndef SIGALRM
-#define 0		/* hmmm... */
-#else
-#define HAVE_SIGALRM
-#endif
-
-#if defined(HAVE_SIGHOLD) && !defined(HAVE_SIGSETMASK)
-static unsigned long __cursigmask; /* keep track of signal mask status */
+#define SIGALRM		SIGUSR1		/* hmmm... */
 #endif
 
 static int	tm_diff ();
 
+#undef INTERRUPTS_DONE
+#ifdef SIG_BLOCK
+#define INTERRUPTS_DONE
+static int disabled = false;
+
 /*
  * IntState disableInterrupts()
  *
@@ -138,25 +137,19 @@ static int	tm_diff ();
 IntState
 disableInterrupts()
 {
-#ifdef HAVE_SIGSETMASK
-  return (sigsetmask(-1));
-#else
-#ifdef HAVE_SIGHOLD
-  unsigned long oldmask = __cursigmask;
-  register int i;
+  sigset_t	  newSet;
+  static sigset_t oldSet;
 
-  __cursigmask = -1;
-  for (i=1; i <= 32; i++) {
-    sighold(i);	 /* want it blocked - ok if it already is */
+  if (disabled) {
+    /* let only the outermost calls actually block and unblock signals */
+    return ((IntState) NULL);
   }
-  return oldmask;
-#else
-  return 0;   /* no recognized interrupt mechanism */
-#endif
-#endif
+  disabled = true;
+  sigfillset(&newSet);
+  sigprocmask(SIG_BLOCK, &newSet, &oldSet);
+  return ((IntState) &oldSet);
 }
 
-
 /*
  * void enableInterrupts(mask)
  *
@@ -172,58 +165,107 @@ disableInterrupts()
  *
  */
 void
-enableInterrupts(mask)
-     IntState mask;
+enableInterrupts(newSet)
+     IntState newSet;
 {
-#ifdef HAVE_SIGSETMASK
-  sigsetmask(mask);
-
-#else
-#ifdef HAVE_SIGHOLD  /* !!! have config check for sigrelse too? */
-  unsigned long oldmask = __cursigmask;
-  register int i;
-
-  __cursigmask = mask;
-  for (i=1; mask != 0; i++, mask >>= 1) {
-    if (oldmask & (0x1 << (i-1))) {
-      sigrelse(i); /* want it unblocked and it is blocked */
-    }
+  if (newSet) {
+    disabled = false;
+    sigprocmask(SIG_SETMASK, (sigset_t *) newSet, NULL);
   }
-#else
-  ;    /* do nothing -- we don't know what to do */
-#endif
+}
 #endif
+
+#if defined(HAVE_SIGSETMASK) && !defined(INTERRUPTS_DONE) /* BSD style */
+#define INTERRUPTS_DONE
+IntState
+disableInterrupts()
+{
+  return (sigsetmask(-1));
 }
 
 void
+enableInterrupts(mask)
+     IntState mask;
+{
+  sigsetmask(mask);
+}
+#endif
+
+#if defined(HAVE_SIGHOLD) && !defined(INTERRUPTS_DONE) /* SVID style */
+#define INTERRUPTS_DONE
+static mst_Boolean disabled;
+
+IntState
+disableInterrupts()
+{
+  register int i;
+
+  if (disabled) {
+    /* let only the outermost calls actually block and unblock signals */
+    return (0);
+  }
+
+  disabled = true;
+  for (i=1; i <= 8 * SIZEOF_LONG; i++) {
+    sighold(i);	 /* want it blocked - ok if it already is */
+  }
+  return (1);
+}
+
+void
+enableInterrupts(mask)
+     IntState mask;
+{
+  register int i;
+
+  if (!mask) {
+    return;
+  }
+
+  disabled = false;
+  for (i=1; i <= 8 * SIZEOF_LONG; i++) {
+    sigrelse(i); /* want it unblocked if it used to be */
+  }
+}
+#endif
+
+#ifndef INTERRUPTS_DONE
+IntState
+disableInterrupts()
+{
+  return ((IntState) 0);   /* no recognized interrupt mechanism */
+}
+
+void
+enableInterrupts(mask)
+     IntState mask;
+{
+  /* no way dude */
+}
+#endif
+
+SigHandler
 setSignalHandler(signum, handlerFunc)
      int signum;
-     RETSIGTYPE (*handlerFunc)(SIG_ARG_TYPE);
+     SigHandler handlerFunc;
 {
 #ifdef _POSIX_VERSION
   /* If we are running on a posix-compliant system, then do
-     things the Posix way. */
-  struct sigaction act;
+   * things the Posix way. */
+  struct sigaction act, o_act;
 
   act.sa_handler = handlerFunc;
-#ifdef SA_ONESHOT
-  act.sa_flags = SA_ONESHOT;
-#else
-#ifdef SA_RESETHAND
-  act.sa_flags = SA_RESETHAND;
-#else
   act.sa_flags = 0;
-#endif
-#endif
 
   sigemptyset (&act.sa_mask);
-  sigaction (signum, &act, (struct sigaction *)NULL);
+  sigaction (signum, &act, &o_act);
+  return o_act.sa_handler;
+
 #else
-  signal(signum, handlerFunc);
+  return signal(signum, handlerFunc);
 #endif
 }
 
-
 
 
 /*
@@ -231,39 +273,82 @@ setSignalHandler(signum, handlerFunc)
  *
  * Description
  *
- * Returns the local time in milliseconds
+ * Returns the local time since midnight in milliseconds
  *
  */
 unsigned long
 getMilliTime()
 {
-#ifdef _WIN32
-  /* time() seems not to work... so we hack. */
-  SYSTEMTIME st;
-  GetLocalTime(&st);
-  return (st.wMilliseconds +
-	  st.wSecond * 1000 +
-	  st.wMinute * 60000 +
-	  st.wHour   * 3600000);
-#else
-#ifdef HAVE_GETTIMEOFDAY
+#undef MILLI_DONE
+#ifdef WIN32
+#define MILLI_DONE
+  /* time() seems not to work... so we hack. This method to 
+   * obtain the time is complex, but it is the most precise. */
+  static long		frequency = 0, frequencyH, adjust = 0;
+  long			milli;
+  LARGE_INTEGER		counter;
+  SYSTEMTIME		st;
+
+  if (!frequency) {
+    if (QueryPerformanceFrequency (&counter)) {
+      /* frequencyH = 1000 * 2^32 / frequency */
+      frequency = counter.HighPart ? -1 : counter.LowPart;
+      frequencyH = MulDiv (1000 * (1 << 16), (1 << 16), frequency);
+    } else {
+      frequency = -1;
+    }
+  }
+
+  if (frequency == -1) {
+    /* QueryPerformanceCounter not supported, always use GetLocalTime */
+    adjust = milli = 0;
+  } else {
+    QueryPerformanceCounter (&counter);
+    /* milli = (high * 2^32 + low) * 1000 / freq
+     *       = high * (1000 * 2^32 / freq) + (low * 1000 / freq)
+     *       = (high * frequencyH) + (low / 4) * 4000 / freq)
+     *
+     * Dividing and multiplying counter.LowPart by 4 is needed because
+     * MulDiv accepts signed integers but counter.LowPart is unsigned. */
+    milli = counter.HighPart * frequencyH;
+    milli += MulDiv(counter.LowPart >> 2, 4000, frequency);
+  }
+
+  if (!adjust) {
+    GetLocalTime(&st);
+    adjust = st.wMilliseconds;
+    adjust += st.wSecond * 1000;
+    adjust += st.wMinute * 60000;
+    adjust += st.wHour   * 3600000;
+    adjust -= milli;
+    milli += adjust;
+  } else {
+    milli += adjust;
+    while (milli > 86400000) {
+      milli -= 86400000;
+      adjust -= 86400000;
+    }
+  }
+  return (milli);
+#endif
+
+#if defined(HAVE_GETTIMEOFDAY) && !defined(MILLI_DONE) /* BSD style */
+#define MILLI_DONE
   struct timeval t;
 
   gettimeofday(&t, nil);
   return (t.tv_sec * 1000 + t.tv_usec / 1000);
-#else
-#ifdef __OS2__
-  time_t t;
-  t = time(NULL);
-  return (t * 1000 / CLOCKS_PER_SEC);
-#else
-  time_t t;
-  struct tms dummy;
+#endif
 
+#ifndef MILLI_DONE
+  clock_t t;
+#ifdef _POSIX_VERSION
+  struct tms dummy;
   t = times(&dummy);
-  return (t * 1000 / HZ);
-#endif
-#endif
+#else /* !_POSIX_VERSION */
+  t = clock();
+#endif /* !_POSIX_VERSION */
+  return ( ((double)t) * 1000 / CLOCKS_PER_SEC);
 #endif
 }
 
@@ -276,8 +361,8 @@ tm_diff (a, b)
      struct tm *a, *b;
 {
   /* Compute intervening leap days correctly even if year is negative.
-     Take care to avoid int overflow in leap day calculations,
-     but it's OK to assume that A and B are close to each other.  */
+   * Take care to avoid int overflow in leap day calculations,
+   * but it's OK to assume that A and B are close to each other.  */
   int a4 = (a->tm_year >> 2) + (TM_YEAR_BASE >> 2) - ! (a->tm_year & 3);
   int b4 = (b->tm_year >> 2) + (TM_YEAR_BASE >> 2) - ! (b->tm_year & 3);
   int a100 = a4 / 25 - (a4 % 25 < 0);
@@ -294,7 +379,7 @@ tm_diff (a, b)
 }
 
 /*
- * long adjustTimeZone(long)
+ * long adjustTimeZone(time_t)
  *
  * Description
  *
@@ -307,7 +392,7 @@ tm_diff (a, b)
  */
 long
 adjustTimeZone(t)
-     long t;
+     time_t t;
 {
   struct tm save_tm, *decoded_time;
   long bias;
@@ -320,17 +405,17 @@ adjustTimeZone(t)
   decoded_time = gmtime (&t);
   bias = tm_diff (&save_tm, decoded_time);
 
-  return(t + bias);
+  return(((long) t) + bias);
 }
 
 long
 currentTimeZoneBias()
 {
-  long now;
+  time_t now;
   long bias;
   struct tm save_tm, *decoded_time;
 
-  time((time_t *) &now);
+  time(&now);
 
 #ifdef LOCALTIME_CACHE
   tzset ();
@@ -343,6 +428,46 @@ currentTimeZoneBias()
   return(bias);
 }
 
+/*
+ * char *currentTimeZoneName()
+ *
+ * Description
+ *
+ * Returns the name of the current time zone.
+ *
+ * Outputs
+ *
+ * As described above, or XXX if the time zone's name could not be found.
+ */
+char *
+currentTimeZoneName()
+{
+  char *zone;
+  zone = getenv("TZ");
+  if (!zone) {
+    long bias = currentTimeZoneBias() / 60;
+#ifdef _WIN32
+    TIME_ZONE_INFORMATION tzi;
+    LPCWSTR name;
+    static char buffer[32];
+    GetTimeZoneInformation(&tzi);
+    zone = buffer;
+    name = (bias == (tzi.Bias + tzi.StandardBias))
+      ? tzi.StandardName : tzi.DaylightName;
+    
+    WideCharToMultiByte(CP_ACP, 0,
+      name, lstrlenW(name), zone, 32,
+      NULL, NULL);
+
+#else
+    /* Maybe we could guess it */
+    zone = "XXX";
+#endif
+  }
+
+  return strdup(zone);
+}
+
 /*
  * long getTime()
  *
@@ -361,7 +486,7 @@ getTime()
   time_t now;
   time(&now);
 
-  return (adjustTimeZone((unsigned long) now));
+  return (adjustTimeZone(now));
 }
 
 
@@ -372,10 +497,7 @@ getTime()
  * Description
  *
  * Set up func as a signal handler to be called after deltaMilli
- * milliseconds. The function should NOT expect anything about the
- * condition in which it is called (e.g. it should not expect to
- * be called from an interrupt handler, even though we pass SIGALRM
- * to it to fake this assumption some way).
+ * milliseconds.
  *
  * Inputs
  *
@@ -391,12 +513,11 @@ getTime()
  * None.
  */
 
-#ifdef _WIN32
+#ifdef WIN32
 struct {
   HANDLE	hNewWaitEvent;
   HANDLE	hCritEvent;
   long		sleepTime;
-  RETSIGTYPE	(*alarmFunc)();
 } alarms;
 
 /* thread for precise alarm callbacks */
@@ -407,11 +528,9 @@ alarmThread(unused)
   WaitForSingleObject(alarms.hNewWaitEvent, INFINITE);
   for(;;) {
     int sleepTime;
-    RETSIGTYPE	(*alarmFunc)();
 
     WaitForSingleObject(alarms.hCritEvent, INFINITE);
     sleepTime = alarms.sleepTime;
-    alarmFunc = alarms.alarmFunc;
     SetEvent(alarms.hCritEvent);
 
     if (sleepTime > 0) {
@@ -420,13 +539,13 @@ alarmThread(unused)
 	continue;
       }
     }
-    alarmFunc(SIGALRM);
+    kill(getpid(), SIGALRM);
     WaitForSingleObject(alarms.hNewWaitEvent, INFINITE);
   }
 }
 
 static void
-initializeWin32Alarms()
+initializeAlarms()
 {
   HANDLE hthread;
   DWORD tid;
@@ -446,6 +565,12 @@ initializeWin32Alarms()
   /* This does not terminate the thread - it only releases our handle */
   CloseHandle(hthread);
 }
+
+#else
+static void
+initializeAlarms()
+{
+}
 #endif
 
 void
@@ -456,42 +581,90 @@ signalAfter(deltaMilli, func)
 /* Please feel free to make this more accurate for your operating system
  * and send me the changes.
  */
+  setSignalHandler(SIGALRM, func);
+
   if (deltaMilli <= 0) {
-    func(SIGALRM);
+    kill(getpid(), SIGALRM);
   } else {
 
-#ifdef _WIN32
+#ifdef WIN32
+#define ALARM_DONE
     WaitForSingleObject(alarms.hCritEvent, INFINITE);
     alarms.sleepTime = deltaMilli;
-    alarms.alarmFunc = func;
     SetEvent(alarms.hCritEvent);
     SetEvent(alarms.hNewWaitEvent);
+#endif
 
-#else
- #ifdef ITIMER_REAL
+#if defined(ITIMER_REAL) && !defined(ALARM_DONE)
+#define ALARM_DONE
     struct itimerval value;
     value.it_interval.tv_sec = value.it_interval.tv_usec = 0;
     value.it_value.tv_sec = deltaMilli/1000;
     value.it_value.tv_usec = (deltaMilli%1000) * 1000;
-    setSignalHandler(SIGALRM, func);
     setitimer(ITIMER_REAL, &value, (struct itimerval *)0);
+#endif
 
- #else
-  #ifdef HAVE_SIGALRM
-    setSignalHandler(SIGALRM, func);
-  #endif
-  #ifdef HAVE_ALARM
-    /* round up to nearest second */
-    alarm((deltaMilli + 999)/ 1000);
-  #else
-    /* ugh: no support for alarms on this system!! */
-    func(SIGALRM);
-  #endif
- #endif
+#if defined(HAVE_FORK) && !defined(ALARM_DONE)
+#define ALARM_DONE
+    static pid_t pid = -1;
+    long end, ticks;
+    if (pid != -1) {
+      kill(pid, SIGTERM);
+    }
+    setSignalHandler(SIGCHLD, SIG_IGN);
+    switch(pid = fork()) {
+      case -1:
+	/* Error, try to recover */
+	kill(getpid(), SIGALRM);
+	break;
+
+      case 0:
+	/* Child process */
+	end = getMilliTime() + deltaMilli;
+	do {
+	  ticks = end - getMilliTime();
+	  if (ticks > 1100) {		/* +100 is arbitrary */
+	    sleep((int) (ticks / 1000))
+	  }
+	} while (ticks > 0);
+	kill (getppid(), SIGALRM)
+	_exit(0);
+    }
+#endif
+
+#if defined(HAVE_ALARM) && !defined(ALARM_DONE)
+#define ALARM_DONE
+    alarm(deltaMilli / 1000);
+#endif
+
+#if defined(ALARM_DONE)
+#undef ALARM_DONE
+#else
+    /* Cannot do anything better than this */
+    kill(getpid(), SIGALRM);
 #endif
   }
 }
 
+void
+relinquish(usecs)
+     int usecs;
+{
+/* Please feel free to make this more accurate for your operating system
+ * and send me the changes.
+ */
+#ifdef WIN32
+  Sleep((usecs + 999) / 1000);
+#else
+#ifdef HAVE_SELECT
+  struct timeval time;
+  time.tv_sec = usecs / 1000000;
+  time.tv_usec = usecs % 1000000;
+
+  select(0, NULL, NULL, NULL, &time);
+#endif
+#endif
+}
 
 
 /*
@@ -510,7 +683,7 @@ signalAfter(deltaMilli, func)
 char *
 getCurDirName()
 {
-#ifdef HAVE_UNISTD_H
+#ifdef HAVE_GETCWD
   char *cwd;
   char *ret;
   unsigned path_max;
@@ -540,14 +713,9 @@ getCurDirName()
   errno = save_errno;
   return (NULL);
 #else
- #ifndef HAVE_GETWD
-  #error "Must have either getcwd or getwd!!"
- #else
-  char	name[PATH_MAX];
-  
+  char	name[PATH_MAX+2];
   getwd(name);
   return (strdup(name));
- #endif
 #endif
 }
 
@@ -575,7 +743,6 @@ getFullFileName(fileName)
 {
   char	*fullFileName;
   static char *fullPath = NULL;
-  extern char *strdup();
 
   if (fileName[0] == '/') { /* absolute, so don't need to change */
     return (strdup(fileName));
@@ -597,6 +764,7 @@ getFullFileName(fileName)
   sprintf(fullFileName, "%s/%s", fullPath, fileName);
   return (fullFileName);
 }
+
 
 
 /*
@@ -629,18 +797,19 @@ getFileModifyTime(fileName)
   if (stat(fileName, &st) < 0) {
     return ((unsigned long)0);
   } else {
-    return (adjustTimeZone((unsigned long) st.st_mtime));
+    return (adjustTimeZone(st.st_mtime));
   }
 }
 
 
 /*
  * mst_Boolean fileIsReadable(fileName)
+ * mst_Boolean fileIsWriteable(fileName)
  *
  * Description
  *
- * Returns true if the file named "fileName" exists and is readable by the
- * current process.  Returns false otherwise.
+ * Returns true if the file named "fileName" exists and is readable or
+ * writable by the current process.  Returns false otherwise.
  *
  * Inputs
  *
@@ -649,7 +818,7 @@ getFileModifyTime(fileName)
  *
  * Outputs
  *
- * True if the file exists and is readable, false otherwise.
+ * True if the file exists and is readable/writeable, false otherwise.
  */
 mst_Boolean
 fileIsReadable(fileName)
@@ -658,21 +827,14 @@ fileIsReadable(fileName)
   return (access(fileName, R_OK) == 0);
 }
 
-OOP
-getOpenFileSize(fd)
-     int fd;
+mst_Boolean
+fileIsWriteable(fileName)
+     char *fileName;
 {
-  struct stat statBuf;
-
-  if (fstat(fd, &statBuf) < 0) {
-    return nilOOP;
-  } else {
-    return fromInt(statBuf.st_size);
-  }
+  return (access(fileName, W_OK) == 0);
 }
 
-
-
+
 FILE *
 openPipe(command, mode)
      char *command, *mode;
@@ -710,7 +872,7 @@ openPipe(command, mode)
     if ((close(fildes[parent_pipe_end]) != 0)
 	|| (close(child_pipe_end) != 0)
 	|| (dup(fildes[child_pipe_end]) != child_pipe_end)) {
-      exit(1);
+      _exit(1);
     }
 
     /* Write a single byte through the pipe to establish it */
@@ -720,11 +882,11 @@ openPipe(command, mode)
       char buff[1];
       read(fildes[child_pipe_end],buff,1);
       if (*buff != 'z') {
-        exit(1);
+        _exit(1);
       }
     }
 
-    exit (system(command) >= 0);
+    _exit (system(command) >= 0);
     /*NOTREACHED*/
 
   default:
@@ -777,12 +939,15 @@ closePipe(file)
 #ifdef HAVE_POPEN
   return (pclose(file));
 #else
+#if defined(HAVE_PIPE) && defined(HAVE_WAITPID)
   int statloc;
-  close(fileno(file));
-  waitpid(-1, &statloc, 0);
-
   /* need to call fclose() to deallocate FILE buffer */
-  return (fclose(file));
+  close(fileno(file));
+  fclose(file);
+
+  while((waitpid((pid_t) -1, &statloc, 0) == -1) && (errno == EINTR)) ;
+  return (statloc);
+#endif
 #endif
 }
 
@@ -791,7 +956,7 @@ FILE *
 openFile(fileName, fileMode)
      char *fileName, *fileMode;
 {
-#if defined(_WIN32) || defined(__OS2__) || defined(MSDOS)
+#if defined(WIN32) || defined(__OS2__) || defined(MSDOS)
   char *fileModeBinary;
   fileModeBinary = alloca(strlen(fileMode) + 2);
   strcpy(fileModeBinary, fileMode);
@@ -803,11 +968,58 @@ openFile(fileName, fileMode)
 }
 
 
+#ifndef HAVE_GETPAGESIZE
+size_t
+getpagesize()
+{
+#ifdef _SC_PAGE_SIZE
+  return sysconf(_SC_PAGE_SIZE);
+#else
+#ifdef _SC_PAGESIZE
+  return sysconf(_SC_PAGESIZE);
+#else
+#ifdef	PAGESIZE
+  return PAGESIZE;
+#else	/* no PAGESIZE */
+#ifdef	EXEC_PAGESIZE
+  return EXEC_PAGESIZE;
+#else	/* no EXEC_PAGESIZE */
+#ifdef	NBPG
+#ifndef	CLSIZE
+#define	CLSIZE 1
+#endif	/* CLSIZE */
+  return NBPG * CLSIZE;
+#else	/* no NBPG */
+#ifdef	NBPC
+  return NBPC
+#else	/* no NBPC */
+  return 4096;			/* Just punt and use reasonable value */
+#endif /* no NBPC */
+#endif /* no NBPG */
+#endif /* no EXEC_PAGESIZE */
+#endif /* no PAGESIZE */
+#endif /* no _SC_PAGESIZE */
+#endif /* no _SC_PAGE_SIZE */
+}
+#endif
+
 #ifndef HAVE_GETDTABLESIZE
 int
 getdtablesize()
 {
-  return (sizeof(short) * 8);	 /* a safe >>>hack<<< value */
+  int tableSize = -1;
+
+#ifdef  _SC_OPEN_MAX            /* POSIX -- preferred */
+  tableSize = sysconf(_SC_OPEN_MAX);
+#else
+#ifdef RLIMIT_NOFILE           /* who is non POSIX but has this? */
+  if (getrlimit(RLIMIT_NOFILE, &rlimit) != -1) {
+    tableSize = rlimit.rlim_max;
+#endif
+#endif
+
+  /* If something went wrong, choose a safe >>>hack<<< value */
+  return (tableSize == -1) ? sizeof(short) * 8 : tableSize;
 }
 
 #endif /* ! HAVE_GETDTABLESIZE */
@@ -824,12 +1036,56 @@ getdtablesize()
 void
 initSysdep()
 {
-#ifdef _WIN32
-  initializeWin32Alarms();
-#endif
+  setSignalHandler (SIGCHLD, SIG_IGN);
+  initializeAlarms();
   tzset ();
 }
 
+void
+flushCode(dest, len)
+     char 	*dest;
+     int	len;
+{
+#ifdef __GNUC__
+#ifdef PPC
+  for (; len > 0; dest += SIZEOF_CHAR_P, len -= SIZEOF_CHAR_P) {
+    __asm__ __volatile__ ("dcbst 0,%0; sync; icbi 0,%0; isync"::"r"(dest));
+  }
+#endif
+
+#ifdef __sparc__
+  __asm__ __volatile__ ("stbar");
+  for (; len > 0; dest += SIZEOF_CHAR_P, len -= SIZEOF_CHAR_P) {
+    __asm__ __volatile__ ("flush %0"::"r"(dest));
+  }
+
+  /* [SPARC Architecture Manual v8, page 139, implementation note #5] */
+  __asm__ __volatile__ ("nop; nop; nop; nop; nop");
+
+#endif
+
+#ifdef __alpha__
+  __asm__ __volatile__ ("call_pal 0x86");
+#endif
+
+#ifdef __mc68000__
+  /* On the 68040 (and 060 I presume) we need to flush the instruction
+   * cache before running new code.  Stupidly, Motorola made cpush a
+   * privledged instruction, so we have to get operating system help
+   * for this.  Don't bother asking: there is no standard there. 
+   * 
+   * We are partially rescued by gcc, which must be taught how to do this
+   * in order to build bound-function trampolines.  This works at least
+   * for A/UX, Linux and NeXT.  The rest of you with 040's should
+   * consider fixing gcc, as you are broken there too.  
+   * 
+   * For those with 020's and 030's, this is effectively a no-op.  */
+
+  __clear_cache(dest, dest + len);
+#endif
+
+#endif /* __GNUC__ */
+}
 
 #ifdef HAVE_STDARG_H
 void debugf (const char *fmt, ...)
@@ -851,10 +1107,10 @@ debugf(va_alist)
 
   vsprintf (buf, fmt, args);
 
-#if defined(_WIN32) && !defined(__CYGWIN__)&& !defined(__CYGWIN32__)
-  /* VC++ has a Win32-friendly debugger, Cygwin has not */
+#if defined(WIN32) && !defined(__CYGWIN__)&& !defined(__CYGWIN32__)
+  /* VC++ has a Win32-friendly debugger, otherwise winning Cygwin has not */
   OutputDebugString (buf);
-#else /* !_WIN32 */
+#else /* !WIN32 */
   {
     static FILE *debFile = NULL;
     if (debFile == NULL) {
@@ -863,7 +1119,25 @@ debugf(va_alist)
     fputs(buf, debFile);
     fflush(debFile);
   }
-#endif /* !_WIN32 */
+#endif /* !WIN32 */
 
   va_end(args);
 }
+
+/*
+ *	void debug()
+ *
+ * Description
+ *
+ *	Used for debugging.  You set a breakpoint in the debug routine in the
+ *	debugger, and have code call it when you want it to stop.  Performs no
+ *	action normally.
+ *
+ */
+void
+debug()
+{
+/* kill(getpid(), SIGTRAP); */
+/* getchar(); */
+}
+
