@@ -1,16 +1,16 @@
-/***********************************************************************
+/******************************** -*- C -*- ****************************
  *
  *	Semantic Tree manipulation module.
  *
- *	$Revision: 1.6.2$
- *	$Date: 1999/08/31 11:23:18$
+ *	$Revision: 1.8.5$
+ *	$Date: 2000/12/27 10:45:49$
  *	$Author: pb$
  *
  ***********************************************************************/
 
 /***********************************************************************
  *
- * Copyright 1990, 91, 92, 94, 95, 99 Free Software Foundation, Inc.
+ * Copyright 1988-92, 1994-95, 1999, 2000 Free Software Foundation, Inc.
  * Written by Steve Byrne.
  *
  * This file is part of GNU Smalltalk.
@@ -34,8 +34,10 @@
 
 #include "gst.h"
 #include "alloc.h"
+#include "obstack.h"
 #include "sym.h"
 #include "tree.h"
+#include "dict.h"
 #include "lex.h"
 #include <stdio.h>
 
@@ -44,15 +46,18 @@ char	*nilName = "(nil)"; /* how to print nil */
 
 mst_Boolean hadError = false;
 
-static TreeNode makeMethodNode(), makeBlockNode(), makeListNode(), 
-		makeExprNode();
+static TreeNode makeListNode(), makeExprNode();
 static TreeNode makeTreeNode();
 static OOP	makeUnarySelector(), makeBinarySelector();
-static void	freeNode(), freeMethodNode(), freeBlockNode(), freeExprNode(),
-		freeListNode(), freeConstNode();
+
+#ifdef prev_obstacks
+/**/static void	freeNode(), freeMethodNode(), freeBlockNode(), freeExprNode(),
+/**/		freeListNode(), freeConstNode();
+#endif
+
 static void	printMethodNode(), printBlockNode(), printExprNode(), 
 		printListNode(), printConstNode(), printNodeType(), indent(),
-		printSelector();
+		printSelector(), printArrayConstructorNode();
 
 
 /* Used only for printing tree node names when debugging */
@@ -73,10 +78,10 @@ static char *nodeTypeNames[] = {
   "blockNodeType", 		/* blockNodeType */
   "cascadedMessageNodeType", 	/* cascadedMessageNodeType */
   "messageListType" 		/* messageListType */
+  "arrayConstructorType"	/* arrayConstructorType */
 };
 
 
-
 
 /*
  *	TreeNode makeArrayElt(elt)
@@ -135,8 +140,14 @@ makeMethod(selectorExpr, temporaries, primitiveIndex, statements)
      TreeNode selectorExpr, temporaries, statements;
      int	primitiveIndex;
 {
-  return (makeMethodNode(selectorExpr, temporaries,
-			 primitiveIndex, statements));
+  TreeNode	result;
+
+  result = makeTreeNode(methodNodeType);
+  result->vMethod.selectorExpr = selectorExpr;
+  result->vMethod.temporaries = temporaries;
+  result->vMethod.primitiveIndex = primitiveIndex;
+  result->vMethod.statements = statements;
+  return (result);
 }
 
 
@@ -306,7 +317,13 @@ TreeNode
 makeBlock(arguments, temporaries, statements)
      TreeNode arguments, temporaries, statements;
 {
-  return (makeBlockNode(arguments, temporaries, statements));
+  TreeNode	result;
+
+  result = makeTreeNode(blockNodeType);
+  result->vBlock.arguments = arguments;
+  result->vBlock.temporaries = temporaries;
+  result->vBlock.statements = statements;
+  return (result);
 }
 
 TreeNode
@@ -330,6 +347,19 @@ makeIntConstant(ival)
   return (result);
 }
 
+TreeNode
+makeByteObjectConstant(boval)
+     ByteObject boval;
+{
+  TreeNode 	result;
+
+  result = makeTreeNode(constExprType);
+  result->vConst.constType = byteObjConst;
+  result->vConst.val.boVal = boval;
+
+  return (result);
+}
+
 TreeNode
 makeFloatConstant(fval)
      double	fval;
@@ -343,19 +373,6 @@ makeFloatConstant(fval)
   return (result);
 }
 
-TreeNode
-makeCharConstant(cval)
-     char	cval;
-{
-  TreeNode 	result;
-
-  result = makeTreeNode(constExprType);
-  result->vConst.constType = charConst;
-  result->vConst.val.cVal = cval;
-
-  return (result);
-}
-
 TreeNode
 makeStringConstant(sval)
      char	*sval;
@@ -369,6 +386,32 @@ makeStringConstant(sval)
   return (result);
 }
 
+TreeNode
+makeOOPConstant(oval)
+     OOP oval;
+{
+  TreeNode 	result;
+
+  result = makeTreeNode(constExprType);
+  result->vConst.constType = oopConst;
+  result->vConst.val.oopVal = oval;
+
+  return (result);
+}
+
+TreeNode
+makeCharConstant(cval)
+     char	cval;
+{
+  TreeNode 	result;
+
+  result = makeTreeNode(constExprType);
+  result->vConst.constType = oopConst;
+  result->vConst.val.oopVal = charOOPAt(cval);
+
+  return (result);
+}
+
 TreeNode
 makeSymbolConstant(symbolNode)
      TreeNode symbolNode;
@@ -376,17 +419,65 @@ makeSymbolConstant(symbolNode)
   TreeNode 	result;
 
   result = makeTreeNode(constExprType);
-  result->vConst.constType = symbolConst;
+  result->vConst.constType = oopConst;
   if (symbolNode) {
-    result->vConst.val.symVal = symbolNode->vExpr.selector;
-    freeNode(symbolNode);
+    result->vConst.val.oopVal = symbolNode->vExpr.selector;
+#ifdef prev_obstacks
+/**/    freeNode(symbolNode);
+#endif
   } else {
-    result->vConst.val.symVal = nilOOP;
+    result->vConst.val.oopVal = nilOOP;
   }
 
   return (result);
 }
 
+/* This function converts an Array constant's format (linked list of its
+ * elements) to a ByteArray constant's format (ByteObject struct).  The code
+ * itself is awful and the list is extremely space inefficient, but consider
+ * that:
+ * a) it makes the parser simpler (Arrays and ByteArrays are treated in almost
+ *    the same way; only, the latter call this function and the former don't).
+ * b) a list is indeed an elegant solution because we don't know the size of
+ *    the byte array until we have parsed it all (that is, until we call this
+ *    function.
+ * c) the ByteObject is the best format for ByteArrays: first, it is the one
+ *    which makes it easiest to make a full-fledged object out of the parse
+ *    tree; second, it is logical to choose it since LargeIntegers use it,
+ *    and ByteArrays are represented exactly the same as LargeIntegers.
+ */
+TreeNode
+makeByteArrayConstant(aval)
+     TreeNode aval;
+{
+  TreeNode      arrayElt, ival;
+  int           len;
+  ByteObject	bo;
+  Byte		*data;
+
+  for(len = 0, arrayElt = aval; arrayElt;
+      len++, arrayElt = arrayElt->vList.next);
+
+  bo = (ByteObject)obstack_alloc(compilationObstack,
+    sizeof(ByteObjectStruct) + len);
+
+  bo->class = byteArrayClass;
+  bo->size = len;
+  data = bo->body;
+
+  /* Now extract the node for each integer constant, storing its value
+   * into the ByteObject */
+  for (arrayElt = aval; arrayElt; arrayElt = arrayElt->vList.next) {
+    ival = arrayElt->vList.value;
+    *data++ = ival->vConst.val.iVal;
+  }
+
+#ifdef prev_obstacks
+/**/    freeNode(aval);
+#endif
+  return (makeByteObjectConstant(bo));
+}
+
 TreeNode
 makeArrayConstant(aval)
      TreeNode aval;
@@ -400,6 +491,19 @@ makeArrayConstant(aval)
   return (result);
 }
 
+TreeNode
+makeArrayConstructor(statements)
+     TreeNode statements;
+{
+  TreeNode 	result;
+
+  result = makeTreeNode(arrayConstructorType);
+  result->vConst.constType = arrayConst;
+  result->vConst.val.aVal = statements;
+
+  return (result);
+}
+
 /*
  *	void addNode(n1, n2)
  *
@@ -434,43 +538,49 @@ freeTree(tree)
     return;
   }
 
-  switch (tree->nodeType) {
-  case methodNodeType:
-    freeMethodNode(tree);
-    break;
-    
-  case blockNodeType:
-    freeBlockNode(tree);
-    break;
+  obstack_free(compilationObstack, NULL);
+  obstack_init(compilationObstack);
 
-  case symbolNodeType:
-  case unaryExprType:
-  case binaryExprType:
-  case keywordExprType:
-  case cascadedMessageNodeType:
-  case statementListType:
-  case returnExprType:
-  case assignExprType:
-    freeExprNode(tree);
-    break;
-    
-  case variableNodeType:
-  case keywordListType:
-  case variableListType:
-  case arrayEltListType:
-  case messageListType:
-    freeListNode(tree);
-    break;
-
-  case constExprType:
-    freeConstNode(tree);
-    break;
-
-  }
+#ifdef prev_obstacks
+/**/  switch (tree->nodeType) {
+/**/  case methodNodeType:
+/**/    freeMethodNode(tree);
+/**/    break;
+/**/    
+/**/  case blockNodeType:
+/**/    freeBlockNode(tree);
+/**/    break;
+/**/
+/**/  case symbolNodeType:
+/**/  case unaryExprType:
+/**/  case binaryExprType:
+/**/  case keywordExprType:
+/**/  case cascadedMessageNodeType:
+/**/  case statementListType:
+/**/  case returnExprType:
+/**/  case assignExprType:
+/**/    freeExprNode(tree);
+/**/    break;
+/**/    
+/**/  case variableNodeType:
+/**/  case keywordListType:
+/**/  case variableListType:
+/**/  case arrayEltListType:
+/**/  case messageListType:
+/**/    freeListNode(tree);
+/**/    break;
+/**/
+/**/  case constExprType:
+/**/    freeConstNode(tree);
+/**/    break;
+/**/
+/**/  }
+#endif
 }
 
 
-
+
+
 
 /***********************************************************************
  *
@@ -479,34 +589,6 @@ freeTree(tree)
  ***********************************************************************/
 
 
-static TreeNode
-makeMethodNode(selectorExpr, temporaries, primitiveIndex, statements)
-     TreeNode selectorExpr, temporaries, statements;
-     int	primitiveIndex;
-{
-  TreeNode	result;
-
-  result = makeTreeNode(methodNodeType);
-  result->vMethod.selectorExpr = selectorExpr;
-  result->vMethod.temporaries = temporaries;
-  result->vMethod.primitiveIndex = primitiveIndex;
-  result->vMethod.statements = statements;
-  return (result);
-}
-
-static TreeNode
-makeBlockNode(arguments, temporaries, statements)
-     TreeNode arguments, temporaries, statements;
-{
-  TreeNode	result;
-
-  result = makeTreeNode(blockNodeType);
-  result->vBlock.arguments = arguments;
-  result->vBlock.temporaries = temporaries;
-  result->vBlock.statements = statements;
-  return (result);
-}
-
 static TreeNode
 makeListNode(nodeType, name, value)
      NodeType nodeType;
@@ -544,7 +626,9 @@ makeTreeNode(nodeType)
 {
   TreeNode result;
 
-  result = (TreeNode)xmalloc(sizeof(struct TreeNodeStruct));
+  result = (TreeNode)obstack_alloc(compilationObstack,
+    sizeof(struct TreeNodeStruct));
+
   result->nodeType = nodeType;
   return (result);
 }
@@ -565,7 +649,8 @@ makeBinarySelector(binaryOp)
   return (internString(binaryOp));
 }
 
-
+
+
 
 /***********************************************************************
  *
@@ -573,88 +658,97 @@ makeBinarySelector(binaryOp)
  *
  ***********************************************************************/
 
-static void
-freeMethodNode(node)
-     TreeNode node;
-{
-  freeTree(node->vMethod.selectorExpr);
-  freeTree(node->vMethod.temporaries);
-  freeTree(node->vMethod.statements);
-  freeNode(node);
-}
+#ifdef prev_obstacks
+/**/static void
+/**/freeMethodNode(node)
+/**/     TreeNode node;
+/**/{
+/**/  freeTree(node->vMethod.selectorExpr);
+/**/  freeTree(node->vMethod.temporaries);
+/**/  freeTree(node->vMethod.statements);
+/**/  freeNode(node);
+/**/}
+/**/
+/**/static void
+/**/freeBlockNode(node)
+/**/     TreeNode node;
+/**/{
+/**/  freeTree(node->vBlock.arguments);
+/**/  freeTree(node->vBlock.temporaries);
+/**/  freeTree(node->vBlock.statements);
+/**/  freeNode(node);
+/**/}
+/**/
+/**/static void
+/**/freeExprNode(node)
+/**/     TreeNode node;
+/**/{
+/**/  freeTree(node->vExpr.receiver);
+/**/  freeTree(node->vExpr.expression);
+/**/  freeNode(node);
+/**/}
+/**/
+/**/static void
+/**/freeListNode(node)
+/**/     TreeNode node;
+/**/{
+/**/  freeTree(node->vList.value);
+/**/  freeTree(node->vList.next);
+/**/  if (node->vList.name) {
+/**/    xfree(node->vList.name);
+/**/  }
+/**/
+/**/  freeNode(node);
+/**/}
+/**/
+/**/static void
+/**/freeConstNode(node)
+/**/     TreeNode node;
+/**/{
+/**/  switch (node->vConst.constType) {
+/**/  case intConst:
+/**/  case floatConst:
+/**/  case oopConst:
+/**/    /* these have no storage of their own */
+/**/    break;
+/**/
+/**/  case byteObjConst:
+/**/    if (node->vConst.val.boVal) {
+/**/      xfree(node->vConst.val.boVal);
+/**/    } else {
+/**/      errorf("Internal error: badly formed tree for byte object constant"); 
+/**/    }
+/**/    break;
+/**/
+/**/  case stringConst:
+/**/    if (node->vConst.val.sVal) {
+/**/      xfree(node->vConst.val.sVal);
+/**/    } else {
+/**/      errorf("Internal error: badly formed tree for string constant"); 
+/**/    }
+/**/    break;
+/**/
+/**/  case arrayConst:
+/**/    freeTree(node->vConst.val.aVal);
+/**/    break;
+/**/
+/**/  default:
+/**/    errorf("Internal error: corrupted tree structure"); 
+/**/  }
+/**/
+/**/  freeNode(node);
+/**/}
+/**/
+/**/
+/**/static void
+/**/freeNode(node)
+/**/     TreeNode node;
+/**/{
+/**/  xfree(node);
+/**/}
+/**/
+#endif
 
-static void
-freeBlockNode(node)
-     TreeNode node;
-{
-  freeTree(node->vBlock.arguments);
-  freeTree(node->vBlock.temporaries);
-  freeTree(node->vBlock.statements);
-  freeNode(node);
-}
-
-static void
-freeExprNode(node)
-     TreeNode node;
-{
-  freeTree(node->vExpr.receiver);
-  freeTree(node->vExpr.expression);
-  freeNode(node);
-}
-
-static void
-freeListNode(node)
-     TreeNode node;
-{
-  freeTree(node->vList.value);
-  freeTree(node->vList.next);
-  if (node->vList.name) {
-    xfree(node->vList.name);
-  }
-
-  freeNode(node);
-}
-
-static void
-freeConstNode(node)
-     TreeNode node;
-{
-  switch (node->vConst.constType) {
-  case intConst:
-  case floatConst:
-  case charConst:
-  case symbolConst:
-    /* these have no storage of their own */
-    break;
-
-  case stringConst:
-    if (node->vConst.val.sVal) {
-      xfree(node->vConst.val.sVal);
-    } else {
-      errorf("Internal error: badly formed tree for string constant"); 
-    }
-    break;
-
-  case arrayConst:
-    freeTree(node->vConst.val.aVal);
-    break;
-
-  default:
-    errorf("Internal error: corrupted tree structure"); 
-  }
-
-  freeNode(node);
-}
-
-
-static void
-freeNode(node)
-     TreeNode node;
-{
-  xfree(node);
-}
-
-
 
 /***********************************************************************
  *
@@ -702,6 +796,10 @@ printTree(node, level)
     printListNode(node, level);
     break;
 
+  case arrayConstructorType:
+    printArrayConstructorNode(node, level);
+    break;
+
   case constExprType:
     printConstNode(node, level);
     break;
@@ -712,6 +810,15 @@ printTree(node, level)
 }
 
 
+static void
+printArrayConstructorNode(node, level)
+     TreeNode node;
+     int level;
+{
+  printNodeType(node, level);
+  printTree(node->vConst.val.aVal, level+1);
+}
+
 static void
 printListNode(node, level)
      TreeNode node;
@@ -803,17 +910,13 @@ printConstNode(node, level)
     printf("float: %g\n", node->vConst.val.fVal);
     break;
 
-  case charConst:
-    printf("char: %c\n", node->vConst.val.cVal);
-    break;
-
   case stringConst:
     printf("string: \"%s\"\n", node->vConst.val.sVal);
     break;
 
-  case symbolConst:
-    printf("symbol: ");
-    printSymbol(node->vConst.val.symVal);
+  case oopConst:
+    printf("oop: ");
+    printObject(node->vConst.val.oopVal);
     printf("\n");
     break;
 
