@@ -2,15 +2,15 @@
  *
  *	Main Module
  *
- *	$Revision: 1.6.2$
- *	$Date: 1999/08/31 11:23:18$
+ *	$Revision: 1.8.5$
+ *	$Date: 2000/12/27 10:45:49$
  *	$Author: pb$
  *
  ***********************************************************************/
 
 /***********************************************************************
  *
- * Copyright 1990, 91, 92, 94, 95, 99 Free Software Foundation, Inc.
+ * Copyright 1988-92, 1994-95, 1999, 2000 Free Software Foundation, Inc.
  * Written by Steve Byrne.
  *
  * This file is part of GNU Smalltalk.
@@ -52,29 +52,3 @@ main(argc, argv)
 
   exit(0);
 }
-
-void testMain(argc, argv)
-     int	argc;
-     char 	**argv;
-{
-/*
-  char	*str;
-  OOP	o;
- */
-
-  smalltalkArgs(argc, argv);
-  initSmalltalk();
-
-#ifdef out_temp /* Sat Feb  8 17:08:18 1992 */
-/**/  o = evalCode("'foo on you' printNl!");
-/**/  msgSendf(nil, "%s %o inspect", o);
-#endif /* out_temp Sat Feb  8 17:08:18 1992 */
-#ifdef out_temp /* Thu Jan  2 22:23:07 1992 */
-/**/
-/**/  msgSendf(nil, "%s %s printNl", "This is a test");
-/**/  msgSendf(&str, "%s %s , %s", "This is a test", " ok?");
-/**/  msgSendf(nil, "%s %s printNl", str);
-#endif /* out_temp Thu Jan  2 22:23:07 1992 */
-  exit(0);
-}
-
